//==========================================================================
//####COPYRIGHTBEGIN####
//                                                                          
// -------------------------------------------                              
// The contents of this file are subject to the Red Hat eCos Public License 
// Version 1.1 (the "License"); you may not use this file except in         
// compliance with the License.  You may obtain a copy of the License at    
// <http://www.redhat.com/>                                                   
//                                                                          
// Software distributed under the License is distributed on an "AS IS"      
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the 
// License for the specific language governing rights and limitations under 
// the License.                                                             
//                                                                          
// The Original Code is eCos - Embedded Configurable Operating System,      
// released September 30, 1998.                                             
//                                                                          
// The Initial Developer of the Original Code is Red Hat.                   
// Portions created by Red Hat are                                          
// Copyright (C) 1998, 1999, 2000, 2001 Red Hat, Inc.                             
// All Rights Reserved.                                                     
// -------------------------------------------                              
//                                                                          
//####COPYRIGHTEND####
//==========================================================================

#ifndef ONCE_HAL_ARCH_H
#define ONCE_HAL_ARCH_H

#include "type.h"

//
// CPSR Register defines
//

#define CPSR_IRQ_DISABLE		0x80	// IRQ disabled when =1
#define CPSR_FIQ_DISABLE		0x40	// FIQ disabled when =1
#define CPSR_THUMB_ENABLE		0x20	// Thumb mode when =1
#define CPSR_FIQ_MODE			0x11
#define CPSR_IRQ_MODE			0x12
#define CPSR_SUPERVISOR_MODE	0x13
#define CPSR_UNDEF_MODE			0x1B

#define CPSR_MODE_BITS          0x1F

#define CPSR_INITIAL (CPSR_IRQ_DISABLE|CPSR_FIQ_DISABLE|CPSR_SUPERVISOR_MODE)
#define CPSR_THREAD_INITIAL (CPSR_SUPERVISOR_MODE)

//--------------------------------------------------------------------------
// Processor saved states:
// The layout of this structure is also defined in "arm.inc", for assembly
// code, which will be generated automatically if this file changes.

#define HAL_THREAD_CONTEXT_FIRST        0
#define HAL_THREAD_CONTEXT_R0           (0-HAL_THREAD_CONTEXT_FIRST)
#define HAL_THREAD_CONTEXT_R4           (4-HAL_THREAD_CONTEXT_FIRST)
#define HAL_THREAD_CONTEXT_R8           (8-HAL_THREAD_CONTEXT_FIRST)
#define HAL_THREAD_CONTEXT_R9           (9-HAL_THREAD_CONTEXT_FIRST)
#define HAL_THREAD_CONTEXT_R10          (10-HAL_THREAD_CONTEXT_FIRST)
#define HAL_THREAD_CONTEXT_LAST         10
#define HAL_NUM_THREAD_CONTEXT_REGS     (HAL_THREAD_CONTEXT_LAST - \
                                          HAL_THREAD_CONTEXT_FIRST+1)

// It seems that r0-r3,r12 are considered scratch by function calls

typedef struct 
{
    // These are common to all saved states
    UINT32  d[HAL_NUM_THREAD_CONTEXT_REGS] ;  // Data regs (r0..r10)
    UINT32  fp;                               // (r11) Frame pointer
    UINT32  ip;                               // (r12)
    UINT32  sp;                               // (r13) Stack pointer
    UINT32  lr;                               // (r14) Link Reg
    UINT32  pc;                               // (r15) PC place holder
                                                  //       (never used)
    UINT32  cpsr;                             // Condition Reg
    // These are only saved for exceptions and interrupts
    UINT32  vector;                           // Vector number
    UINT32  svc_sp;                           // saved svc mode sp

} HAL_SavedRegisters;

//-------------------------------------------------------------------------
// Exception handling function.
// This function is defined by the kernel according to this prototype. It is
// invoked from the HAL to deal with any CPU exceptions that the HAL does
// not want to deal with itself. It usually invokes the kernel's exception
// delivery mechanism.

externC void hal_deliver_exception( UINT32 code, UINT32 data );


//-------------------------------------------------------------------------
// Context Initialization
// Initialize the context of a thread.
// Arguments:
// _sparg_ name of variable containing current sp, will be changed to new sp
// _thread_ thread object address, passed as argument to entry point
// _entry_ entry point address.
// _id_ bit pattern used in initializing registers, for debugging.

#define HAL_THREAD_INIT_CONTEXT( _sparg_, _thread_, _entry_, _id_ )         \
    MACRO_START                                                         	\
    register UINT32 _sp_ = ((UINT32)_sparg_) &~15;                      	\
    register HAL_SavedRegisters *_regs_;                                    \
    int _i_;                                                                \
    _regs_ = (HAL_SavedRegisters *)((_sp_) - sizeof(HAL_SavedRegisters));   \
    for( _i_ = HAL_THREAD_CONTEXT_FIRST; _i_ <= HAL_THREAD_CONTEXT_LAST;    \
           _i_++ )                                                          \
        (_regs_)->d[_i_] = (_id_)|_i_;                                      \
    (_regs_)->d[00] = (UINT32)(_thread_); /* R0 = arg1 = thread ptr */    	\
    (_regs_)->sp = (UINT32)(_sp_);        /* SP = top of stack      */    	\
    (_regs_)->lr = (UINT32)(_entry_);     /* LR = entry point       */    	\
    (_regs_)->pc = (UINT32)(_entry_);     /* PC = [initial] entry point */	\
    (_regs_)->cpsr = (CPSR_THREAD_INITIAL); /* PSR = Interrupt enabled */   \
    _sparg_ = (UINT32)_regs_;                                          		\
    MACRO_END

//--------------------------------------------------------------------------
// Context switch macros.
// The arguments are pointers to locations where the stack pointer
// of the current thread is to be stored, and from where the sp of the
// next thread is to be fetched.

externC void hal_thread_switch_context( UINT32 to, UINT32 from );
externC void hal_thread_load_context( UINT32 to )
    __attribute__ ((noreturn));

#define HAL_THREAD_SWITCH_CONTEXT(_fspptr_,_tspptr_) \
        hal_thread_switch_context((UINT32)_tspptr_,  \
                                  (UINT32)_fspptr_);

#define HAL_THREAD_LOAD_CONTEXT(_tspptr_)            \
        hal_thread_load_context( (UINT32)_tspptr_ );

//--------------------------------------------------------------------------
// Execution reorder barrier.
// When optimizing the compiler can reorder code. In multithreaded systems
// where the order of actions is vital, this can sometimes cause problems.
// This macro may be inserted into places where reordering should not happen.

#define HAL_REORDER_BARRIER() asm volatile ( "" : : : "memory" )

//--------------------------------------------------------------------------
// Breakpoint support
// HAL_BREAKPOINT() is a code sequence that will cause a breakpoint to happen
// if executed.
// HAL_BREAKINST is the value of the breakpoint instruction and 
// HAL_BREAKINST_SIZE is its size in bytes.

#define _stringify1(__arg) #__arg
#define _stringify(__arg) _stringify1(__arg)

#define HAL_BREAKINST_ARM          0xE7FFDEFE
#define HAL_BREAKINST_ARM_SIZE     4
#define HAL_BREAKINST_THUMB        0xbebe  // illegal instruction currently
#define HAL_BREAKINST_THUMB_SIZE   2

extern UINT32 __arm_breakinst;
extern UINT16 __thumb_breakinst;
#define HAL_BREAKINST_ADDR(x) (((x)==2)? \
                              ((void*)&__thumb_breakinst) : \
                              ((void*)&__arm_breakinst))

//--------------------------------------------------------------------------
// Thread register state manipulation for GDB support.

// GDB expects the registers in this structure:
//   r0..r10, fp, ip, sp, lr, pc  - 4 bytes each
//   f0..f7                       - 12 bytes each (N/A on ARM7)
//   fps                          - 4 bytes       (N/A on ARM7)
//   ps                           - 4 bytes

// Translate a stack pointer as saved by the thread context macros above into
// a pointer to a HAL_SavedRegisters structure.
#define HAL_THREAD_GET_SAVED_REGISTERS( _sp_, _regs_ )  \
        (_regs_) = (HAL_SavedRegisters *)(_sp_)


//--------------------------------------------------------------------------
// HAL setjmp

#define ARC_JMP_BUF_SIZE 16  // Actually 11, but some room left over

typedef UINT32 hal_jmp_buf[ARC_JMP_BUF_SIZE];

externC int hal_setjmp(hal_jmp_buf env);
externC void hal_longjmp(hal_jmp_buf env, int val);

//---------------------------------------------------------------------------
// Table definition macros.
// The ARM assembler has a slightly different syntax for some of this, so
// we have to define alternatives here.
// Note that the __string() and __xstring() macros used here are defined in
// hal_tables.h.

#ifndef HAL_TABLE_BEGIN
#define HAL_TABLE_BEGIN( _label, _name )                                    \
__asm__(".section \"" __string(.ecos.table.##_name##.begin) "\",\"aw\"\n"   \
	".globl " __xstring(LABEL_DEFN(_label)) "\n"                        	\
	".type    " __xstring(LABEL_DEFN(_label)) ",object\n"               	\
	".p2align 2\n"                                                          \
__xstring(LABEL_DEFN(_label)) ":\n"                                         \
	".previous\n"                                                           \
       )
#endif

#ifndef HAL_TABLE_END
#define HAL_TABLE_END( _label, _name )                                      \
__asm__(".section \"" __string(.ecos.table.##_name##.finish) "\",\"aw\"\n"  \
	".globl " __xstring(LABEL_DEFN(_label)) "\n"                        	\
	".type    " __xstring(LABEL_DEFN(_label)) ",object\n"               	\
	".p2align 2\n"                                                          \
__xstring(LABEL_DEFN(_label)) ":\n"                                         \
	".previous\n"                                                           \
       )
#endif

//---------------------------------------------------------------------------

// Minimal and sensible stack sizes: the intention is that applications
// will use these to provide a stack size in the first instance prior to
// proper analysis.  Idle thread stack should be this big.

//    THESE ARE NOT INTENDED TO BE MICROMETRICALLY ACCURATE FIGURES.
//           THEY ARE HOWEVER ENOUGH TO START PROGRAMMING.
// YOU MUST MAKE YOUR STACKS LARGER IF YOU HAVE LARGE "AUTO" VARIABLES!

// This is not a config option because it should not be adjusted except
// under "enough rope" sort of disclaimers.

// A minimal, optimized stack frame, rounded up - no autos
#define HAL_STACK_FRAME_SIZE (4 * 20)

// Stack needed for a context switch: this is implicit in the estimate for
// interrupts so not explicitly used below:
#define HAL_STACK_CONTEXT_SIZE (4 * 20)

// Interrupt + call to ISR, interrupt_end() and the DSR
#define HAL_STACK_INTERRUPT_SIZE \
    ((4 * 20) + 2 * HAL_STACK_FRAME_SIZE)

// Space for the maximum number of nested interrupts, plus room to call functions
#define HAL_MAX_INTERRUPT_NESTING 4

#define HAL_STACK_SIZE_MINIMUM \
        (HAL_MAX_INTERRUPT_NESTING * HAL_STACK_INTERRUPT_SIZE + \
         2 * HAL_STACK_FRAME_SIZE)

#define HAL_STACK_SIZE_TYPICAL \
        (HAL_STACK_SIZE_MINIMUM + \
         16 * HAL_STACK_FRAME_SIZE)

//--------------------------------------------------------------------------
// Macros for switching context between two eCos instances (jump from
// code in ROM to code in RAM or vice versa).
#define ARC_HAL_SAVE_GP()
#define ARC_HAL_RESTORE_GP()

#endif // ONCE_HAL_ARCH_H
// End of hal_arch.h
