#ifndef ONCE_PCI_H
#define ONCE_PCI_H
//=============================================================================
//
//      pci.h
//
//      PCI library
//
//=============================================================================
//####COPYRIGHTBEGIN####
//                                                                          
// -------------------------------------------                              
// The contents of this file are subject to the Red Hat eCos Public License 
// Version 1.1 (the "License"); you may not use this file except in         
// compliance with the License.  You may obtain a copy of the License at    
// http://www.redhat.com/                                                   
//                                                                          
// Software distributed under the License is distributed on an "AS IS"      
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the 
// License for the specific language governing rights and limitations under 
// the License.                                                             
//                                                                          
// The Original Code is eCos - Embedded Configurable Operating System,      
// released September 30, 1998.                                             
//                                                                          
// The Initial Developer of the Original Code is Red Hat.                   
// Portions created by Red Hat are                                          
// Copyright (C) 1998, 1999, 2000 Red Hat, Inc.                             
// All Rights Reserved.                                                     
// -------------------------------------------                              
//                                                                          
//####COPYRIGHTEND####
//=============================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    jskov, from design by nickg 
// Contributors: jskov
// Date:         1999-08-09
// Purpose:      PCI configuration access
// Usage:
//              #include </io/pci.h>
// Description: 
//             This library provides a set of routines for accessing 
//             the PCI bus configuration space in a portable manner. 
//             This is provided by two APIs. The high level API (defined
//             by this file) is used by device drivers, or other code, to
//             access the PCI configuration space portably. The low level
//             API (see pci_hw.h) is used  by the PCI library itself
//             to access the hardware in a platform-specific manner and
//             may also be used by device drivers to access the PCI
//             configuration space directly.
//
//####DESCRIPTIONEND####
//
//=============================================================================

#include "pci_cfg.h"
#include "type.h"

//------------------------------------------------------------------
// The PCI memory space can span 64 bits, IO space only 32 bits
typedef ULONG PCI_ADDRESS64;
typedef UINT32 PCI_ADDRESS32;

//------------------------------------------------------------------
// Macros for manufacturing and decomposing device ids
typedef UINT32 pci_device_id;           // PCI device ID

#define PCI_DEV_MAKE_ID(__bus,__devfn) (((__bus)<<16)|((__devfn)<<8))
#define PCI_DEV_GET_BUS(__devid) ((__devid>>16)&0xFF)
#define PCI_DEV_GET_DEVFN(__devid) ((__devid>>8)&0xFF)

#define PCI_NULL_DEVID 0xffffffff
#define PCI_NULL_DEVFN 0xffff

//------------------------------------------------------------------
// PCI device data definitions and structures

typedef enum { 
    PCI_HEADER_NORMAL = 0, 
    PCI_HEADER_BRIDGE = 1,
    PCI_HEADER_CARDBUS_BRIDGE = 2
} pci_header_type;

typedef struct                          // PCI device data
{
    pci_device_id devid;            // ID of this device

    // The following fields are read out of the
    // config space for this device.

    UINT16  vendor;                 // vendor ID
    UINT16  device;                 // device ID
    UINT16  command;                // command register
    UINT16  status;                 // status register
    UINT32  class_rev;              // class+revision
    UINT8   cache_line_size;        // cache line size
    UINT8   latency_timer;          // latency timer
    pci_header_type header_type;    // header type
    UINT8   bist;                   // Built-in Self-Test
    UINT32  base_address[6];        // Memory base address registers

    // The following fields are used by the resource allocation
    // routines to keep track of allocated resources.
    UINT32 num_bars;
        
    UINT32 base_size[6];            // Memory size for each base address
    UINT32 base_map[6];             // Physical address mapped

    UINT32 hal_vector;            // HAL interrupt vector used by
                                        // device if int_line!=0

    // One of the following unions will be filled in according to
    // the value of the header_type field.

    union
    {
        struct
        {
            UINT32  cardbus_cis;    // CardBus CIS Pointer
            UINT16  sub_vendor;     // subsystem vendor id
            UINT16  sub_id;         // subsystem id
            UINT32  rom_address;    // ROM address register
            UINT8   cap_list;       // capability list
            UINT8   reserved1[7];
            UINT8   int_line;       // interrupt line
            UINT8   int_pin;        // interrupt pin
            UINT8   min_gnt;        // timeslice request
            UINT8   max_lat;        // priority-level request
        } normal;
        struct
        {
	    UINT8	pri_bus;		// primary bus number
	    UINT8	sec_bus;		// secondary bus number
	    UINT8	sub_bus;		// subordinate bus number
	    UINT8	sec_latency_timer;	// secondary bus latency
	    UINT8	io_base;
	    UINT8	io_limit;
	    UINT16	sec_status;		// secondary bus status
	    UINT16	mem_base;
	    UINT16	mem_limit;
	    UINT16	prefetch_base;
	    UINT16	prefetch_limit;
	    UINT32	prefetch_base_upper32;
	    UINT32	prefetch_limit_upper32;	
	    UINT16	io_base_upper16;
	    UINT16	io_limit_upper16;
            UINT8   reserved1[4];
            UINT32  rom_address;    	// ROM address register
            UINT8   int_line;       	// interrupt line
            UINT8   int_pin;        	// interrupt pin
	    UINT16	control;		// bridge control
        } bridge;
        struct
        {
            // Not yet supported
        } cardbus_bridge;
    } header;
} pci_device;
    
//------------------------------------------------------------------------
// Init
externC void pci_init( void );

//------------------------------------------------------------------------
// Common device configuration access functions

// This function gets the PCI configuration information for the
// device indicated in devid. The common fields of the pci_device
// structure, and the appropriate fields of the relevant header union
// member are filled in from the device's configuration space. If the
// device has not been enabled, then this function will also fetch
// the size and type information from the base address registers and
// place it in the base_size[] array.
externC void pci_get_device_info ( pci_device_id devid, 
                                        pci_device *dev_info );


// This function sets the PCI configuration information for the
// device indicated in devid. Only the configuration space registers
// that are writable are actually written. Once all the fields have
// been written, the device info will be read back into *dev_info, so
// that it reflects the true state of the hardware.
externC void pci_set_device_info ( pci_device_id devid, 
                                       pci_device *dev_info );

//------------------------------------------------------------------------
// Device find functions

// Searches the PCI bus configuration space for a device with the
// given vendor and device ids. The search starts at the device
// pointed to by devid, or at the first slot if it contains
// PCI_NULL_DEVID. *devid will be updated with the ID of the next
// device found. Returns true if one is found and false if not.
externC int pci_find_device( UINT16 vendor, UINT16 device,
                                      pci_device_id *devid );


// Searches the PCI bus configuration space for a device with the
// given class code.  The search starts at the device pointed to by
// devid, or at the first slot if it contains PCI_NULL_DEVID.
// *devid will be updated with the ID of the next device found.
// Returns true if one is found and false if not.
externC int pci_find_class( UINT32 dev_class,
                                     pci_device_id *devid );



// Searches the PCI bus configuration space for a device whose properties
// match those required by the match_func, which the user supplies.  The
// match_func's arguments are vendor, device, class exactly as they might
// be in the two APIs above.  The additional parameter is for any state
// which a caller might wish available to its callback routine.
// The search starts at the device pointed to by
// devid, or at the first slot if it contains PCI_NULL_DEVID.  *devid
// will be updated with the ID of the next device found.  Returns true if
// one is found and false if not.
typedef int (pci_match_func)( UINT16,/* vendor */
                                       UINT16,/* device */
                                       UINT32,/* class  */
                                       void * /* arbitrary user data */ );
externC int pci_find_matching( pci_match_func *matchp,
                                        void * match_callback_data,
                                        pci_device_id *devid );


// Searches the PCI configuration space for the next valid device
// after cur_devid. If cur_devid is given the value
// PCI_NULL_DEVID, then the search starts at the first slot. It
// is permitted for next_devid to point to the cur_devid.  Returns
// true if another device is found and false if not.
externC int pci_find_next( pci_device_id cur_devid, 
                                    pci_device_id *next_devid );

//------------------------------------------------------------------------
// Resource Allocation
// These routines allocate memory and IO space to PCI devices.

// Allocate memory and IO space to all base address registers using
// the current memory and IO base addresses in the library. If
// dev_info does not contain valid base_size[] entries, then the
// result is false.
externC int pci_configure_device( pci_device *dev_info );

// Allocate memory and IO space for all devices found on the given
// bus and its subordinate busses. This routine recurses when a
// PCI-to-PCI bridge is encountered. The next_bus argument points
// to a variable holding the bus number of the next PCI bus to
// be allocated when a bridge is encountered. This routine returns
// true if successful, false if unsuccessful.
externC int pci_configure_bus( UINT8 bus,
					UINT8 *next_bus );

// These routines set the base addresses for memory and IO mappings
// to be used by the memory allocation routines. Normally these base
// addresses will be set to default values based on the platform,
// these routines allow those to be changed by application code if
// necessary.
externC void pci_set_memory_base( PCI_ADDRESS64 base, int addr_ext);
externC void pci_set_io_base( PCI_ADDRESS32 base );

// These routines allocate memory or IO space to the base address
// register indicated by bar. The base address in *base will be
// correctly aligned and the address of the next free location will
// be written back into it if the allocation succeeds. If the base
// address register is of the wrong type for this allocation, or
// dev_info does not contain valid base_size[] entries, the result is
// false.
externC int pci_allocate_memory( pci_device *dev_info,
                                          UINT32 bar, 
                                          PCI_ADDRESS64 *base );
externC int pci_allocate_io( pci_device *dev_info,
                                      UINT32 bar, 
                                      PCI_ADDRESS32 *base );

// Translate the device's PCI interrupt (INTA#-INTD#) to the
// associated HAL vector. This may also depend on which slot the
// device occupies. If the device may generate interrupts, the
// translated vector number will be stored in vec and the result is
// true. Otherwise the result is false.
externC int pci_translate_interrupt( pci_device *dev_info,
                                              UINT32 *vec );


//----------------------------------------------------------------------
// Specific device configuration access functions

// Read functions
// These functions read registers of the appropriate size from the
// configuration space of the given device. They should mainly be
// used to access registers that are device specific. General PCI
// registers are best accessed through pci_get_device_info().
externC void pci_read_config_uint8( pci_device_id devid,
             				              UINT8 offset, UINT8 *val);
externC void pci_read_config_uint16( pci_device_id devid,
             				              UINT8 offset, UINT16 *val);
externC void pci_read_config_uint32( pci_device_id devid,
										  UINT8 offset, UINT32 *val);

// Write functions
// These functions write registers of the appropriate size to the
// configuration space of the given device. They should mainly be
// used to access registers that are device specific. General PCI
// registers are best accessed through
// pci_get_device_info(). Writing the general registers this way
// may render the contents of a pci_device structure invalid.
externC void pci_write_config_uint8( pci_device_id devid,
                                         UINT8 offset, UINT8 val);
externC void pci_write_config_uint16( pci_device_id devid,
                                          UINT8 offset, UINT16 val);
externC void pci_write_config_uint32( pci_device_id devid,
                                          UINT8 offset, UINT32 val);


//----------------------------------------------------------------------
// Functions private to the PCI library. These should only be used by
// tests.
externC int pci_allocate_memory_priv(pci_device *dev_info,
                                              UINT32 bar,
                                              PCI_ADDRESS64 *base,
                                             PCI_ADDRESS64 *assigned_addr);
externC int pci_allocate_io_priv( pci_device *dev_info,
                                           UINT32 bar, 
                                           PCI_ADDRESS32 *base,
                                           PCI_ADDRESS32 *assigned_addr);


//----------------------------------------------------------------------
// Bus probing limits.
#define PCI_MAX_BUS                      8  // Eight is enough?
#define PCI_MAX_DEV                      8
#define PCI_MAX_FN                       8
#define PCI_MAX_BAR                      6

//-----------------------------------------------------------------------------
#endif // ifndef ONCE_PCI_H
// End of pci.h
