/* syslog.c
---------------------------------------------------------------------------
                 Copyright (c) 2002, 2003 Intel Corporation
						 All rights reserved.
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer. 
* Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution. 
* Neither the name of Intel Corporation nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission. 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                      
---------------------------------------------------------------------------
system: IXP2400
subsystem: DIAG
author: gvaddadi, Mar 6, 02
revisions: ccm, April 21, 02
		   dalsraja, June 6, 02
--------------------------------------------------------------------------
*/

#include "syslog.h"
#include "misc_func.h"
#include "hal_ixdp2400.h"

extern void eprintf(char *f, ...);
extern int flash_erase(void *addr, int len, void **err_addr);
extern int flash_program(void *_addr, void *_data, int len, void **err_addr);

static int CURRENT_VERBOSE_LEVEL = 0;


/*******************************************************************************
Function: syslog_init() will initialize the System Log Buffer with the cookie.
The value 0x54525545 is used as cookie which is represents the ASCII 'TRUE'. Each
time the syslog is wriitten or read, it is checked.
Arguments:None
Return Value:None
********************************************************************************/
void syslog_init(void)
{
	void *err_addr;
	SYSLOG_CNTRL_BLOCK cntrl_block;
	
	eprintf("Initialising Syslog...\n");

	cntrl_block.cookie = SYSLOG_VALID;
	cntrl_block.rsvd1 = cntrl_block.rsvd2 /*= cntrl_block.rsvd3 =  cntrl_block.rsvd4 = cntrl_block.rsvd5 = cntrl_block.rsvd6 = cntrl_block.rsvd7 */= 0;

	flash_program((void *)(SYSLOG_START_ADDR), (unsigned char *)&(cntrl_block),
				sizeof(SYSLOG_CNTRL_BLOCK), (void **)&err_addr);
	
	return;
}


/*******************************************************************************
Function: syslog_chg_verbose(void) will display the current verbose level and 
allows to change it.
Arguments: None
Return Vaue: SYSLOG_OK if change is sucess else SYSLOG_ERROR
********************************************************************************/
void syslog_chg_verbose(int new_level)
{
	eprintf("Current Verbose level is %d\n", CURRENT_VERBOSE_LEVEL);

	if (syslog_validity_chk() == SYSLOG_ERROR)
	{
		syslog_clear();
		syslog_init();
	}		

	CURRENT_VERBOSE_LEVEL = new_level;

	eprintf("\nNew Verbose level is %d\n", CURRENT_VERBOSE_LEVEL);

	return;	
}


/*******************************************************************************
Function: syslog_validity_chk(void) will check for the cookie 0x54525545 at 
SYSLOG_START_ADDR.
Arguments: None
Return Vaue: SYSLOG_OK if cookie matches else SYSLOG_ERROR
********************************************************************************/
int syslog_validity_chk(void)
{
	unsigned char buf[4];
	unsigned int cookie = SYSLOG_VALID;
	
	flash_read((void *)SYSLOG_START_ADDR, buf, 4);

    if (strncmp(buf, (char *)&cookie, 4) == 0)
	{
        return SYSLOG_OK;
	}
    else
	{
        return SYSLOG_ERROR;
	}
}


/*********************************************************************************
Function: syslog_clear() will erase all the log entries when requested by the user. 
Arguments: None
Return Value: none
***********************************************************************************/

void syslog_clear(void) 
{
	void *err_addr;

	eprintf("Clearing Syslog...\n");
	flash_erase((void*)SYSLOG_START_ADDR, SYSLOG_SIZE, (void*)&err_addr);
	return;
}


/*************************************************************************************
Function: syslog_show() will display the log entries when requested by the user. 
If the user enters the optional parameter, which indicates how many entries to display, 
it displays those many entries form the last. By default it display all the entries.
Arguments: no_of_entries - number of entries to display from the lats entry. 
			0 will display all entries
Return Value: none

**************************************************************************************/

void syslog_show(int num_of_entries_to_display)
{
//	volatile unsigned char* start;
	int status,i;
	static SYSLOG_INFO info;
//	int no_of_entries = 0;
	register PSYSLOG_DATA buf;
	
	if(syslog_validity_chk())
	{
		status = compute_last_logentry_ptr(&info);
	
//		eprintf("From syslog show: status =%d, last_entry = %x, no=%d\n",status, info.last_entry, info.num_of_entries);

	/*Compute the address of the starting logentry to be displayed from NoOfLogEntries 
	and last_entry.*/
	
		if(status==SYSLOG_EMPTY)
		{
			return;
		}

/*		if((no_of_entries != 0) && (no_of_entries < info.num_of_entries) && (status == SYSLOG_OK))
		{
			start = info.last_entry - (64 * no_of_entries);	
		}
		else
		{
			start = (volatile unsigned char *)SYSLOG_START_ADDR;
			no_of_entries = info.num_of_entries;
		}
*/

		if (num_of_entries_to_display > info.num_of_entries)
		{
			num_of_entries_to_display = info.num_of_entries;
		}

		for(i = 0; i < num_of_entries_to_display; i++)
		{
//			eprintf(" Showing entry at 0x%x\n",((unsigned int)FIRST_ENTRY_PTR + (i * 64)));
			eprintf("Entry number %d\n", i);
			buf = (SYSLOG_DATA *)(FIRST_ENTRY_PTR + (i * 64));
			syslog_display(buf);
		}
	}/*syslog_validity_chk*/
	else
	{
		eprintf("Syslog not valid!!!\n");
	}

}


/*****************************************************************************
Function: syslog_display(0 will display the syslog entry in the required format.
Arguments: ptr to syslog entry to be displayed
Return value: none
******************************************************************************/

void syslog_display(SYSLOG_DATA *x)
{
//		int timestamp;
//		unsigned char type;
//		unsigned char source_comp;
//		unsigned short int error_code;
//		char desc[56];

//	eprintf("Timestamp: %d\n",x->timestamp);
	
	switch(x->type)
	{
		case 1:	eprintf(" Type:	Error\n");
				break;
		case 2:	eprintf(" Type:	Warning\n");
				break;
		case 3:	eprintf(" Type:	DBG MSG1\n");
				break;
		case 4:	eprintf(" Type:	DBG MSG2\n");
				break;
		case 5:	eprintf(" Type:	DBG MSG3\n");
				break;
	}
	
	switch(x->source_comp)
	{
		case 1:	eprintf(" Source:	Master_BM\n");
				break;
		case 2:	eprintf(" Source:	Master_Diag\n");
				break;
		case 5:	eprintf(" Source:	Slave_BM\n");
				break;
		case 6:	eprintf(" Source:	Slave_Diag\n");
				break;
	}
	
//	eprintf("Errorcode:	%x\n", x->error_code);

	eprintf(" Desc:	%s\n", x->desc);
	eprintf("\n**********\n");
	return;
	
}


/***********************************************************************************
Function: compute_last_logentry_ptr() will search for a 64 byte window to compute the 
the addr of the last entry in the syslog.
Arguments: SYSLOG_INFO to update the last_entry addr and no. of entries
Return Value: SYSLOG_FULL if syslog is full, SYSLOG_OK if succesfull in computing last_entry addr 
SYSLOG_EMPTY- if syslog is empty.
************************************************************************************/

int compute_last_logentry_ptr(SYSLOG_INFO *x)
{	
	/*status 0- error, 1-success, 2 syslog emty*/
	static char buf[64];
	int status = SYSLOG_ERROR, cnt, i;
	x->num_of_entries = 0;

	x->last_entry = (unsigned char *)FIRST_ENTRY_PTR;
	cnt = 0;

	while((cnt != 64) && (x->last_entry < (unsigned char *)(SYSLOG_START_ADDR + SYSLOG_SIZE - 1)))
	{
		flash_read(x->last_entry, &buf[0], 64);

		cnt = 0;
		for(i = 0; i < 64; i++)
		{
			if(buf[i] != (char)0xFF)
			{
				break;
			}
			else
			{
				cnt++;
			}
		}

		if(cnt == 64)
		{
//			eprintf("Found the empty window!! at 0x%x\n", x->last_entry);
			status = SYSLOG_OK;
			break;
		}
		else
		{
			x->last_entry+=64;
			x->num_of_entries++;
		}
	}

	
	if(x->last_entry == (unsigned char *)FIRST_ENTRY_PTR)
	{
		eprintf("Syslog is empty!!!!\n");
		status = SYSLOG_EMPTY;
	}
	
	if(x->last_entry >= (unsigned char*)(SYSLOG_START_ADDR + SYSLOG_SIZE - SYSLOG_ENTRY_SIZE - 1))
	{
		status = SYSLOG_FULL;
		eprintf("Syslog is full\n");
	}
	else if((x->last_entry + (5 * 64)) > (unsigned char *)(SYSLOG_START_ADDR + SYSLOG_SIZE - 1))
	{
		eprintf("Warning: Syslog will be full soon and will be erased!!!\n");
	}

	return status;
}


/*************************************************************************
Function: syslog_dump() will store a syslog entry to flash.
Arguments: SYSLOG_DATA -buf containing the log entry 
			size- size of entry which is 64 bytes
Return value: SYSLOG_OK if sucessful
			  SYSLOG_ERROR if error
**************************************************************************/

int syslog_dump(SYSLOG_DATA *buf, int len)
{
	void *err_addr;

	int current_logging_level = CURRENT_VERBOSE_LEVEL;

	SYSLOG_INFO x;
	if(current_logging_level == LOG_OFF)
	{
		eprintf("*** System Log is off currently ***\n");
		return SYSLOG_ERROR;
	}
	else
	{
		if(buf->type > current_logging_level)
		{
			eprintf("Inappropriate Logging/Verbose level\n");
			return SYSLOG_ERROR;
		}

		if(syslog_validity_chk())
		{
			if(compute_last_logentry_ptr(&x))
			{
				flash_program(x.last_entry, (char *)buf, 64, (void*)&err_addr);
//				eprintf("addr = 0x%X\n", (unsigned int)x.last_entry);
			}
			else
			{
				eprintf("Syslog Full !!! Erasing Syslog....\n");
				syslog_clear();
				syslog_init();

				flash_program((void*)FIRST_ENTRY_PTR, (char *)buf, 64, (void*)&err_addr);
			}
		}
	}

	return SYSLOG_OK;
}
