/* uc_load.c
 *
 *------------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *------------------------------------------------------------------------------
 *
 *
 *  system: IXP2400
 *  subsystem: DIAG
 *  author: dalsraja, March, 2002
 *  revisions: dalsraja, 1 May 2002 - incorporated code review comment changes
 * 
 * 
 * -----------------------------------------------------------------------------
 */

#include "common.h"
#include "diagstruct.h"
#include "diag_utils.h"
#include "uc_load.h"

extern UINT32 t_all_verbose;


static int GetUengBase(unsigned short Uengine);
static int IsUengActive(unsigned short Uengine);
static int IsUengCtxEnabled(unsigned short Uengine);
static void Reset_Uengine(unsigned short Uengine);
static void ClearReset_Uengine(unsigned short Uengine);

//*****************************************************************************
// Function:	int GetUengBase(unsigned short Uengine)
// Description:	Gets the CSR base address of a specific microengine
// Inputs:		Uengine -> Microengine numbers Sausalito 0 - 4 and 16 - 19
// 				Uengine -> Microengine numbers Castine, i.e. 0 - 7 and 16 - 23
// Outputs:		Returns the CSR base address of the microengine
//*****************************************************************************
static int GetUengBase(unsigned short Uengine)
{
#ifndef CASTINE		// if Sausalito
	if ((Uengine <= 3) || ((Uengine >= 16) && (Uengine <= 19)))
#else				// if Castine
	if ((Uengine <= 7) || ((Uengine >= 16) && (Uengine <= 23)))
#endif
	{
		return UENG_CSR_BASE_ADDR(Uengine);
	}
	else
	{
		return UENG_INVALID;
	}
}


//************************************************************************
// Function:	int IsUengActive(unsigned short Uengine)
// Description:	Check the state of the microengine
// Inputs:		Uengine -> Microengine numbers Sausalito 0 - 4 and 16 - 19
// 				Uengine -> Microengine numbers Castine 0 - 7 and 16 - 23
// Outputs:		Returns the state of the microengine
//************************************************************************
static int IsUengActive(unsigned short Uengine)
{
	if(GetUengBase(Uengine) == UENG_INVALID)
	{
		return(UENG_INVALID);
	}

	if((GET32(RESET_1_REG) & (1 << Uengine)) == 0)
	{
		return(UENG_NOT_RESET);
	}
	
	return(UENG_RESET);
}


//*************************************************************************
// Function:	int IsUengCtxEnabled(unsigned short Uengine)
// Description:	Check if the contects of the threads are enabled
// Inputs:		Uengine -> Microengine numbers Sausalito 0 - 4 and 16 - 19
// 				Uengine -> Microengine numbers Castine 0 - 7 and 16 - 23
// Outputs:		Returns the state of the microengine
//*************************************************************************
static int IsUengCtxEnabled(unsigned short Uengine)	// IXDP2400 - The whole function is new
{
	int ctx_enables_val, ctr, ueng_base;

	ueng_base = GetUengBase(Uengine);
	
	ctx_enables_val = ((GET32(ueng_base + CTX_ENABLES) & CTX_ENABLE_MASK) >> 8);
	
	if(ctx_enables_val != 0)			// if any of the ctx_enables bits are set
	{									// display a message stating which ctx is still active
		eprintf("Context ");
		
		for (ctr = 0; ctr < 8; ctr++)
		{
			if (((ctx_enables_val >> ctr) & 1) == 1)	// if the context is enabled, display the context number
			{
				eprintf("%d, ", ctr);
			}
		}

		eprintf("are active\n");

		return(UENG_ACTIVE);
	}

	return(UENG_INACTIVE);
}


//*************************************************************************
// Function:	void Reset_Uengine(unsigned short Uengine)
// Description:	Calling this function will assert a reset on the 
//				microengine specified
// Inputs:		Uengine -> Microengine numbers Sausalito 0 - 4 and 16 - 19
// 				Uengine -> Microengine numbers Castine 0 - 7 and 16 - 23
// Outputs:		None
//*************************************************************************
static void Reset_Uengine(unsigned short Uengine)
{
	int reset_reg_val;

	reset_reg_val = GET32(RESET_1_REG) | (1 << Uengine);

	PUT32(RESET_1_REG, reset_reg_val);		// Assert the reset for the uengine
}


//*************************************************************************
// Function:	void ClearReset_Uengine(unsigned short Uengine)
// Description:	Calling this function will de-assert the reset on
//				the specified microengine
// Inputs:		Uengine -> Microengine numbers Sausalito 0 - 4 and 16 - 19
// 				Uengine -> Microengine numbers Castine 0 - 7 and 16 - 23
// Outputs:		None
//*************************************************************************
static void ClearReset_Uengine(unsigned short Uengine)
{
	int reset_reg_val;

	reset_reg_val = GET32(RESET_1_REG) & ~(1 << Uengine);

	PUT32(RESET_1_REG, reset_reg_val);		// De-assert the reset for the unegine
}


// IXDP2400:Consider the whole function below to be new as the loading
//			of microcode to the microstore is different compared to IXP1200
//***********************************************************************************************************
// Function:	int DownLoadUcode(unsigned short Uengine, unsigned short SizeUcode, unsigned int *Ucode)
// Description:	Calling this function will load the microcode into the microstore of a specified microengine
// Inputs:		Uengine -> Microengine numbers Sausalito 0 - 4 and 16 - 19
// 				Uengine -> Microengine numbers Castine 0 - 7 and 16 - 23
//				SizeUcode	->	Bytesize of the microcode to be loaded
//				*Ucode		->	Pointer to the array holding the microcode data to be loaded to the ustore
// Outputs:		Return the status of the operation
//***********************************************************************************************************
int DownLoadUcode(unsigned short Uengine, unsigned short SizeUcode, unsigned int *Ucode)
{
	register PDiagCommon acL = (PDiagCommon) ACADDRESS;
	int status, Uaddr, ueng_base, i, Ucode_ele;

	Ucode_ele = SizeUcode / 8;	// Tot num of elements(rows) to write into the ustore
	if(t_all_verbose != MINIMAL)
		eprintf("Size of ucode: %d for Uengine %d\n", Ucode_ele, Uengine);

	ueng_base = GetUengBase(Uengine);
	
	//  Check if the specified microengine value is a valid microengine number
	if(ueng_base == UENG_INVALID)
	{
		eprintf("Microengine %d is not a valid microengine number\n", Uengine);
		return(UENG_INVALID);
	}

	// Check the size of the microcode to be loaded into the microstore with the actual
	// size of the mircostore
	// Note: acL->UStoreSize is a struct defined in diagstruct.h and the value for it
	//		 can be found in test_mngr.c in function Diag_Struct_Init()
	if(Ucode_ele > acL->UStoreSize)
	{
		eprintf("Size of microcode (%d bytes) exceeds size of microstore (%d bytes)\n", SizeUcode, acL->UStoreSize);
		return(UENG_BADSIZE);
	}

	// Set the specified microengine to the Reset state by setting the respective
	// bit on the Reset_1 register
	Reset_Uengine(Uengine);

	hal_delay_us(25000);				// delay for 25 ms

	// Check if the microengine specified is in the reset mode
	status = IsUengActive(Uengine);
	if(status != UENG_RESET)
	{
		if (status == UENG_NOT_RESET)
		{
			eprintf("Microengine %d is not in reset state\n", Uengine);
		}

		return(status);
	}

	// Clear the reset bit of the respective microengine in the RESET_1 register
	ClearReset_Uengine(Uengine);

	hal_delay_us(25000);				// delay for 25 ms

	// Check if the contexts of the microengine specified have all been disabled
	status = IsUengCtxEnabled(Uengine);
	if(status != UENG_INACTIVE)
	{
		if (status == UENG_ACTIVE)
		{
			eprintf("Microengine %d contexts not disabled\n", Uengine);
		}

		return(status);
	}

	Uaddr =  ADDR_EN_BITPOS;				
	PUT32((ueng_base + USTORE_ADDR), Uaddr);	// Set bit 31 (ECS bit) of the USTORE_ADDRESS register
											// to indicate that microstore access will be performed

	for(i = 0; i < (Ucode_ele * 2); i+=2)		// for each microstore element to be written
	{
		PUT32((ueng_base + USTORE_DATA_L), Ucode[i]);		// Write the LSB 32 bit in first
		PUT32((ueng_base + USTORE_DATA_U), Ucode[i + 1]);	// Then write the MSB 8 bits
		// NOTE: After the USTORE_DATA_U register is written to, the pointer auto-increments and
		// points to the next location in the microstore.
		// NOTE: Each microstore instruction is 40 bits long, thus the LSB 32 bits and the MSB 8 bits
	}

	// For each microstore element, read the content of the microstore and verify against the original data
	for(i = 0; i < (Ucode_ele * 2); i+=2)
	{
		int ustore_data_l, ustore_data_u;

		PUT32((ueng_base + USTORE_ADDR), (Uaddr | (i >> 1)));
		hal_delay_us(10);

		ustore_data_l = GET32(ueng_base + USTORE_DATA_L);
		ustore_data_u = GET32(ueng_base + USTORE_DATA_U);
		if((ustore_data_l != Ucode[i]) || (ustore_data_u != Ucode[i + 1]))
		{
			eprintf("***Ustore error at offset %d for Uengine %d\n", i>>1, Uengine);
			eprintf("USTORE_DATA_L: actual = 0x%08x, expected = 0x%08x\n", ustore_data_l, Ucode[i]);
			eprintf("USTORE_DATA_U: actual = 0x%08x, expected = 0x%08x\n", ustore_data_u, Ucode[i + 1]);
			return(UENG_USTORE_ERROR);
		}
	}

	PUT32((ueng_base + USTORE_ADDR), 0);	// unset bit 31 of ustore address reg after completing writing to ustore

	for (i = 0; i < 8; i++)							// For each of the 8 threads/contexts
	{
		PUT32((ueng_base + CSR_CTX_POINTER), i);			// Selects which ctx is going to be indirecly referenced
		PUT32((ueng_base + CTX_STS_IND), 0);				// Set the ctx PC to 0
		PUT32((ueng_base + CTX_SIG_EVNT_IND), 0);			// Set the ctx signal event -- in CTX_SIG_EVENTS_INDIRECT
		PUT32((ueng_base + CTX_WAKEUP_EVNT_IND), 1);	// Set the ctx wakeup event to voluntary
	}

	PUT32((ueng_base + CC_ENABLE), CUR_COND_CODE_EN);		// Enable condition code in CC_ENABLE

	// Enable ctx 0 to 7 in CTX_ENABLES, clear CTL_STR_PAR_ER bit and BREAKPOINT bit
	PUT32((ueng_base + CTX_ENABLES), (CTX_ENABLE_MASK | (1 << 29) | (1 << 27)));

	return(UENG_SUCCESS);
}


void Clear_ME(unsigned short Uengine)
{
	Reset_Uengine(Uengine);
	hal_delay_us(25000);				// delay for 25 ms
	ClearReset_Uengine(Uengine);
}


void Clear_All_ME(void)
{
	Clear_ME(0);
	Clear_ME(1);
	Clear_ME(2);
	Clear_ME(3);
	Clear_ME(16);
	Clear_ME(17);
	Clear_ME(18);
	Clear_ME(19);
}
