/* gpio.c
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 *
 *  system: IXP2400
 *  subsystem: DIAG
 *  author: ccm, Jan 21, 02
 *  revisions:
 * 
 * 
 * --------------------------------------------------------------------------
 */

#include "misc_func.h"
#include "diagstruct.h"
#include "gpio.h"
#include "i2c.h"
#include "diag_utils.h"
#include "i2c_test.h"
#include "led.h"
#include "syslog.h"

// static function prototypes
static void GPIO_Help(void);
static void GPIO_Detection_Test(void);
static void GPIO_Level_Detect_Test(void);
static void GPIO_Edge_Detect_Test(void);
static void GPIO_Reg_Test(void);


//
// Function: GPIO_Help
//
// Description: Displays all the options available for the GPIO test
//				on the console.
//
static void GPIO_Help(void)
{
	eprintf("> t GPIO <option>, where,\n");
	eprintf("<option> are\n");
	eprintf("d - GPIO detection test\n");
	eprintf("r - GPIO register test\n");
	eprintf("e - GPIO edge detection test\n");
	eprintf("l - GPIO level detection test\n");	
	eprintf("a - all tests\n");
	eprintf("h - help\n");
}

//
// Function: GPIO_Test
//
// Description: Calls the appropriate function(s) of the GPIO test
//				depending on user input.
//
void GPIO_Test(void)
{
	char opt;
	register PDiagCommon acL = (PDiagCommon) ACADDRESS;

	acL->Err = OK;
	opt = acL->argv[2][0];
	
	switch(opt)
	{
		case 'd': 	GPIO_Detection_Test();
					break;

		case 'r': 	GPIO_Reg_Test();
					break;

		case 'e': 	GPIO_Edge_Detect_Test();
					break;

		case 'l': 	GPIO_Level_Detect_Test();
					break;

		case 'a':	GPIO_Detection_Test();
					GPIO_Reg_Test();
					GPIO_Edge_Detect_Test();
					GPIO_Level_Detect_Test();
					break;

		case 'h':
		default: 	GPIO_Help();
	}
	
	if(acL->Err == ERROR)
		eprintf("GPIO TEST FAIL!!!\n");
}

//
// Function: GPIO_Detection_Test
//
// Description: Reads and displays the current pin direction, level,
//				edge detection, level sensitivity, and interrupt of the
//				GPIO pins
//
static void GPIO_Detection_Test(void)
{
	int temp;

	temp = GET32(GPIO_PDPR);
	eprintf("GPIO Registers:\n");
	eprintf("  Pin Direction: 0x%02x\n", temp);

	temp = GET32(GPIO_PLR);
	eprintf("  Pin Level: 0x%02x\n", temp);
	// If the above does not work then use similar alt method as PDPR
	// but with '0' as low and '1' as high

	temp = GET32(GPIO_REDR);
	eprintf("  Rising Edge Detect Enable: 0x%02x\n", temp);
	// If the above does not work then use similar alt method as PDPR
	// but with '0' as disable and '1' as enable

	temp = GET32(GPIO_FEDR);
	eprintf("  Falling Edge Detect Enable: 0x%02x\n", temp);
	// If the above does not work then use similar alt method as PDPR
	// but with '0' as disable and '1' as enable

	temp = GET32(GPIO_LSHR);
	eprintf("  Level Sensitive High Detect Enable: 0x%02x\n", temp);
	// If the above does not work then use similar alt method as PDPR
	// but with '0' as disable and '1' as enable

	temp = GET32(GPIO_LSLR);
	eprintf("  Level Sensitive Low Detect Enable: 0x%02x\n", temp);
	// If the above does not work then use similar alt method as PDPR
	// but with '0' as disable and '1' as enable

	temp = GET32(GPIO_INER);
	eprintf("  Interrupt Enable: 0x%02x\n", temp);
	// If the above does not work then use similar alt method as PDPR
	// but with '0' as disable and '1' as enable

	Set_LED("Gd D");
	slowport_code[0] = SP_GPIO;
	slowport_code[1] = SP_DONE;
	dump_slowport(slowport_code, 2, SP_NON_FATAL);
	
}

//
// Function: GPIO_Reg_Test
//
// Description: Displayes and stores the existing GPIO register
//				values. Then, check the capability to pull all
//				the register values to a HIGH by setting and
//				reading back the register values. Finally, 
//				restore the registers back to their original values.
//
static void GPIO_Reg_Test(void)
{
	int PDPR, POPR, PLR, temp, status = OK;
	register PDiagCommon acL = (PDiagCommon) ACADDRESS;

	eprintf("GPIO Register Test started... \n");

	// Save existing register values of PDPR and POPR
	PDPR = GET32(GPIO_PDPR);
	POPR = GET32(GPIO_POPR);
	PLR = GET32(GPIO_PLR);

	// Display the current values of PDPR, POPR, and PLR
	eprintf("PDPR: 0x%02x\n", PDPR);
	eprintf("POPR: 0x%02x\n", POPR);
	eprintf("PLR: 0x%02x\n", PLR);

	// Set pins to output by setting bits[7:0] in GPIO_PDSR
	PUT32(GPIO_PDSR, 0xFF);
	// Read GPIO_PDPR to ensure pin directions are correct
	// are configured to output
	temp = GET32(GPIO_PDPR);
	if((temp & 0xFF) != 0xFF)
	{
		eprintf("Setting of GPIO_PDSR FAIL! GPIO_PDPR = %02x\n", temp);
		status = ERROR;
	}

	// Clear pins to input by setting bits[7:0] in GPIO_PDCR
	PUT32(GPIO_PDCR, 0xFF);
	// Read GPIO_PDPR and GPIO_PLR to ensure pins
	// are configured to input
	temp = GET32(GPIO_PDPR);
	if((temp & 0xFF) != 0x0)
	{
		eprintf("Clearing of GPIO_PDCR FAIL! GPIO_PDPR = %02x\n", temp);
		status = ERROR;
	}

	// Set pins to output by setting bits[7:0] in GPIO_PDPR
	PUT32(GPIO_PDPR, 0xFF);
	// Read GPIO_PDPR and GPIO_PLR to ensure pins
	// are configured to output
	temp = GET32(GPIO_PDPR);
	if((temp & 0xFF) != 0xFF)
	{
		eprintf("Setting of GPIO_PDPR FAIL! GPIO_PDPR = %02x\n", temp);
		status = ERROR;
	}

	// Set pins to level high by setting bits[7:0] in GPIO_POSR
	PUT32(GPIO_POSR, 0xFF);
	// Read GPIO_POPR to ensure pins are configured to level high
	temp = GET32(GPIO_POPR);
	if((temp & 0xFF) != 0xFF)
	{
		eprintf("Setting of GPIO_POSR FAIL!\n. GPIO_POPR = %02x\n", temp);
		status = ERROR;
	}

	// Clear pins to level low by setting bits[7:0] in GPIO_POCR
	PUT32(GPIO_POCR, 0xFF);
	// Read GPIO_POPR to ensure pins are configured to level low
	temp = GET32(GPIO_POPR);
	if((temp & 0xFF) != 0x0)
	{
		eprintf("Clearing of GPIO_POCR FAIL!\n. GPIO_POPR = %02x\n", temp);
		status = ERROR;
	}

	// Set pins to level high by setting bits[7:0] in GPIO_POPR
	PUT32(GPIO_POPR, 0xFF);
	temp = GET32(GPIO_POPR);
	if((temp & 0xFF) != 0xFF)
	{
		eprintf("Setting of GPIO_POPR FAIL!\n. GPIO_POPR = %02x\n", temp);
		status = ERROR;
	}

	// Restore POPR and PDPR registers
	PUT32(GPIO_PDPR, PDPR);
	PUT32(GPIO_POPR, POPR);

	
	eprintf("GPIO Register Test... ");
	if(status == OK)
	{
		eprintf("PASS\n");
		Set_LED("Gr P");
		slowport_code[0] = SP_GPIO;
		slowport_code[1] = SP_PASS;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}
	else
	{
		eprintf("FAIL!\n");
		Set_LED("Gr F");
		slowport_code[0] = SP_GPIO;
		slowport_code[1] = SP_FAIL;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);

		// prepare structure for dumping
		syslog.type = TYPE_ERROR;
		syslog.source_comp = MASTER_DIAG;	// Determine master/slave
		strcpy(syslog.desc, "GPIO Register Test FAIL!");

		// write to syslog
		syslog_dump(&syslog, sizeof(SYSLOG_DATA));		

		acL->Err = ERROR;
	}
}

//
// Function: GPIO_Edge_Detect_Test
//
// Description: Set GPIO for rising edge detection and calls
//				i2c_read to trigger a rising edge. Checks for
//				rising edge detection.
//				Set GPIO for falling edge detection and calls
//				i2c_read to trigger a falling edge. Checks for
//				falling edge detection.
//
static void GPIO_Edge_Detect_Test(void)
{
	int temp;
	unsigned char buffer[2];
	register PDiagCommon acL = (PDiagCommon) ACADDRESS;

	// Set Pin 6(EDGE_LEVEL_GPIO_TO_TEST) to Detect Rising Edge
	PUT32(GPIO_REDR, EDGE_LEVEL_GPIO_TO_TEST);
	// Clear Pin 6(EDGE_LEVEL_GPIO_TO_TEST) of previous edge detect status
	PUT32(GPIO_EDSR, EDGE_LEVEL_GPIO_TO_TEST);

	// Make a I2C read operation to trigger rising edge
	i2c_read(SLAVE_ADDR_0, 0x1, buffer);

	temp = GET32(GPIO_EDSR);
	eprintf("GPIO Rising Edge Detect... ");
	if( (temp & EDGE_LEVEL_GPIO_TO_TEST) != 0)
	{
		eprintf("PASS\n");
		Set_LED("GerP");
		slowport_code[0] = SP_GPIO;
		slowport_code[1] = SP_PASS;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}
	else
	{
		eprintf("FAIL!\n");
		Set_LED("GerF");
		slowport_code[0] = SP_GPIO;
		slowport_code[1] = SP_FAIL;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);

		// prepare structure for dumping
		syslog.type = TYPE_ERROR;
		syslog.source_comp = MASTER_DIAG;	// Determine master/slave
		strcpy(syslog.desc, "GPIO Edge Detect Rising Test FAIL!");

		// write to syslog
		syslog_dump(&syslog, sizeof(SYSLOG_DATA));

		acL->Err = ERROR;
	}

	// Clear all pins to disable Detect Rising Edge
	PUT32(GPIO_REDR, 0x0);
	// Set Pin 6(EDGE_LEVEL_GPIO_TO_TEST) to Detect Falling Edge
	PUT32(GPIO_FEDR, EDGE_LEVEL_GPIO_TO_TEST);
	// Clear Pin 6(EDGE_LEVEL_GPIO_TO_TEST) of previous edge detect status
	PUT32(GPIO_EDSR, EDGE_LEVEL_GPIO_TO_TEST);

	// Make a I2C read operation to trigger falling edge
	i2c_read(SLAVE_ADDR_0, 0x1, buffer);

	temp = GET32(GPIO_EDSR);
	eprintf("Falling Edge Detect... ");
	if( (temp & EDGE_LEVEL_GPIO_TO_TEST) != 0)
	{
		eprintf("PASS\n");
		Set_LED("GefP");
		slowport_code[0] = SP_GPIO;
		slowport_code[1] = SP_PASS;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}
	else
	{
		eprintf("FAIL!\n");
		Set_LED("GefF");
		slowport_code[0] = SP_GPIO;
		slowport_code[1] = SP_FAIL;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);

		// prepare structure for dumping
		syslog.type = TYPE_ERROR;
		syslog.source_comp = MASTER_DIAG;	// Determine master/slave
		strcpy(syslog.desc, "GPIO Edge Detect Falling Test FAIL!");

		// write to syslog
		syslog_dump(&syslog, sizeof(SYSLOG_DATA));

		acL->Err = ERROR;
	}
}

//
// Function: GPIO_Level_Detect_Test
//
// Description: Sets GPIO for level HIGH sensitivity and calls
//				i2c_read to trigger a level HIGH. Checks for
//				level HIGH detection.
//				Sets GPIO for level LOW sensitivity and calls
//				i2c_read to trigger a level LOW. Checks for
//				level LOW detection.
//
static void GPIO_Level_Detect_Test(void)
{
	int temp;
	unsigned char buffer[2];
	register PDiagCommon acL = (PDiagCommon) ACADDRESS;

	// Set Pin 6(EDGE_LEVEL_GPIO_TO_TEST) to Enable Level Sensitive High
	PUT32(GPIO_LSHR, EDGE_LEVEL_GPIO_TO_TEST);
	// Clear Pin 6(EDGE_LEVEL_GPIO_TO_TEST) of previous Level Sensitive status
	PUT32(GPIO_LDSR, EDGE_LEVEL_GPIO_TO_TEST);

	// Make a I2C read operation to trigger a level change
	i2c_read(SLAVE_ADDR_0, 0x1, buffer);

	// Read GPIO_LDSR to determine if a Level Sensitive High was detected
	temp = GET32(GPIO_LDSR);
	eprintf("GPIO Level Sensitive High... ");
	if( (temp & EDGE_LEVEL_GPIO_TO_TEST) != 0)
	{
		eprintf("PASS\n");
		Set_LED("GlhP");
		slowport_code[0] = SP_GPIO;
		slowport_code[1] = SP_PASS;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}
	else
	{
		eprintf("FAIL!\n");
		Set_LED("GlhF");
		slowport_code[0] = SP_GPIO;
		slowport_code[1] = SP_FAIL;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);

		// prepare structure for dumping
		syslog.type = TYPE_ERROR;
		syslog.source_comp = MASTER_DIAG;	// Determine master/slave
		strcpy(syslog.desc, "GPIO Level Detect High Test FAIL!");

		// write to syslog
		syslog_dump(&syslog, sizeof(SYSLOG_DATA));

		acL->Err = ERROR;
	}

	// Clear Pin 6(EDGE_LEVEL_GPIO_TO_TEST) to Disable Level Sensitive High
	PUT32(GPIO_LSHR, 0);
	// Clear Pin 6(EDGE_LEVEL_GPIO_TO_TEST) of previous Level Sensitive status
	PUT32(GPIO_LDSR, EDGE_LEVEL_GPIO_TO_TEST);
	// Set Pin 6(EDGE_LEVEL_GPIO_TO_TEST) to Enable Level Sensitive Low
	PUT32(GPIO_LSLR, EDGE_LEVEL_GPIO_TO_TEST);

	// Make a I2C read operation to trigger a level change
	i2c_read(SLAVE_ADDR_0, 0x1, buffer);

	// Read GPIO_LDSR to determine if a Level Sensitive Low was detected
	temp = GET32(GPIO_LDSR);
	eprintf("Level Sensitive Low... ");
	if( (temp & EDGE_LEVEL_GPIO_TO_TEST) != 0)
	{
		eprintf("PASS\n");
		Set_LED("GllP");
		slowport_code[0] = SP_GPIO;
		slowport_code[1] = SP_PASS;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}
	else
	{
		eprintf("FAIL!\n");
		Set_LED("GllF");
		slowport_code[0] = SP_GPIO;
		slowport_code[1] = SP_FAIL;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);

		// prepare structure for dumping
		syslog.type = TYPE_ERROR;
		syslog.source_comp = MASTER_DIAG;	// Determine master/slave
		strcpy(syslog.desc, "GPIO Level Detect Low Test FAIL!");

		// write to syslog
		syslog_dump(&syslog, sizeof(SYSLOG_DATA));

		acL->Err = ERROR;
	}
}
