/* scratch.c
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 *
 *  system: IXDP2400
 *  subsystem: DIAG
 *  author: dalsraja, Feb 8, 2002
 *  revisions:	dalsraja, May 7, 2002
 *				
 * 
 * 
 * --------------------------------------------------------------------------
 */

#include "scratch.h"
#include "common.h"
#include "diagstruct.h"
#include "syslog.h"
#include "memory.h"
#include "diag_utils.h"
#include "hal_ixdp2400.h"
#include "led.h"

static void Scratchpad_Help(void);
static void Config_Scratch(int ringnum, int *Base_Addr, int *Head_Addr, int *Tail_Addr);
static int scratch_pattern_wr_rd(UINT32 pattern);

extern UINT32 t_all_verbose;

// Function: static void Scratchpad_Help(void)
//
// Provides the help which is displayed on the console screen
// for the Scratchpad tests
static void Scratchpad_Help(void)
{
	eprintf("> t Xscale s <option>,\n");
	eprintf("where <option> are\n");
	eprintf("\tp - Scratchpad pattern write-read test\n");
	eprintf("\tr - Scratch ring test\n");
	eprintf("\ta - All scratch tests\n");
	eprintf("\th - help message\n");
}


// Function: void Xscale_Scratchpad_Test(void)
//
// Determine which Scratchpad test is called - pattern test or scratch ring test
//		Pattern - writes and reads a pattern to all the Scratchpad and compares it
//		Ring - Configures the 16 scratch rings and checks its functionality
void Xscale_Scratchpad_Test(void)
{
	char opt[16];
	register PDiagCommon acL = (PDiagCommon) ACADDRESS;

	strcpy(opt, acL->argv[3]);//opt = acL->argv[3][0];
	
	switch(opt[0])
	{
		case 'p': 	Scratch_Pattern_Test();
					break;

		case 'r': 	ScratchRing_Test();
					break;

		case 'a': 	Scratch_Pattern_Test();
					ScratchRing_Test();
					break;

		case 'h':
		default:	Scratchpad_Help();
	}

	return;	
}


// Ring_#_Base register
//
//  31:30         29:14            13:9       8:0
// +------+---------------------+--------+-----------+
// | Size |       Unused        |  Base  |   Unused  |
// +------+---------------------+--------+-----------+
//
// Size - Ring size, i.e. 00 - 128, 01 - 256, 10 - 512, 11 - 1024
// Base - Ring base address - refer to document for more info


//********************************************************************************************
// Function: void ScratchRing_Test(void)
//
// Configures a scratch ring, writes data to the ring, then reads the data and compares
// whether the data written is the same as the data read from the ring
// Indirectly tests the writing and reading of data to the whole 16KB or scratch
//********************************************************************************************
void ScratchRing_Test(void)
{
	int read_data, ctr, flag, temp_value, read_data_direct_old, read_data_direct_new;
	int ring_num, ring_size, ring_base, max_ring_base, ring_data;
	int ring_base_addr, ring_head_addr, ring_tail_addr;
	int scratch_getput_inst;
	register PDiagCommon acL = (PDiagCommon) ACADDRESS;
	int scratch_ringsize[] = {128, 256, 512, 1024};

	ring_base_addr = 0;
	ring_head_addr = 0;
	ring_tail_addr = 0;
	flag = OK;

	for (ring_num = 0; ring_num < 16; ring_num++)		// For each of the rings, i.e. 0 to 15
	{

		// Get the ring address
		Config_Scratch(ring_num, &ring_base_addr, &ring_head_addr, &ring_tail_addr);

		// Get the scratch ring Get/Put instruction address
		scratch_getput_inst = SCR_RING_INSTR_ADDR + (ring_num << 2);

		for (ring_size = 0; ring_size < 4; ring_size++)		// For each of the ring sizes
		{
			// Calculate the ring's maximum possible base address
			max_ring_base = SCRATCH_SIZE_LW / scratch_ringsize[ring_size];
			
			// For each of the possible ring base addresses, range varies for each ring size
			for (ring_base = 0; ring_base < max_ring_base; ring_base++)
			{
				int rd_back_csr;

				// Value to be written to RING_#_BASE register
				temp_value = (int)((ring_size << 30) | (ring_base << (ring_size + 9)));

				// Write value into RING_#_BASE register
				PUT32(ring_base_addr, temp_value);
				
				// Write 0 into RING_#_HEAD reg to init
				PUT32(ring_head_addr, 0);

				// Write 0 into RING_#_TAIL reg to init
				PUT32(ring_tail_addr, 0);

				rd_back_csr = GET32(ring_tail_addr);
				if(rd_back_csr != 0)
					eprintf("Scratch CSR is not updated\n");

				// pattern written to ring is
				// +-------+-------------+-------+------------------+
				// | 31:28 |    27:16    | 15:14 |       13:0       |
				// +-------+-------------+-------+------------------+
				//
				// [31:28] - Ring number
				// [27:18] - Fixed pattern of 0xA5A
				// [17:14] - ring size
				// [13:0]  - scratch address
				ring_data = (ring_num<<28) | (SCRATCHRING_PATTERN<<16) | (ring_size<<14) | (ring_base << (ring_size + 9));

				for (ctr = 0; ctr < scratch_ringsize[ring_size]; ctr++)
				{
					// Write pattern to scratch ring
					PUT32(scratch_getput_inst, (ring_data + (ctr << 2)));

					// Read data from scratch ring
					read_data = GET32(scratch_getput_inst);


					// Compare written and read data, if not the same
					if (read_data != (ring_data + (ctr << 2)))	
					{
						eprintf("FAIL! Scratchpad ring %d  Write value=0x%08x  Read value=0x%08x...FAIL!\n", ring_num, (ring_data + (ctr << 2)), read_data);
						flag = ERROR;
					}

					// This checks if the ring completes the full circle and re-starts
					// from location zero of the ring
					if (ctr == (scratch_ringsize[ring_size] - 1))
					{
						// eprintf("Checking if Scratch ring %d wraps around...\n", ring_num);
						// Read the data of the Scratchpad directly before performing
						// a Put to the Ring. This will be used to confirm that the ring
						// wrapped around to start from the beginning again
						read_data_direct_old = GET32(SCRATCHPAD_BASE + (ring_base << (ring_size + 9)));

						// Write pattern to scratch ring
						PUT32(scratch_getput_inst, (ring_data + 1));

						// Read data from scratch ring
						read_data = GET32(scratch_getput_inst);

						// Read the data of the Scratchpad directly after performing
						// a Put and Get to the Ring. This will be used to confirm that
						// the ring wrapped around to start from the beginning again
						read_data_direct_new = GET32(SCRATCHPAD_BASE + (ring_base << (ring_size + 9)));
						
						// Compare data written to and read from the ring.
						// If not the same, it is an error.
						if (read_data != (ring_data + 1))
						{
							eprintf("FAIL! Scratch ring %d  Ring put=0x%08x  Ring get=0x%08x\n", ring_num, (ring_data + 1), read_data);
							flag = ERROR;
						}

						// If data read directly from scratch (not ring access) before 
						// and after the PUT and GET from the ring are the same, the ring
						// did not wrap around, i.e. it is an error.
						if  (read_data_direct_old == read_data_direct_new)
						{
							eprintf("FAIL! Scratch ring %d did not wrap around. Old value=0x%08x  New value=0x%08x  Expected=0x%08x\n", ring_num, read_data_direct_old, read_data_direct_new, (ring_data + 1));
							flag = ERROR;
						}


						// If data read directly from scratch (not ring access) and the
						// data written to ring are not the same, it is an error.
						if (read_data_direct_new != read_data)
						{
							eprintf("FAIL! Scratch ring %d  Direct scratch read=0x%08x  Expected=0x%08x\n", ring_num, read_data_direct_new, (ring_data + 1));
							flag = ERROR;
						}
					}
				}
			}
		}
	}
	
	if (flag == OK)
	{
		if(t_all_verbose != MINIMAL)
			eprintf("Scratch ring test... PASS\n");
		Set_LED("XSrP");
		slowport_code[0] = SP_XSC;
		slowport_code[1] = SP_PASS;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}
	else
	{
		// prepare structure for dumping
		syslog.type = TYPE_ERROR;
		syslog.source_comp = MASTER_DIAG;	// Determine master/slave
		strcpy(syslog.desc, "Scratch ring test FAIL!");

		// write to syslog
		syslog_dump(&syslog, sizeof(SYSLOG_DATA));

		acL->Err = ERROR;

		eprintf("Scratch ring test... FAIL!\n");
		Set_LED("XSrF");
		slowport_code[0] = SP_XSC;
		slowport_code[1] = SP_FAIL;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}		
}


//***************************************************************************
// Function: void Scratch_Pattern_Test(void) 
//
// Initiates the Scratchpad pattern test
//***************************************************************************
void Scratch_Pattern_Test(void)
{
	int status1, status2;
	register PDiagCommon acL = (PDiagCommon) ACADDRESS;

	// Note: The Scratchpad write/read continues through the whole Scratchpad memory
	// even if an arror was detected. It does not exit this test if an error is found
	// until it reaches the end of the Scratchpad memory block
	// Use defined pattern
	status1 = scratch_pattern_wr_rd((UINT32)SCRATCH_PATTERN);

	if (status1 == OK)
	{
		if(t_all_verbose != MINIMAL)
			eprintf("Scratchpad pattern 0x%X test... PASS\n", SCRATCH_PATTERN);
	}
	else
	{
		eprintf("Scratchpad pattern 0x%X test... FAIL!\n", SCRATCH_PATTERN);
	}


	// Use inverse of defined pattern
	status2 = scratch_pattern_wr_rd((UINT32)~SCRATCH_PATTERN);

	if (status2 == OK)
	{
		if(t_all_verbose != MINIMAL)
			eprintf("Scratchpad pattern 0x%X test... PASS\n", ~(SCRATCH_PATTERN));
	}
	else
	{
		eprintf("Scratchpad pattern 0x%X test... FAIL!\n", ~(SCRATCH_PATTERN));
	}


	if ((status1 == ERROR) || (status2 == ERROR))
	{
		// prepare structure for dumping
		syslog.type = TYPE_ERROR;
		syslog.source_comp = MASTER_DIAG;	// Determine master/slave
		strcpy(syslog.desc, "Scratch pattern test FAIL!");

		// write to syslog
		syslog_dump(&syslog, sizeof(SYSLOG_DATA));

		acL->Err = ERROR;

		Set_LED("XSpF");
		slowport_code[0] = SP_XSC;
		slowport_code[1] = SP_FAIL;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}
	else
	{
		Set_LED("XSpP");
		slowport_code[0] = SP_XSC;
		slowport_code[1] = SP_PASS;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}
}


//***************************************************************************
// Function: static void Config_scratch(int ringnum, 
//							int *Base_Addr, int *Head_Addr, int *Tail_Addr)
//
// Gets the address for the ring base, head and tail registers
//***************************************************************************
static void Config_Scratch(int ringnum, int *Base_Addr, int *Head_Addr, int *Tail_Addr)
{
	*Base_Addr = RING_BASE(ringnum);
	*Head_Addr = RING_HEAD(ringnum);
	*Tail_Addr = RING_TAIL(ringnum);
}


//***************************************************************************
// Function: static int scratch_pattern_wr_rd(UINT32 pattern) 
//
// Writes a specified pattern to the each location on the Scratchpad and 
// reads the value back and compares it with the written value
//***************************************************************************
static int scratch_pattern_wr_rd(UINT32 pattern)
{
	int flag, ctr, read_value;

	flag = OK;

	if(t_all_verbose != MINIMAL)
		eprintf("Writing pattern %x to Scratchpad...\n", pattern);
	
	for (ctr=0; ctr<SCRATCH_SIZE; ctr+=4)			// Move through the entire Scratchpad
	{
		PUT32((SCRATCHPAD_BASE+ctr), pattern);			// Write the pattern to specified location
		read_value = GET32(SCRATCHPAD_BASE+ctr);		// Read the value from specified location

		// Compare written and read data. 
		// If not the same, display a message and set error flag
		if (read_value != pattern)		
		{
			eprintf("Error at Address:0x%x     Read Value: 0x%08x\n", (int)(SCRATCHPAD_BASE+ctr), read_value);
			flag = ERROR;
		}
	}

	return flag;
}
