/* uart.c
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 *
 *  system: IXP2400
 *  subsystem: DIAG
 *  author: ccm, Jan 21, 02
 *  revisions:
 * 
 * 
 * --------------------------------------------------------------------------
 */

#include "misc_func.h"
#include "diagstruct.h"
#include "diag_utils.h"
#include "uart.h"
#include "console_monitor.h"
#include "led.h"
#include "syslog.h"
#include "hal_ixdp2400.h"
#include "hal_platform_ints.h"

// static function prototypes
static void UART_Help(void);
static void UART_Dump_Reg(void);
static void UART_Reg_Test(void);
static void UART_Polling_Test(void);
static void UART_Int_Test(void);
static void UART_FIFO_Polling_Test(void);
static void UART_FIFO_Int_Test(void);
static void _LED_Display_Cause_Of_Failure(void);
static int _Test_Register(unsigned int reg_addr, unsigned int var);

char read_buffer[128];
char write_buffer[128];
volatile int write_counter = 0, read_counter = 0;
volatile int int_flag = 0;

extern int serial_osc;

//
// Function: UART_Help
//
// Description: Displays all the options available for the UART test
//				on the console.
//
static void UART_Help(void)
{
	eprintf("> t uart <option>,\n");
	eprintf("where <option> are\n");
	eprintf("d - Dump all UART readable registers\n");
	eprintf("r - UART Register test\n");
	eprintf("q - UART Non-FIFO polling test\n");
	eprintf("j - UART Non-FIFO interrupt test\n");
	eprintf("p - UART FIFO polling test\n");
	eprintf("i - UART FIFO interrupt test\n");
	eprintf("a - All UART test\n");
	eprintf("h - help\n");
}

//
// Function: UART_Test
//
// Description: Calls the appropriate function(s) of the UART test
//				depending on user input.
//
void UART_Test(void)
{
	char opt;
	register PDiagCommon acL = (PDiagCommon) ACADDRESS;
	int loops, dcache_status;

	acL->Err = OK;	
	opt = acL->argv[2][0];
	if(acL->argc < 4)
		loops = 1;
	else
		loops = acL->decArg[3];

	dcache_status = get_CP15() & DCACHE_BIT;
	if(dcache_status != 0)
		dcache_off();
	
	switch(opt)
	{
		case 'd': 	UART_Dump_Reg();
					break;

		case 'r': 	while(loops--)
						UART_Reg_Test();
					break;

		case 'q': 	while(loops--)
						UART_Polling_Test();
					break;

		case 'p':	while(loops--)
						UART_FIFO_Polling_Test();
					break;
		
		case 'j': 	while(loops--)
						UART_Int_Test();
					break;

		case 'i':	while(loops--)
						UART_FIFO_Int_Test();
					break;

		case 'a':	while(loops--)
					{
						UART_Reg_Test();
						UART_Polling_Test();
						UART_FIFO_Polling_Test();
						UART_Int_Test();
						UART_FIFO_Int_Test();
					}
					break;

		case 'h':
		default: 	UART_Help();
	}
	if(dcache_status != 0)
		dcache_on();

	if(acL->Err == ERROR)
		eprintf("UART TEST FAIL!!!\n");
}


//
// Function: UART_Dump_Reg
//
// Description: Displays the register values of:
//				DLRL, DLRH, IER, IIR, LCR, LSR, and SPR
//
static void UART_Dump_Reg(void)
{
	UINT8 temp, dlrl, dlrh;

	PUT8(UART_LCR, (GET8(UART_LCR) | (1<<7)));	// Bit 7 = 1; DLAB = 1

	// Read DLRL and DLRH
	dlrl = GET8(UART_DLRL);
	dlrh = GET8(UART_DLRH);

	PUT8(UART_LCR, (GET8(UART_LCR) & 0x7F));	// Bit 7 = 0; DLAB = 0

	// Requires DLAB = 0 to access THR for serial output

	// Display DLRL and DLRH
	eprintf("DLRL - 0x%02x\n", dlrl);
	eprintf("DLRH - 0x%02x\n", dlrh);

	// Display IER
	temp = GET8(UART_IER);
	eprintf("IER  - 0x%02x\n", temp);

	// Display IIR
	temp = GET8(UART_IIR);
	eprintf("IIR  - 0x%02x\n", temp);

	// Display LCR
	temp = GET8(UART_LCR);
	eprintf("LCR  - 0x%02x\n", temp);

	// Display LSR
	temp = GET8(UART_LSR);
	eprintf("LSR  - 0x%02x\n", temp);

	// Display SPR
	temp = GET8(UART_SPR);
	eprintf("SPR  - 0x%02x\n", temp);
}

//
// Function: UART_Reg_Test
//
// Description: Checks for expected settings in the UART:
//
static void UART_Reg_Test(void)
{
	unsigned int lcr, iir, ier, dlrl, dlrh, status = OK;

	lcr = GET8(UART_LCR);
	if((lcr & 0x3) != 0x3)
	{
		eprintf("LCR Reg Read FAIL! Read 0x%x\n", lcr);
		status = ERROR;
		_LED_Display_Cause_Of_Failure();
	}

	iir = GET8(UART_IIR);
	if((iir & 0xc0) != 0xc0)	// check bits[7:6] are set for "FIFO mode is selected"
	{
		eprintf("IIR Reg Read FAIL! Read 0x%x\n", iir);
		status = ERROR;
		_LED_Display_Cause_Of_Failure();
	}

	ier = GET8(UART_IER);
	if((ier & 0x40) != 0x40)	//check if UART Unit is enabled
	{
		eprintf("IER Reg Read FAIL! Read 0x%x\n", ier);
		status = ERROR;
		_LED_Display_Cause_Of_Failure();
	}

	PUT8(UART_LCR, (GET8(UART_LCR) | (1<<7)));	// Bit 7 = 1; DLAB = 1

	dlrl = GET8(UART_DLRL);
	dlrh = GET8(UART_DLRH);

	if((dlrl & 0xFF) != (((serial_osc/DEFAULT_BAUD)/16)&0xFF))
	{
		eprintf("DLRL Reg Read FAIL! Read 0x%x\n", dlrl);
		status = ERROR;
		_LED_Display_Cause_Of_Failure();
	}

	if((dlrh & 0xFF) != (((serial_osc/DEFAULT_BAUD)/16) >>8))
	{
		eprintf("DLRH Reg Read FAIL! Read 0x%x\n", dlrh);
		status = ERROR;
		_LED_Display_Cause_Of_Failure();
	}

	PUT8(UART_LCR, (GET8(UART_LCR) & 0x7F));		// Bit 7 = 0; DLAB = 0
	PUT8(UART_IER, (GET8(UART_IER) & ~ENABLE_UART));// Disable UART
	PUT8(UART_LCR, (GET8(UART_LCR) | (1<<7)));		// Bit 7 = 1; DLAB = 1

	// *** Test UART_DLRL Register ***
	if(_Test_Register(UART_DLRL, dlrl) == ERROR)
	{
		PUT8(UART_LCR, (GET8(UART_LCR) & 0x7F));	// Bit 7 = 0; DLAB = 0
		PUT8(UART_IER, (GET8(UART_IER) | ENABLE_UART));	// Enable UART
		hal_delay_us(10000);
		eprintf("UART_DLRL WRV FAIL!\n");
		status = ERROR;
	}

	// *** Test UART_DLRH Register ***
	if(_Test_Register(UART_DLRH, dlrh) == ERROR)
	{
		PUT8(UART_LCR, (GET8(UART_LCR) & 0x7F));	// Bit 7 = 0; DLAB = 0
		PUT8(UART_IER, (GET8(UART_IER) | ENABLE_UART));	// Enable UART
		hal_delay_us(10000);
		eprintf("UART_DLRH WRV FAIL!\n");
		status = ERROR;
	}
	
	// *** Test UART_IER Register ***
	PUT8(UART_LCR, (GET8(UART_LCR) & 0x7F));	// Bit 7 = 0; DLAB = 0
	if(_Test_Register(UART_IER, ier) == ERROR)
	{
		PUT8(UART_IER, (GET8(UART_IER) | ENABLE_UART));	// Enable UART
		hal_delay_us(10000);
		eprintf("UART_IER WRV FAIL!\n");
		status = ERROR;
	}

	// *** Test UART_LCR Register ***
	if(_Test_Register(UART_LCR, lcr) == ERROR)
	{
		PUT8(UART_IER, (GET8(UART_IER) | ENABLE_UART));	// Enable UART
		hal_delay_us(10000);
		eprintf("UART_LCR WRV FAIL!\n");
		status = ERROR;
	}

	PUT8(UART_FCR, 0x7);
	hal_delay_us(1000);
	PUT8(UART_IER, (GET8(UART_IER) | ENABLE_UART));	// Enable UART
	hal_delay_us(10000);

	eprintf("UART Reg Test... ");

	if(status == OK)
	{
		eprintf("PASS\n");
		Set_LED("Ur P");
		slowport_code[0] = SP_UART;
		slowport_code[1] = SP_PASS;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}
	else	// FAIL
	{
		eprintf("FAIL!\n");
		Set_LED("Ur F");
		slowport_code[0] = SP_UART;
		slowport_code[1] = SP_FAIL;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
		_LED_Display_Cause_Of_Failure();

		// prepare structure for dumping
		syslog.type = TYPE_ERROR;
		syslog.source_comp = MASTER_DIAG;	// Determine master/slave
		strcpy(syslog.desc, "UART Register Test FAIL!");

		// write to syslog
		syslog_dump(&syslog, sizeof(SYSLOG_DATA));
	}
}

//
// Function: UART_Polling_Test
//
// Description: Tests UART in Non-FIFO Polling Mode
//
static void UART_Polling_Test(void)
{
	char string_test[16], stored_ier;

	stored_ier = GET8(UART_IER);	// store original IER value
	PUT8(UART_FCR, 0);	// Disable Transmitter and Receiver FIFOs

	// Receiver Data Available Interrupt Disable
	// Transmit Data Request Interrupt Disable
	// Receiver Line Status Interrupt Disable
	// Receiver Time Out Interrupt Disable
	PUT8(UART_IER, ENABLE_UART);

	eprintf("\n\rPlease enter \"test*uart\" as an input: ");
	
	getCmdLine(string_test);
	eprintf("UART Non-FIFO Polling Test... ");
	if(strcmp(string_test, "test*uart") == 0)
	{
		eprintf("PASS\n");
		Set_LED("Uq P");
		slowport_code[0] = SP_UART;
		slowport_code[1] = SP_PASS;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}
	else
	{
		eprintf("FAIL!\n");
		Set_LED("Uq F");
		slowport_code[0] = SP_UART;
		slowport_code[1] = SP_FAIL;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
		_LED_Display_Cause_Of_Failure();

		// prepare structure for dumping
		syslog.type = TYPE_ERROR;
		syslog.source_comp = MASTER_DIAG;	// Determine master/slave
		strcpy(syslog.desc, "UART Polling Test FAIL!");

		// write to syslog
		syslog_dump(&syslog, sizeof(SYSLOG_DATA));
	}

	// Restore FCR and IER values
	PUT8(UART_FCR, 0x1);	// Re-enable FIFO mode
	PUT8(UART_IER, stored_ier);
}


//
// Function: UART_FIFO_Polling_Test
//
// Description: Tests UART in FIFO Polling Mode
//
static void UART_FIFO_Polling_Test(void)
{
	int stored_ier;
	char string_test[16];

	stored_ier = GET8(UART_IER);	// store original IER value
	// Enable Transmit and Receive FIFOs (bit 0)
	// Reset Receiver and Transmitter FIFOs (bits [2:1])
	PUT8(UART_FCR, 0x7);

	// Receiver Data Available Interrupt Disable
	// Transmit Data Request Interrupt Disable
	// Receiver Line Status Interrupt Disable
	// Receiver Time Out Interrupt Disable
	PUT8(UART_IER, ENABLE_UART);

	eprintf("\nPlease enter \"test*uart\" as an input: ");

	getCmdLine(string_test);
	eprintf("UART FIFO Polling Test... ");
	if(strcmp(string_test, "test*uart") == 0)
	{
		eprintf("PASS\n");
		Set_LED("Up P");
		slowport_code[0] = SP_UART;
		slowport_code[1] = SP_PASS;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
	}
	else
	{
		eprintf("FAIL!\n");
		Set_LED("Up F");
		slowport_code[0] = SP_UART;
		slowport_code[1] = SP_FAIL;
		dump_slowport(slowport_code, 2, SP_NON_FATAL);
		_LED_Display_Cause_Of_Failure();

		// prepare structure for dumping
		syslog.type = TYPE_ERROR;
		syslog.source_comp = MASTER_DIAG;	// Determine master/slave
		strcpy(syslog.desc, "UART FIFO Polling Test FAIL!");

		// write to syslog
		syslog_dump(&syslog, sizeof(SYSLOG_DATA));
	}

	// Restore FCR and IER
	PUT8(UART_FCR, 0x1);	// Re-enable FIFO mode
	PUT8(UART_IER, stored_ier);
}

//
// Function: UART_Int_Test
//
// Description: Test UART in Non-FIFO Interrupt Mode
//
static void UART_Int_Test(void)
{
	int counter, stored_ier;
	int_flag = 0;
	read_counter = 0;
	write_counter = 0;

	bzero(read_buffer, 128);
	strcpy(write_buffer, "\n\rPlease enter \"test*uart\" as an input: ");

	stored_ier = GET8(UART_IER);	// store original IER value

	// Receiver Data Available Interrupt Enable
	// Transmit Data Request Interrupt Enable
	// Receiver Line Status Interrupt Enable
	// Receiver Time Out Interrupt Disable
	PUT8(UART_FCR, 0x0);	// Disable Transmitter and Receiver FIFOs
	PUT8(UART_IER, 0x7 | ENABLE_UART);
	
	// Verify Transmitter interrupt
	counter = 10;
	while((int_flag != WRITE_INT) && counter--);

	if(int_flag == WRITE_INT)
	{
		int_flag = 0;
		while(!KEYPRESS && !int_flag);

		if(int_flag == READ_INT)
		{
			if(strcmp(read_buffer, "test*uart") == 0)
			{
				// Restore FCR and IER values
				PUT8(UART_IER, stored_ier);
				PUT8(UART_FCR, 0x7);	// Re-enable FIFO mode
				eprintf("\nUART Non-FIFO Interrupt Test... PASS\n");
				Set_LED("Uj P");
				slowport_code[0] = SP_UART;
				slowport_code[1] = SP_PASS;
				dump_slowport(slowport_code, 2, SP_NON_FATAL);

				return;
			}
		}
	}

	// Restore FCR and IER values
	PUT8(UART_IER, stored_ier);
	PUT8(UART_FCR, 0x7);	// Re-enable FIFO mode
	
	eprintf("\nUART Non-FIFO Interrupt Test... FAIL!\n");
	Set_LED("Uj F");
	slowport_code[0] = SP_UART;
	slowport_code[1] = SP_FAIL;
	dump_slowport(slowport_code, 2, SP_NON_FATAL);
	_LED_Display_Cause_Of_Failure();

	// prepare structure for dumping
	syslog.type = TYPE_ERROR;
	syslog.source_comp = MASTER_DIAG;	// Determine master/slave
	strcpy(syslog.desc, "UART Interrupt Test FAIL!");

	// write to syslog
	syslog_dump(&syslog, sizeof(SYSLOG_DATA));

//	PUT32(IRQ_ENABLE_CLR_REG, UART_INT);	// Disable UART Interrupt
	return;
}

//
// Function: UART_FIFO_Int_Test
//
// Description: Tests UART in FIFO Interrupt Mode
//
static void UART_FIFO_Int_Test(void)
{
	int stored_ier, counter;
	int_flag = 0;
	read_counter = 0;
	write_counter = 0;

	bzero(read_buffer, 128);
	strcpy(write_buffer, "\n\rPlease enter \"test*uart\" as an input: ");

	stored_ier = GET8(UART_IER);	// store original IER value
	// Enable Transmit and Receive FIFOs (bit 0)
	// Reset Receiver and Transmitter FIFOs (bits [2:1])
	// 00 = 1 byte in FIFO causes interrupt (bits [7:6])
	PUT8(UART_FCR, 0x7);
	hal_delay_us(1000);


//	PUT32(IRQ_ENABLE_SET_REG, UART_INT);	// Enable UART Interrupt

	// Receiver Data Available interrupt enabled (bit 0)
	// Transmit Data Request interrupt enabled (bit 1)
	// Receiver Line Status interrupt enabled (bit 2)
	// Receiver Time Out interrupt disabled (bit 4)
	PUT8(UART_IER, 0x7 | ENABLE_UART);
	// As soon as the Transmit Data Request is enabled it will trigger
	// an interrupt to display "write buffer"

	counter = 10;
	while((int_flag != WRITE_INT) && counter--);

	if(int_flag == WRITE_INT)
	{
		int_flag = 0;
		while(!int_flag);

		if(int_flag == READ_INT)
		{
			if(strcmp(read_buffer, "test*uart") == 0)	// match
			{
				// Restore FCR and IER
				PUT8(UART_FCR, 0x1);	// Re-enable FIFO mode
				PUT8(UART_IER, stored_ier);
				hal_delay_us(1000);

				eprintf("UART FIFO Interrupt Test... PASS\n");
				Set_LED("Ui P");
				return;
			}
		}
	}

	// Restore FCR and IER
	PUT8(UART_FCR, 0x1);	// Re-enable FIFO mode
	PUT8(UART_IER, stored_ier);
	hal_delay_us(1000);
	
	eprintf("UART FIFO Interrupt Test... FAIL!\n");
	Set_LED("Ui F");
	_LED_Display_Cause_Of_Failure();

	return;
}

//
// Function: UART_Interrupt_Handler
//
// Description: Handles Interrupts generated by serial
//
void UART_Interrupt_Handler(void)
{
	int iir;
	volatile int temp;
	unsigned char temp_read;

	Set_LED("UART");

	iir = GET8(UART_IIR);

	if((iir & IIR_FIFOES) == FIFO_MODE)
	{
		if((iir & 0xF) == 0x2)	// Transmit Interrupt
		{
			Set_LED("F TX");
			eprintf("%s", write_buffer);
			PUT8(UART_IER, (GET8(UART_IER) & ~(0x2)));
			int_flag = WRITE_INT;
		}

		else if((iir & 0xF) == 0x4)	// Receive Interrupt
		{
			Set_LED("F RX");
			do
			{
				temp_read = GET8(UART_RBR);
				PUT8(UART_THR, temp_read);
				
				// for debug: PUT8(ALPHANUM_DIS_DATA_REG + (read_counter%4), temp_read);

				if(temp_read == '\r')
				{
					PUT8(UART_IER, ENABLE_UART);	// Disable interrupt mode
					int_flag = READ_INT;
					PUT8(UART_THR, '\n');
					break;
				}
				read_buffer[read_counter++] = temp_read;
				
				temp = GET8(UART_LSR) & LSR_DR;
			} while(temp);
		}
	}
	else	// NON_FIFO_MODE
	{
		if((iir & 0xF) == 0x2)	// Transmit Interrupt
		{
			Set_LED("NFTX");
			write_counter = 0;
			do
			{
				if( (GET8(UART_LSR) & (1<<6)) != 0)
					PUT8(UART_THR, write_buffer[write_counter++]);//eprintf("%c", write_buffer[write_counter++]);
			}while(write_buffer[write_counter] != '\0');
			PUT8(UART_IER, (GET8(UART_IER) & ~(0x2)));
			int_flag = WRITE_INT;
		}

		else if((iir & 0xF) == 0x4)	// Receive Interrupt
		{
			Set_LED("NFRX");
			temp_read = GET8(UART_RBR);
		
			do
			{
				hal_delay_us(1000);
				read_buffer[read_counter] = temp_read;
				PUT8(UART_THR, read_buffer[read_counter]);
				read_counter++;
			
				do
				{
					temp = GET8(UART_LSR) & LSR_DR;
				} while(!temp);

				temp_read = GET8(UART_RBR);
			} while(temp_read != '\r');
			PUT8(UART_IER, ENABLE_UART);	// Disable Interrupt mode
			int_flag = READ_INT;
		}
	}
}

//
// Function: _LED_Display_Cause_Of_Failure
//
// Description: Checks the LSR and display the kind of error
//
static void _LED_Display_Cause_Of_Failure(void)
{
	unsigned int lsr;

	if( (GET8(UART_IIR) & 0x6) == 0x6)	// Receive Error
	{
		// Check LSR and display cause of failure through Set_LED(char* buffer)
		lsr = GET8(UART_LSR);

		if( (lsr & (1<<1)) != 0)	// bit 1 set: Overrun Error
		{
			Set_LED(" OE ");
		}
		if( (lsr & (1<<2)) != 0)	// bit 2 set: Parity Error
		{
			Set_LED(" PE ");
		}
		if( (lsr & (1<<3)) != 0)	// bit 3 set: Framing Error
		{
			Set_LED(" FE ");
		}
		if( (lsr & (1<<7)) != 0)	// bit 7 set: FIFO Error Status
		{
			Set_LED("FFE ");
		}
	}
}

//
// Function: _Test_Register
//
// Description: Stores original register value, write-read-verify and then retore
//
static int _Test_Register(unsigned int reg_addr, unsigned int var)
{
	PUT8(UART_SPR, var);	// Store IER into Scratchpad register for temp

	PUT8(reg_addr, 1);		// Write a one to UART_IER
	var = GET8(reg_addr);	// Read back UART_IER
	if(var != 1)			// Check if IER is one
	{
		_LED_Display_Cause_Of_Failure();
		return ERROR;
	}

	PUT8(reg_addr, 0);		// Write a zero to UART_IER
	var = GET8(reg_addr);	// Read back UART_IER
	if(var != 0)			// Check if IER is zero
	{
		_LED_Display_Cause_Of_Failure();
		return ERROR;
	}

	var = GET8(UART_SPR);	// Read original IER value from Scratchpad Register
	PUT8(reg_addr, var);	// Write back the original IER

	return OK;
}
