
/*
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2001 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 *
 *   AUTHOR:         $Author: gdvaddad $
 *   DATE:           $Date: 2002/12/08 06:38:39 $
 *   LAST MODIFIED:  $Modtime: 4/03/02 3:20p $
 *   REV#:           $Revision: 1.3 $
 *   LOGFILE:        $Logfile: /Embedded/source/drivers/ixf/bb_util.c $
 *   WORKFILE:       $Workfile: bb_util.c $
 *
 * --------------------------------------------------------------------------
 * Project: ALL
 *
 * Purpose: Implement error handling & utility portion of driver software
 *			
 * Notes:
 * Public Procedures:
 *
 * Support Functions for Error Handling and DebugCfg Linkage
 *    bb_Sysfail                 Fatal Error Handler
 *    bb_Exception               Exception Handler
 *    bb_Warning                 Warning Message Handler
 *
 * Support Functions for Chip Drivers
 *    bb_Make16ByteWorkString    Build 16-byte J0, J1, J2 Work String
 *    bb_Make64ByteWorkString    Build 64-byte J1 Work String
 *    bb_InsertCRC7              Calculate & insert Tx CRC7 byte
 *
 * Support Functions - used only for UNIT-TEST (Use printf and putchar)
 *    bb_HexDump                 Dump bytes in hex
 *    bb_PrintReg                Print 8-bit reg
 *    bb_Print2RegsBE            Print 16-bit value from 2 regs (Big Endian)
 *    bb_Print2RegsLE            Print 16-bit value from 2 regs (Little Endian)
 *
 * User defined Procedures:
 *    PutString                  Put string
 *    Exit                       Exit during Sysfail
 *
 * --------------------------------------------------------------------------
 * CompilerFlags: 
 * --------------------------------------------------------------------------
 * --------------------------------------------------------------------------
 *   $Log: bb_util.c,v $
 *   Revision 1.3  2002/12/08 06:38:39  gdvaddad
 *   phase 5c/6
 *
 *   Revision 1.2  2002/10/25 20:24:49  machambe
 *   no message
 *
 *   Revision 1.1  2002/10/05 00:08:38  machambe
 *   no message
 *
 *   Revision 1.1  2002/07/21 01:34:38  machambe
 *   no message
 *
 * 
 * 33    4/05/02 12:34p Machambe
 * Changing Build Type to APPL_HANDLES_ERR
 * 
 * 32    12/04/01 4:22p Pkjain2
 * 
 * 31    11/15/01 10:00a Machambe
 * Changing bb_util.h to ixf_api_d.h
 * 
 * 29    5/03/01 3:53p Machambe
 * 
 * --------------------------------------------------------------------------
 */

#ifdef VXWORKS
#include "vxWorks.h"
#endif

#include "common_types.h"
#include "common_def.h"
#include "ixf_api_d.h"     /* Get Common System function interface files */  

#include <stdio.h>         /* Get printf variants, namely vsprintf */
#include <stdarg.h>        /* Get defs for variable length arg list */


/***********************************************************************
 * Max limits 
 **********************************************************************/

#define bb_BUFF_LEN     256  /* Buf Len for formatted print line */
#define bb_MAX_MSG_LEN  (bb_BUFF_LEN - 2) /* leave room for \n and \0 */

char*               bb_pFuncName;  /* -> printable function name,
                                             for bb_Sysfail, bb_Exception */

/***********************************************************************
 * User defined error handling functions
 **********************************************************************/

static bb_BuildType_e      BuildType=bb_APPL_HANDLES_ERRS;  /* Build type */


/***********************************************************************/
/***********************************************************************
 * Procedure Name:   Exit
 *
 * Description:      User defined Exit routine called by bb_Sysfail
 *                   during bb_PRODUCTION_BUILD
 *
 * End
 **********************************************************************/
void Exit(void)
{
    while(1)
    {/*Default Exit */
    }
}
/***********************************************************************
 * Procedure Name:   PutString
 *
 * Description:      User defined PutString
 *
 * End
 **********************************************************************/
void PutString(char* pString)
{
    printf("%s", pString);    /* Default User defined function */
}

/***********************************************************************/


/***********************************************************************/
/***********************************************************************
 * Procedure Name:   bb_Sysfail
 *
 * Description:      Fatal Error handler
 *
 * PDL:
 *    Remember that a fatal error has occured
 *    Write out Fatal Error header
 *    Format string, with variable data, into string, using vsprintf
 *    If string overwrote area allocated for it
 *       Put out msg, using caller's PutString callback, if one exists
 *       Shorten msg to fit
 *    Endif
 *    Add \n and null terminator to string
 *    Put out formatted msg, using caller's PutString callback, if one exists
 *    If application handles errors
 *       return
 *    Endif
 *    If production build,
 *       Take user's Fatal Error exit, if one exists
 *    Else [Development build, or no Fatal Error Handler exists]
 *       Hang - to preserved debug info
 *    Endif
 * End
 **********************************************************************/
void bb_Sysfail          /* System Failure routine supplied by SFT  */
(char*   pFormat,     /* Format string, suitable for printf call */
...)                 /* 0-n pieces of specific data, as appropriate
                            for the specified format */

{
    boolean    StackCorrupted;
    uint       StringLen, X;
    va_list    pArg;          /* -> each arg in variable length arg list */
    char       FormattedString[bb_MAX_MSG_LEN]; /* place to hold formatted msg */
    bb_Error_e ErrorType;



    /* Format and display user's msg */

    va_start(pArg, pFormat);/* pArg -> first arg, after pFormat */

    if (pFormat[0] == 0x00)
    {
        ErrorType = va_arg(pArg, bb_Error_e);
        PutString("\n");
    }
    else
    {
        /* Put out Sysfail Header msg, including External Function's Name */
        PutString("\nSYSFAIL SYSFAIL SYSFAIL SYSFAIL in ");
        StringLen = vsprintf(FormattedString, pFormat, pArg);
        /* Format msg into buff */
        if (StringLen >= bb_MAX_MSG_LEN)
        {
            StringLen = bb_MAX_MSG_LEN;
            StackCorrupted = TRUE;
        }
        else
            StackCorrupted = FALSE;

        FormattedString[StringLen++]  = '\n';  /* Add Newline */
        FormattedString[StringLen]    = '\0';  /* Put back null terminator */

        PutString(FormattedString);  /* Put out fatal err msg */

        if (StackCorrupted)
            PutString("bb_Sysfail: Sysfail msg too long; stack corrupted\n");

        /* Depending on Build Type - Hang, Reset or Return */
    }

    va_end(pArg);           /* done - clean up */
    if (BuildType == bb_APPL_HANDLES_ERRS)   /* If appl handles errors */
    return;  /* return now */

    /* If Production Build -- Take Fatal Error exit (to reset) */
    else if (BuildType == bb_PRODUCTION_BUILD)
    Exit();          /* Take fatal error exit now */
    else if (BuildType == bb_DEVELOPMENT_BUILD)
    goto HangAfterSysfail;  /* Hang to preserve debug info */
    else              /* If undefined & probably unintialized */
        return;        /* Don't hang - just return now */

    /* Development Build -- Hang here to preserve debug info */
HangAfterSysfail:
    X++; /* Don't optimize this loop out */
    goto HangAfterSysfail;


}  /* end bb_Sysfail */


/***********************************************************************/
/***********************************************************************
 * Procedure Name:   bb_Warning
 *
 * Description:      Warning Error handler
 *
 * PDL:
 *    Write out Warning Error header
 *    Format string, with variable data, into string, using vsprintf
 *    If string overwrote area allocated for it
 *       Put out msg, using caller's PutString callback, if one exists
 *       Shorten msg to fit
 *    Endif
 *    Add \n and null terminator to string
 *    Put out formatted msg, using caller's PutString callback, if one exists
 * End
 **********************************************************************/
void bb_Warning            /* Handle Warning */
(char*   pFormat,       /* Format string, suitable for printf call */
...)                   /* 0-n pieces of specific data, as appropriate
                              for the specified format */
{
    boolean    StackCorrupted;
    uint       StringLen;
    va_list    pArg;          /* -> each arg in variable length arg list */
    char       FormattedString[bb_MAX_MSG_LEN]; /* place to hold formatted msg */



    /* Format and display user's msg */
    va_start(pArg, pFormat);/* pArg -> first arg, after pFormat */
    StringLen = vsprintf(FormattedString, pFormat, pArg);
    /* Format msg into buff */
    if (pFormat[0] == 0x00)
    {
        /* ErrorType = va_arg(pArg, bb_Error_e); */
        PutString("\n");
    }
    else
    {
        /* Put out Warning Header msg, including External Function's Name */
        PutString("\nWARNING for ");

        if (StringLen >= bb_MAX_MSG_LEN)
        {
            StringLen = bb_MAX_MSG_LEN;
            StackCorrupted = TRUE;
        }
        else
            StackCorrupted = FALSE;

        FormattedString[StringLen++]  = '\n';  /* Add Newline */
        FormattedString[StringLen]    = '\0';  /* Put back null terminator */

        PutString(FormattedString);  /* Put out user's err msg */

        if (StackCorrupted)
        {
            bb_pFuncName = "bb_Warning";
            bb_Sysfail("Warning msg too long; stack corrupted\n");
        }
    }
    va_end(pArg);           /* done - clean up */

} /* end bb_Warning */


/***********************************************************************
 * Useful common suppport functions provided for specific chip drivers
 **********************************************************************/
/***********************************************************************
 * Procedure Name:   bb_Make16ByteWorkString
 *
 * Description:      Build 16-byte J0, J1, J2 Work String:
 *                   0, Jn string, Pad of 0's
 *
 * Conditions for Use:
 *    pWorkString != NULL
 *
 * Notes:
 *    WorkString is printable; i.e. there will always be a 0 at end,
 *    even if max length.
 *
 * End
 **********************************************************************/
uchar* bb_Make16ByteWorkString   /* Build 16-byte J0, J1, J2 Work String:
 0, Jn string, Pad of 0's;
 Return -> Work String */
(char*   pString)             /* -> J0, J1, J2 null-terminated string */
                                 /* Note:
                                    if pString = NULL, WorkString = 0's */
{
    static uchar WorkString[18];  /* Leave room for null terminator */
    uchar* pWorkString = WorkString;
    int   Len;
    char  Zero=0;

    if (pString == 0)
        pString = &Zero;     /* If pString = NULL, make WorkString = 0's */

    *pWorkString++ = 0;     /* Put 0 at start of string */

    Len = 15;
    while (Len-- > 0)       /* move rest of string, up to max of 15 bytes */
    {
        if ((*pWorkString++ = (uchar)*pString++) == 0)
            /* move string byte or null-term */
            break;            /* if moved null-term - done with user's string */
    }
    while (Len-- > 0)
        *pWorkString++ = 0;  /* pad rest of work string, with 0's */

    *pWorkString = 0;       /* Add null terminator after workstring,
    to make printable */

    return WorkString;

} /* end bb_Make16ByteWorkString */

/***********************************************************************/
/***********************************************************************
 * Procedure Name:   bb_Make64ByteWorkString
 *
 * Description:      Build 64-byte J1 Work String:
 *                   J1 string, Pad of 0's
 *
 * Conditions for Use:
 *    pWorkString != NULL
 *
 * Notes:
 *    WorkString is printable; i.e. there will always be a 0 at end,
 *    even if max length.
 *
 * End
 **********************************************************************/
uchar* bb_Make64ByteWorkString   /* Build 64-byte J1 Work String:
 J1 string, Pad of 0's */
(char*   pString)             /* -> J1 null-terminated string */
                                 /* Note:
                                    if pString = NULL, WorkString = 0's */
{
    static uchar WorkString[66];  /* Leave room for null terminator */
    uchar* pWorkString = WorkString;
    int   Len;
    char  Zero=0;

    if (pString == 0)
        pString = &Zero;     /* If pString = NULL, make WorkString = 0's */

    Len = 64;
    while (Len-- > 0)       /* move string, up to max of 64 bytes */
    {
        if ((*pWorkString++ = (char)*pString++) == 0)
            /* move string byte or null-term */
            break;            /* if moved null-term - done with user's string */
    }
    while (Len-- > 0)
        *pWorkString++ = 0;  /* pad rest of work string, with 0's */

    *pWorkString = 0;       /* Add null terminator after workstring,
    to make printable */

    return WorkString;

} /* end bb_Make64ByteWorkString */

/***********************************************************************/
/***********************************************************************
 * Procedure Name:   bb_InsertCRC7
 *
 * Description:      Calculate & insert CRC7 byte,
 *                   at start of work string
 *
 * Conditions for Use:
 *    Work string is exactly 16 bytes, with dummy byte at start.
 *
 * End
 **********************************************************************/
void bb_InsertCRC7             /* Calculate & insert CRC7 byte,
                                    at start of work string */
(uchar*  pWorkString)         /* -> Work String, from function above, with
                                    dummy byte at start of string, for CRC7 */
{
    typedef struct _bits
    {
unsigned b0 : 1;
unsigned b1 : 1;
unsigned b2 : 1;
unsigned b3 : 1;
unsigned b4 : 1;
unsigned b5 : 1;
unsigned b6 : 1;
unsigned b7 : 1;
    }  BITS;

    int   i;
    BITS  crc, xd, j;
    uchar *pCrc, *pXd;
    uchar *p;

    p = pWorkString;        /* -> first byte of Work String */

    *p    = 0x80;           /* Initialize CRC7 byte in string */
    pCrc  = (uchar *) &crc; /* get around type checking */
    pXd   = (uchar *) &xd;  /* or using a using a Union */
    *pXd  = *pCrc = 0;      /* Initialize CRC7 byte */

    for (i = 0; i < 16;  i++, p++)
    {
      #if(0)   /* Replace this */
        memcpy (&j, p, 1);    /* sizeof (j) is 1 byte */
      #endif
        *((uchar*)&j) = *p;

        *pXd = (uchar) (*p ^ *pCrc);
        crc.b7 = xd.b7 ^ xd.b6 ^ xd.b3;
        crc.b6 = xd.b6 ^ xd.b5 ^ xd.b2;
        crc.b5 = xd.b5 ^ xd.b4 ^ xd.b1;
        crc.b4 = xd.b4 ^ xd.b3 ^ j.b0;
        crc.b3 = xd.b6 ^ xd.b2;
        crc.b2 = xd.b5 ^ xd.b1;
        crc.b1 = xd.b7 ^ xd.b4 ^ j.b0;
    }
    *pCrc = (uchar) (*pCrc >> 1);
    *pCrc |= 0x80;          /* Set MSB to identify CRC7 byte*/
    *pWorkString = *pCrc;   /* Insert CRC7 at start of Work String */

} /* end bb_InsertCRC7 */

/***********************************************************************
 * Useful common suppport functions provided for Debug Printing
 * NOTE: -- For UNIT-TESTING ONLY: Uses putchar and printf
 **********************************************************************/

#ifdef UNIT_TEST

/***********************************************************************
 * Procedure Name:   bb_HexDump
 *
 * Description:      Dump n bytes in hex
 *
 * Conditions for Use:
 *    none
 *
 * End
 **********************************************************************/
void bb_HexDump(char* pVar, uint VarLen) /* Dump VarLen bytes at pVar */

{
    uchar Byte;
    uint  Val;

    while (VarLen--)           /* Dump var in hex */
    {
        Byte = *pVar++;

        Val = (Byte & 0xF0) >> 4;
        putchar(Val < 10 ? Val + '0' : Val - 10 + 'A');

        Val = Byte & 0x0F;
        putchar(Val < 10 ? Val + '0' : Val - 10 + 'A');
    }
}

/***********************************************************************
 * Procedure Name:   bb_BinByte
 *
 * Description:      Dump 1 byte in binary
 *
 * Conditions for Use:
 *    none
 *
 * End
 **********************************************************************/
void bb_BinByte(uchar Byte) /* Dump byte in binary */

{
    uchar Mask;
    int   k;

    k = 4;
    Mask = 0x80;
    while (k-- > 0)   /* for first 4 bits */
    {
        if (Mask & Byte)  /* if bit set */
            putchar('1');
        else
            putchar('0');
        Mask >>= 1;
    }
    putchar(' ');  /* space between first & last 4 bits */

    k = 4;
    while (k-- > 0)   /* for first 4 bits */
    {
        if (Mask & Byte)  /* if bit set */
            putchar('1');
        else
            putchar('0');
        Mask >>= 1;
    }
}

/***********************************************************************
 * Procedure Name:   bb_PrintReg
 *
 * Description:      Print:
 *                   Title Reg %p = %X\n
 *
 * Conditions for Use:
 *    none
 *
 * End
 **********************************************************************/
void bb_PrintReg     /* Print 1 8-bit reg, with title */
(char*                  pTitle,  /* e.g.: "InitOht: OCR1" */
bb_RegPointer_type     pReg)    /* -> Reg */
{
    printf("    %s Reg Address 0x%x = ", pTitle, pReg);

    bb_HexDump((char*)pReg, 1); /* Dump 1 byte - in hex */
    putchar(' ');
    putchar('(');
    bb_BinByte(*(uchar*)pReg);   /* Dump 1 byte in binary */
    putchar(')');
    putchar('\n');    /* End line */
}

/***********************************************************************
 * Procedure Name:   bb_Print2RegsBE   Big Endian (MSB in low byte)
 *
 * Description:      Print:
 *                   Title Regs %p-%p = %X%X\n
 *
 * Conditions for Use:
 *    none
 *
 * Notes:
 *    Print 16-bit value from 2 regs
 *
 * End
 **********************************************************************/
void bb_Print2RegsBE /* Print 2 8-bit regs (Big Endian) (MSB in low byte) */
(char*                pTitle,  /* e.g.: "LOF_LMN" */
bb_RegPointer_type   pMsbReg) /* -> MSB Reg */

{
    printf("    %s Regs 0x%x-0x%x = ", pTitle, *(uchar*)pMsbReg, *((uchar*)pMsbReg + 1));

    bb_HexDump((char*)pMsbReg,1);     /* Dump MSB */
    bb_HexDump((char*)pMsbReg+1,1);   /* Dump LSB */

    putchar(' ');
    putchar('(');
    bb_BinByte(*(uchar*)pMsbReg);      /* Dump MSB in binary */
    putchar(' ');
    putchar(' ');
    bb_BinByte(*((uchar*)pMsbReg+1));  /* Dump LSB in binary */
    putchar(')');
    putchar('\n');
    /* End line */
}

/***********************************************************************
 * Procedure Name:   bb_Print2RegsLE   Little Endian (MSB in High byte)
 *
 * Description:      Print:
 *                   Title Regs %p-%p = %X%X\n
 *
 * Conditions for Use:
 *    none
 *
 * Notes:
 *    Print 16-bit value from 2 regs
 *
 * End
 **********************************************************************/
void bb_Print2RegsLE /* Print 2 8-bit regs (Little Endian) (MSB in High byte) */
(char*              pTitle,  /* e.g.: "B2_SET_WIN_SIZE" */
bb_RegPointer_type pMsbReg) /* -> MSB Reg */
{
    printf("    %s Regs 0x%x-0x%x = ", pTitle, (uchar*)pMsbReg, (uchar*)pMsbReg - 1);

    bb_HexDump((char*)pMsbReg,1);     /* Dump MSB */
    bb_HexDump((char*)pMsbReg-1,1);   /* Dump LSB */

    putchar(' ');
    putchar('(');
    bb_BinByte(*(uchar*)pMsbReg);      /* Dump MSB in binary */
    putchar(' ');
    putchar(' ');
    bb_BinByte(*((uchar*)pMsbReg-1));  /* Dump LSB in binary */
    putchar(')');
    putchar('\n');          /* End line */
}

#endif   /* UNIT_TEST */


