/******************************************************************************
*
* Module: eed.cpp
*
* Description: Ethernet Module Definitions
*
* Project:     HSI VisionWare -- vWare
*
* Functions:
*
* Notes:
*
* History:		See Visual SourceSafe History
*
* Disclaimer: Copyright 1999-2000 Wind River Systems, Inc. All Rights Reserved
*
*******************************************************************************/
#include "vm.h"
#include "eed.h"
#include "estack.h"
#include "daemon.h"
#include "tcp.h"
#include "cmd.h"
#include "cmds.h"
#include "vconsole.h"


extern tEEDRIVER Eedriver;
extern TelnetClass Port1234Daemon;
extern TelnetClass Port1235Daemon;
extern TelnetClass Port1236Daemon;

extern "C" int CheckForPush(char *, unsigned int);

EED Eed;
void (*eedCallback) (UINT8 *, UINT32) = NULL;	/* ethernet callback for application use */

EED::EED(void)
{
	AU *pAU;
	UINT32 AllocationSize = (ALLOCATIONSIZE / sizeof(AU)) * sizeof(AU);
	
	// Initialize the allocation units.
	pAU = (AU *)(pAllocationUnits = (void *)BufferMalloc(AllocationSize));

	pAllocationUnitsEnd = (void *)((UINT32)pAllocationUnits + AllocationSize);
	pNextAllocationUnit = pAllocationUnits;
	while (pAU < pAllocationUnitsEnd)
	{
		pAU->Units = 1;
		pAU->ID = 0;
		pAU++;
	}

	// Initialize the global MAC and IP addresses
	MyIPADDR = ntohl(*(UINT32 *)(GetShellVarPointer(SHELLVAR_IP)));
	memcpy((void *)&MyMAC, GetShellVarPointer(SHELLVAR_MAC), sizeof(MAC));

	EED::Open(&hDevice);

	// Init this global pointer so that daemon objects can find us.
	daemon::EthernetDriver = this;
}

DSTATUS EED::Open(HANDLE *pHandle, void *pArgs)
{
	if (!hDevice)
		// Call the board-specific initialization routine.
		hDevice = *pHandle = Eedriver.Open(&Eedriver);

	// Add an entry to the log if the device fails to open
	if (!hDevice)
		vLOG("EED::EED: Ethernet driver open failed", NULL, NULL);

	// Call ConsoleInit if this is supposed to be the default command channel.
#if (COMMANDCHANNELOUT == DEFAULTETHERNET)
	ConsoleInit(this);
#endif

	return STATUS_OK;	
}

/*  
 *  The Close() method does NOT use the dev parameter.  The dev parameter
 *  is a relic from previous (non-C++) versions of this driver.  It
 *  is included here to match the Read/Write/Ioctl functions.
 */
DSTATUS EED::Close(HANDLE dev)
{
	if (hDevice)
		Eedriver.Close(hDevice);
	hDevice = 0;	
	
	/* If we are using the ethernet port console, close it too  */
	if (&Eed == pCommandChannel)
		pCommandChannel = NULL;
		
	return STATUS_OK;
}

/***************************************************************************
* Function:	Write(HANDLE dev, char *buffer, unsigned int nBytes)
*
* Purpose:	Write packet out channel
*
* Returns:	DSTATUS
****************************************************************************/
DSTATUS EED::Write(HANDLE dev, char *buffer, unsigned int nBytes)
{
	UINT32 Push = CheckForPush(buffer, nBytes) ? TCP_PUSHREQUEST : 0;

	if (Port1234Daemon.State == TCPSTATE_ESTABLISHED)
		Port1234Daemon.Write((UINT8 *)buffer, nBytes, Push);
	else
	{
		if (Port1236Daemon.State == TCPSTATE_ESTABLISHED)
			Port1236Daemon.Write((UINT8 *)buffer, nBytes, Push);
		else
		{
			if (Port1235Daemon.State == TCPSTATE_ESTABLISHED)
				Port1235Daemon.Write((UINT8 *)buffer, nBytes, Push);
		}
	}
	return PACKET_OK;
}
   
/***************************************************************************
* Function:	Ioctl(HANDLE dev, unsigned int funcCode, void *pArgs)
*
* Purpose:	IO settings for channel
*
* Returns:	DSTATUS
****************************************************************************/
DSTATUS EED::Ioctl(HANDLE dev, unsigned int funcCode, void *pArgs)
{
	return Eedriver.Ioctl(hDevice, funcCode, pArgs);
}

/***************************************************************************
* Function:	GetTxBuffer(UINT32 Length)
*
* Purpose:	Get next allocation unit
*
* Returns:	UINT8 *
****************************************************************************/
UINT8 *EED::GetTxBuffer(UINT32 Length)
{
	int i;
	UINT32 UnitsNeeded = (Length + AUOVERHEAD + (sizeof(AU) - 1))/sizeof(AU);
	UINT32 ContiguousUnits;
	union
	{
		AU *pAU;
		UINT8 *pUINT8;
		void *pVoid;
		UINT32 value;
	} pSearch;

	pSearch.pVoid = pNextAllocationUnit;

	for (;;)
	{   
		ContiguousUnits = 0;
		
		while ((pSearch.pVoid < pAllocationUnitsEnd)
				&& (ContiguousUnits < UnitsNeeded))
		{
			switch ((UINT32)(pSearch.pAU->ID))
			{
				case 0:				// allocation unit is free
					ContiguousUnits += pSearch.pAU->Units;
					break;

				case 1:				// allocation unit is claimed but not yet assigned
					ContiguousUnits = 0;
					break;

				default:			// allocation unit is assigned
					if (Eedriver.IsFreeBuffer(hDevice, pSearch.pAU->ID))
					{				// OK to free these allocation units
						pSearch.pAU->ID = 0;
						ContiguousUnits += pSearch.pAU->Units;
					}
					else
						ContiguousUnits = 0;
			}
			pSearch.pAU += pSearch.pAU->Units;
		}

		if (ContiguousUnits >= UnitsNeeded)
		{
			if (ContiguousUnits > UnitsNeeded)
			{
				pSearch.pAU -= (i = (ContiguousUnits - UnitsNeeded));
				pSearch.pAU->ID = 0;
				pSearch.pAU->Units = i;
			}

			pNextAllocationUnit = pSearch.pVoid;   
			pSearch.pAU -= UnitsNeeded;
			pSearch.pAU->ID = 1;
			pSearch.pAU->Units = UnitsNeeded;
			return (pSearch.pUINT8 + AUOVERHEAD);
		}
		
		pSearch.pVoid = pAllocationUnits;
	}
}

/***************************************************************************
* Function:	Send(UINT8 *, UINT32)
*
* Purpose:	Send a packet composed in a single buffer.
*
* Returns:	nothing
****************************************************************************/
void EED::Send(UINT8 *pBuffer, UINT32 Length)
{
	Eedriver.Write(hDevice, pBuffer, Length, 0, 0, 0, 0);
}

/***************************************************************************
* Function:	Send(UINT8 *, UINT32, UINT8 *, UINT32)
*
* Purpose:	Send a packet composed in two buffers.
*
* Returns:	nothing
****************************************************************************/
void EED::Send(UINT8 *pBuffer1, UINT32 Length1, UINT8 *pBuffer2, UINT32 Length2)
{
	Eedriver.Write(hDevice, pBuffer1, Length1, pBuffer2, Length2, 0, 0);
}

/***************************************************************************
* Function:	Send(UINT8 *, UINT32, UINT8 *, UINT32, UINT8 *, UINT32)
*
* Purpose:	Send a packet composed in three buffers.
*
* Returns:	nothing
****************************************************************************/
void EED::Send(UINT8 *pBuffer1, UINT32 Length1, UINT8 *pBuffer2, UINT32 Length2,
		UINT8 *pBuffer3, UINT32 Length3)
{
	Eedriver.Write(hDevice, pBuffer1, Length1, pBuffer2, Length2,
			pBuffer3, Length3);
}

void EED::FreeBuffer(UINT8 *pData)
{
	AU *pAU;
	int i, j;

	// If the buffer pointed to by this BD is part of the communal TX buffer
	// (pAllocationUnits) then free it.
	if ((pData >= pAllocationUnits) && (pData < pAllocationUnitsEnd))
	{
		pAU = (AU *)(pData - AUOVERHEAD);
		for (j = i = pAU->Units; i; --i)
		{
			pAU->Units = 1;
			pAU->ID = 0;
			pAU++;
		}
	}
}

extern "C" void EedriverFreeBuffer(UINT8 *pData)
{
	Eed.FreeBuffer(pData);
}

void EED::AssignBuffer(UINT8 *pData, UINT32 ID)
{
	AU *pAU;

	if ((pData >= pAllocationUnits) && (pData < pAllocationUnitsEnd))
	{
		pAU = (AU *)((UINT8 *)pData - AUOVERHEAD);
		pAU->ID = ID;
	}
}

extern "C" void EedriverAssignBuffer(UINT8 *pData, UINT32 ID)
{
	Eed.AssignBuffer(pData, ID);
}

Component("Ethernet", COMPONENT_ETHERNET, Eed);
extern "C" void Ethernet(void *pIn, void *pOut)
{
	pCommandChannel = &Eed;
}

extern "C" void ProcessPacket(ETHERNETPACKET *p, unsigned int length);

extern "C" void EedriverProcessPacket(UINT8 *p, unsigned int length)
{

	if (eedCallback != NULL)	/* use callback if set */
		eedCallback (p, length);
	else
    	ProcessPacket((ETHERNETPACKET *)p, length);
}

/***************************************************************************
* Function: Ethernet
*
****************************************************************************/
Command(CMD_ETHERNET, "ethernet", Ethernet, WHILERUNNING, "Ethernet", "switch to the ethernet channel");
CommandEnd();

/***************************************************************************
 * Function: void EthernetLaunch(tOsParams *pOS)
 * 
 * Purpose:	 Called prior to launching an OS.  Completes the Ethernet and
 * 	         TCP/IP fields of the tOsParams structure that gets passed to
 *           the operating system.  Also calls the driver's Close function.
 * 
 * Returns:  nothing
 ***************************************************************************/	  
extern "C" void EthernetLaunch(tOsParams *pOS)
{
	universal u;

	u.pVoid = GetShellVarPointer(SHELLVAR_IP);
	pOS->IpAddress = *u.pWord;
	
	u.pVoid = GetShellVarPointer(SHELLVAR_SUBNET);
	pOS->Subnet = POS_LSWAP(*u.pWord);

	pOS->Subnet = *u.pWord;
	
	u.pVoid = GetShellVarPointer(SHELLVAR_GATEWAY);
	pOS->DefaultGateway = *u.pWord;
	
	u.pVoid = GetShellVarPointer(SHELLVAR_REMOTEIP);
	pOS->PrimaryHost = *u.pWord;
	
	u.pVoid = GetShellVarPointer(SHELLVAR_MAC);
	memcpy(&pOS->MacBytes, u.pVoid, 6);

	Eed.Close(Eed.hDevice);
}

Fini(0, EthernetLaunch);
