/* tcp.h - TCP protocol and Telnet support  */

/* Copyright 2002 Wind River Systems, Inc.  */

/*
modification history
--------------------
01a,-------,BL   SeqCompare & Nortel modifications
01a,11Jun02,CJW  Several changes/additions:
                 -  Added packet history support
                 -  Added state history support
                 -  Fixed SEQCMP bug
01b,27Jun02,RCS  Added comment to vector class relating to out of sequence packets
				 - Extended class to include ProcessOOOPS
*/

#ifndef __INCtcph
#define __INCtcph

#include "estack.h"
#include "eed.h"
#include "daemon.h"
#include "circular.h"
#include "timer.h"
#include "ip.h"

/* defines */

/*
 * TXBYTETHRESHOLD is the minimum number of unpushed bytes that must
 * accumulate before data will be sent (unless TXTIMETHRESHOLD elapses).
 */
 
#define TXBYTETHRESHOLD 64

/*
 * TXTIMETHRESHOLD is the maximum length of time that any unpsushed data
 * will remain unsent (in milliseconds).
 */
 
#define TXTIMETHRESHOLD 100

/*
 * TCBS is the maximum number of unacknowledged transmitted packets
 * that can be extant at any given time.
 */
 
#define TCBS    10

/*
 * TXMAXTRIES is the maximum number of times a packet will be sent without
 * receiving an ACK before the connection is broken.
 */
 
#define TXMAXTRIES      5


/* OOOS is the maximum number of out-of-order packets which can be buffered. */
 
#define OOOPS           4

/* MAX_HISTORY is the number of values in the state history trace */

#define MAX_HISTORY     20

/* MAX_PACKET_HISTORY is the number of packets in the packet history trace */

#define MAX_PACKET_HISTORY  30

#define MSL             30      /* Maximum segment life -  30 seconds */

/*
 * SEQCMP - sequence space comparator
 * This handles sequence space wrap-around. Overflow/Underflow makes
 * the result below correct (-,0,+) for any a,b in the sequence
 * space.
 */
 
#define SEQCMP(a,b) ((INT32)((a)-(b)))

/* typedefs */

/* enum TCPSTATE is the set of all possible TCP states */

enum TCPSTATE
{
    TCPSTATE_CLOSED,
    TCPSTATE_LISTEN,
    TCPSTATE_SYNSENT,
    TCPSTATE_SYNRCVD,
    TCPSTATE_ESTABLISHED,
    TCPSTATE_FINWAIT1,
    TCPSTATE_FINWAIT2,
    TCPSTATE_CLOSEWAIT,
    TCPSTATE_CLOSING,
    TCPSTATE_LASTACK,
    TCPSTATE_TIMEWAIT
};

/* TCB is a type used to maintain a list of unacknowledged packets */

typedef struct
{
    TIMEOUT t;
    UINT8 *pData;
    UINT32 SeqNumber;
    UINT16 Bytes;
    UINT8 Tries;
    UINT8 Flags;
    UINT32 SeedValue;
    UINT32 Options;
    UINT64 BaseTime;
    UINT32 LastTimer;
    UINT16 UrgentPointer;
    UINT16 Padding;
} TCB;

/* OOOP is used to store Out-Of-Order Packets (OOOP) */

typedef struct
{
    CIRCULARBUFFER cb;
    UINT32 Options;
    UINT8 Flags;
    UINT8 HeaderLength;
    UINT16 UrgentPointer;
    UINT32 SeqNumber;
    UINT32 Length;
} OOOP;

#ifdef TCP_PACKET_HISTORY

/* Data stored and displayed with each packet in the packet history */

typedef struct _PacketHistoryType
{
    UINT32 SeqNo;
    UINT32 DataLength;
    UINT16 Checksum;
} PacketHistoryType;

#endif

/* function declarations */

/* 
 * InitialSeqNumber() reads the time base to obtain an initial 
 * packet sequence number  
 */
 
extern "C" UINT32 InitialSeqNumber(void);

/* Class definition for a TCP socket. */

class TCPSocket : public IPSocket
{
protected:
    TIMEOUT        AggregateTimer;  /* Data packets are sent when this      */
                                    /* timer expires.                       */
    TIMEOUT        AckTimer;        /* Expires after 200 ms, indicating     */
                                    /* need to send an ACK packet.          */
    TIMEOUT        ZeroProbeTimer;  /* Unused                               */
    TCB            Tcbs[TCBS];      /* Array of unacknowledged packets      */
    CIRCULARBUFFER TxCB;            /* Transmit buffer                      */
    CIRCULARBUFFER RxCB;            /* Receive buffer                       */
    OOOP           Ooops[OOOPS];    /* Array of out of order packets        */
    UINT32         ActiveOoops;     /* number of out of order packets still */
                                    /* stored in Ooops[].                   */
    UINT32         LastError;       /* Unused                               */
    UINT32         LocalSeqNumber;  /* next byte I will send                */
    UINT32         RemoteSeqNumber; /* next byte I expect to receive        */
    UINT32         HighestDatagram; /* highest datagram ID received         */
    UINT32         LocalAckNumber;  /* highest ACK I received               */
    UINT32         RemoteAckNumber; /* highest ACK I sent                   */
    UINT32         SRTT;            /* smoothed packet roundtrip time       */
    UINT32         SDEV;            /* smoothed packet roundtrip deviation  */
    UINT32         RTO;             /* retransmission timeout               */
    UINT32         LocalMSS;        /* Local maximum segment size           */
    UINT32         RemoteMSS;       /* Remote maximum segment size          */
    UINT32         RemoteWindowSize;/* Maximum amount of data that may be   */
                                    /* sent                                 */
    TCPSTATE       InitialState;    /* Starting state of the socket.        */
    static TCPSocket * pTcpList;    /* List of all sockets                  */
    TCPSocket *    pNext;           /* Pointer to next socket in list       */
    
#ifdef TCP_STATE_HISTORY
    
    TCPSTATE stateHistory[MAX_HISTORY];  /* State history trace buffer      */
    UINT8    historyIndex;               /* Current index for history trace */
    
    void addHistory(TCPSTATE newState);  /* Adds new state to history trace */
    void ShowStateHistory(void);         /* Displays state history trace    */

#endif

#ifdef TCP_PACKET_HISTORY
    
    PacketHistoryType packetHistory[MAX_PACKET_HISTORY]; /* Packet history  */
                                                         /* trace buffer    */
    UINT8             packetHistoryIndex;                /* Current index   */
                                                         /* of packet trace */
                                                         
    /* Adds a packet to the packet history trace */
    
    void addPacket(UINT32 SeqNo, UINT16 Checksum, UINT32 DataLength);
    
    void ShowPacketHistory(void);    /* Display the current packet history  */                 

#endif


    /* Timeout functions */
    
    void PacketTimeout(TIMEOUT *p);
    void AckTimeout(TIMEOUT *p);
    void AggregateTimeout(TIMEOUT *p);
    
    /* Packet transmit functions */
    
    TCPPACKET *ConstructHeader(TCB *p);
    void SendAggregateData(UINT32 Flags);
    void Send(TCB *p);
    void SendFlags(UINT32 Flags, UINT32 Options = 0);

	/* Received packet handling */    
    UINT32 ReceiveWindowSize(void);
    void ProcessInOrderPacket(TCPPACKET *pTCP, UINT8 *pData, UINT32 Length);
    void Receive(void *v, UINT32 TcpPacketLength);
    void TargetedReceive(void *v, UINT32 TcpPacketLength);
	void ProcessOOOPS(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);

    /* Socket termination functions */
    
    void Abort(void);
    void SendReset(void);
    
    /* State handling functions */
    
    void ChangeState(TCPSTATE newState);  
    void Established(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);
    void Listen(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);
    void Closed(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);
    void SynSent(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);
    void SynRcvd(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);
    void FinWait1(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);
    void FinWait2(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);
    
    /* 
     * This function is called when a packet is received 
     * while in the TIMEWAIT state 
     */
     
    void TimeWait(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);  
    
    /* This function is called when the timer expires */
    
    static void TimeWait(TCPSocket *);   
    
    void Closing(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);
    void CloseWait(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);
    void LastAck(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);
    void TCPAccepted();

public:

	/* Constructor */
	
    TCPSocket(UINT8 Protocol, UINT16 TargetPort, UINT8 Flags, 
              UINT32 TxCBSize, UINT32 RxCBSize, TCPSTATE InitialState);
              
    /* Write data through the socket */              
    UINT32 Write(UINT8 *pBuffer, UINT32 nBytes, UINT32 Push);
    
    /* 
     * Function required to be defined by child classes.  Should handle
     * processing of packet.
     */
     
	/* The checksum on the pTCP parameter is not guaranteed when processing
		out of order packets.
	*/
    virtual void Vector(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength) = 0;
    
    /* Current state */
    
    TCPSTATE State;
};

class TelnetClass : public TCPSocket
{
public:
    TelnetClass(UINT8 Protocol, UINT16 TargetPort, UINT8 Flags, UINT32 TxCBSize, UINT32 RxCBSize, TCPSTATE InitState);
    void Vector(TCPPACKET *pTCP, UINT8 *pData, UINT32 DataLength);
};

#endif /* __INCtcph */
