/******************************************************************************
*
* Module:		tftpclient.cpp
*
* Description:	TFTP Functions
*
* Project:		HSI VisionWare -- vWare
*
* Functions:
*
* Notes:
*
* History:		See Visual SourceSafe History
*	BL - Handle TFTP Sequence Errors
*
* Disclaimer:	Copyright 1999-2000 Wind River Systems, Inc. All Rights Reserved
*
*******************************************************************************/
//  Includes
#include "vm.h"
#include "daemons.h"
#include "cmd.h"
#include "ffs.h"
#include "tcp.h"
#include "tftp.h"

//  External References
extern UINT32 ErrorAddress;
extern char *pExtension;
extern char *pFilename;
extern char *pFilespec;
extern char *pHost;
extern UINT8 ExtensionType;
extern UINT32 HostIP;
extern "C" int DeconstructFilename(char *p, BOOL HostRequired);
extern FFSHANDLE *gpHandle;
extern "C" void BeginDownload(UINT32 Bias);
extern UINT32 NextLoadAddress;
extern "C" UINT32 LoadMemory(UINT8 *, UINT32);
extern UINT8 *DownloadAddress;

//  Global Declarations
TFTPClient Tftp(17, 1800, 0, 69);
Component("TFTP", COMPONENT_TFTP, Tftp);

/****************************************************************************
* Function:	CopyChunk(UINT8 *pBuffer, UINT32 Length);
*
* Purpose:	copy a received TFTP packet to a file
*
* Returns:	status
****************************************************************************/
UINT32 CopyChunk(UINT8 *pBuffer, UINT32 Length)
{
	UINT32 BytesWritten;

	if (gpHandle)
	{
		if (WriteFile(gpHandle, pBuffer, Length, &BytesWritten))
		{
			FfsResult(pFilename);
			CloseFile(gpHandle);
			gpHandle = 0;
			return 3;
		}
		return 0;
	}
	return 3;
}

/****************************************************************************
* Function:	TftpWait(void)
*
* Purpose:	Once a TFTP transfer has been initiated this routine is called
*			to eat cycles until the result of the transfer is known.
*
* Returns:	0 = success
*			1 = ARP timeout
*			2 = write-protect error
*			3 = a flash failure
****************************************************************************/
extern "C" UINT32 TftpWait(void)
{
	UINT32 retval = 1;

	while (Tftp.State == TFTPSTATE_WORKING)
		;

	switch(Tftp.State)
	{
		case TFTPSTATE_COMPLETE:
			retval = 0;
			break;

		case TFTPSTATE_TIMEDOUT:
			Result(STACKERROR_TFTPTIMEOUT, "no response from server");
			break;

		case TFTPSTATE_WRITEPROTECT:
			Result(STACKERROR_WRITEPROTECT, "address 0x%08X is protected", ErrorAddress);
			break;

		case TFTPSTATE_ARPTIMEOUT:
			Result(STACKERROR_ARPTIMEOUT, "ARP timeout");
			break;

		case TFTPSTATE_FILEERROR:
			Result(STACKERROR_FILEERROR, "file error");
			break;

		// BL - Added Error Handling for Sequence Errors
		case TFTPSTATE_SEQERROR:
			Result(STACKERROR_TFTPSERVERABORT, "TFTP Sequence Error");
			break;

		case TFTPSTATE_SERVERERROR:
			Result(STACKERROR_TFTPSERVERABORT,
					"TFTP server aborted with error code %d, message is \"%s\"",
					Tftp.ErrorCode, Tftp.ErrorMsg);
			break;

		default:
			Result(STACKERROR_UNKNOWNERROR, "unknown TCP/IP error");
			break;
	}

	Tftp.State = TFTPSTATE_CLOSED;
	ProcessMode = 0;
	Tftp.pFfs->CloseFile(Tftp.Handle);
	return retval;
}

/****************************************************************************
* Function:	TftpDownload(UINT32 HostIP, UINT8 *Filename,
*						UINT32 (*Vector)(UINT8 *, UINT32));
*
* Purpose:	Download a file from a TFTP server.
*
* Returns:	see TftpWait()
****************************************************************************/
extern "C" UINT32 TftpDownload(UINT32 HostIP, UINT8 *Filename, UINT32 (*Vector)(UINT8 *, UINT32))
{
	Tftp.GetFile((IPADDR)HostIP, Filename, Vector);
	return TftpWait();
}

/****************************************************************************
* Function:	TftpUpload(UINT32 HostIP, UINT8 *Filename);
*
* Purpose:	Upload a file to a TFTP server.
*
* Returns:	see TftpWait()
****************************************************************************/
extern "C" UINT32 TftpUpload(UINT32 HostIP, UINT8 *Filename)
{
	Tftp.PutFile((IPADDR)HostIP, Filename);
	return TftpWait();
}

/****************************************************************************
* Function:	putfile(char *pIn, void *pOut)
*
* Purpose:	Command to copy a file to a TFTP server
*
* Returns:	void
****************************************************************************/
extern "C" void putfile(char *pIn, void *pOut)
{
	UINT32 Status;
	char *Name;

	if (DeconstructFilename(pIn, TRUE))
		return;

	Status = TftpUpload(HostIP, (unsigned char *)pFilespec);
	if (!Status)
		Result(0, "1 file copied");
}

Command(CMD_PUT, "put", putfile, 0, "", "copy a file to a TFTP host");
Arg(CHAR64ARG, "Filename", "");
CommandEnd();

/****************************************************************************
* Function:	copyfile(in_copyfile *pIn, void *pOut);
*
* Purpose:	Command to copy a file from a TFTP server with optional rename
*
* Returns:	void
****************************************************************************/
typedef struct
{
	char Filespec[64];
	char Newname[64];
	UINT32 InitialIP;
} in_copyfile;

extern "C" void copyfile(in_copyfile *pIn, void *pOut)
{
	UINT32 Status;
	char *Name;
	UINT32 InitialIP;

	if (DeconstructFilename(pIn->Filespec, TRUE))
		return;

	if (strlen(pIn->Newname))
		Name = pIn->Newname;
	else
		Name = pFilename;

	if (pIn->InitialIP)
		InitialIP = pIn->InitialIP;
	else
		InitialIP = 0xFFFFFFFF;

	if (!(gpHandle = (FFSHANDLE *)CreateFile(Name, (0xFFFFFFFF & ~ATTRIBTYPE) /*| ExtensionType */,
			InitialIP, 0xFFFFFFFF)))
	{
		FfsResult(pFilename);
		return;
	}

	BeginDownload(0);
	Status = TftpDownload(HostIP, (unsigned char *)pFilespec, CopyChunk);
	ProcessMode = 0;
	CloseFile(gpHandle);
	gpHandle = 0;
	if (!Status)
		Result(0, "1 file copied");
	else
		DeleteFile(Name);
}

Command(CMD_COPY, "copy", copyfile, 0, "", "copy a file from a TFTP host");
Arg(CHAR64ARG, "Filename", "");
OptArg(CHAR64ARG, "New name", "", "");
Switch("pc", UINT32ARG, "Initial Program Counter", "");
CommandEnd();

