/***************************************************************************
*
* Module:		i8255x.c
*
* Description:  Ethernet driver for the Intel 8255x chip
*
* Project:		HSI VisionWare -- vWare
*
* Functions:
*
* Notes:
*	This is an interrupt driven driver. The read function is not
*	implemented for poll mode.
*
*	This driver must provide a declaration of Eedriver of type
*	tEEDRIVER.  Most drivers will want to "superclass" this
*	structure by defining their own structure whose first field
*	is of type tEEDRIVER.  Any driver-specific initialization data
*	can then be added to the new structure after the tEEDRIVER field.
*	An instance of this new structure must then be named Eedriver.
*	A pointer to this structure will be passed to the Open function.
*
*	This driver must manage its own receive buffers.  VisionWARE
*	will allocate and manage the transmit buffers in a way that
*	obviates the need to allocate a full-sized transmit buffer
*	for each transmit buffer descriptor.  All this driver need do
*	is call EedriverAssignBuffer whenever a buffer is assigned to
*	a transmit BD, and call EedriverFreeBuffer whenever a transmit
*	BD is about to be reused.  Also, this driver must provide an
*	IsFreeBuffer function which takes a pointer to a BD and returns
*	true if the BD is free.
*
* History:		See Visual SourceSafe History
*
* Disclaimer:	Copyright 1999-2000 Wind River Systems, Inc. All Rights Reserved
*
****************************************************************************/
//  Includes
#include "vm.h"
#include "lib.h"
#include "eedriver.h"
#include "pci.h"
#include "i8255x.h"
#ifdef PCI_AUTO_ON
#include "pciconfiglib.h"
#include "pciautoconfig.h"
#include "board.h"
#endif

//  Local Defines
#ifndef ALIGN
  #define ALIGN(a,x) ( ((unsigned)(a) + (unsigned)(x) - 1) & ~((unsigned)(x) - 1) )
#endif

#define SCG_DEBUG	1	/* Debugging 1:On 0:Off  */
#define BSP_LAN1_FLAGS	0	// KLUDGE

#if SCG_DEBUG
  #define DBG_PRINT(x)  Print x
#else
  #define DBG_PRINT(x)
#endif

#define CARDS_NUM	1
#define BOARD_FLAGS	1
#define MIB		1

#if 1 /* XXX */
#define DelayNmsec(x) { volatile int i,j,k; for(i=0;i<1000;++i) for(j=0;j<(x);++j)k=0; }
#else
#define DelayNmsec(x) { volatile int i,j,k; for(i=0;i<10000;++i) for(j=0;j<(x);++j)k=0; }
#endif
#define DelayUsecs(x) DelayNmsec(x)

#define MIB_INC(x)   (x)++
#define MIB_INC_X(x, val)   x  +=  val

#define wait_scb_cmd_zero(x) while(x->cr_uc & 0xf0)

#ifndef LAN_TCB
  #define LAN_TCB 16
#endif

#ifndef LAN_PKB
  #define LAN_PKB 16
#endif

//  Local Declarations
static ULONG		intel_cfg_addr[CARDS_NUM];
static const ULONG	min_rfds = MIN_RFDS;
static ULONG		if_num[CARDS_NUM], ip_addr[CARDS_NUM];
static int			mib_outdiscards[CARDS_NUM];

//  Forward References
static ULONG	cu_idle_processing(LAN_INTF *);
static ULONG	start_ru(LAN_INTF *);
static void		remove_filled_frames(LAN_INTF *);
static void		eat_time(volatile ULONG count);
static void		add_dynamic_rfd(LAN_INTF *, PKB *);
static PKB		*get_pkb(LAN_INTF *);
static void		restart_ru(LAN_INTF *);
static void		ru_start(LAN_INTF *);
static int		merge_cbl(LAN_INTF *);
static ULONG	add_dynamic_cb(LAN_INTF *, TXB *);
static void		data_init(LAN_INTF *);
static void		pkb_init(LAN_INTF *);
static ULONG	recv_init(LAN_INTF *, ULONG count);
static ULONG	diagnose(LAN_INTF *);
static ULONG	configure(LAN_INTF *);
static ULONG	set_ind_addr(LAN_INTF *);
static void		ni_isr(LAN_INTF *);
static void		ack_cmd(LAN_INTF *);
void 			ni_pna_init(ULONG if_num);
static int		EthernetInit(LAN_INTF *);
static void		InitBD(LAN_INTF *);
static HANDLE	EthOpen(LAN_INTF *);
static DSTATUS	EthClose(LAN_INTF *);
static DSTATUS	EthWrite(LAN_INTF *, UINT8 *, UINT32, UINT8 *, UINT32, UINT8 *, UINT32);
static DSTATUS	EthIoctl(LAN_INTF *, UINT32 , void *);
static int		EthIsFreeBuffer(LAN_INTF *, void *);
static int		EthIsr(LAN_INTF *);

//  Global Declarations
LAN_INTF Eedriver = {
  {
    (EEDRIVEROpen)		EthOpen,
    (EEDRIVERClose)		EthClose,
    (EEDRIVERWrite)		EthWrite,
    (EEDRIVERIoctl)		EthIoctl,
    (EEDRIVERIsFreeBuffer)	EthIsFreeBuffer
   },
};

static int sromSizeBits;
void getSromSize(SCB_CRS *ioaddr)
{
	int i;
	USHORT data,readdata;

	ioaddr->eeprom_ctrl = SSWAP(EECS_BIT);

	for (i=2;i>=0;i--)
	{
		data = (6 & (1 << i)) ? EEDI_BIT : 0;
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT | data);
		DelayNMsec(1);
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT | EESK_BIT | data);
		DelayNMsec(1);
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT | data);
		DelayNMsec(1);
	}

	ioaddr->eeprom_ctrl = SSWAP(EECS_BIT);
	DelayNMsec(1);
	readdata = ioaddr->eeprom_ctrl;

	for (i=1;i<=12;i++)
	{
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT);
		DelayNMsec(1);
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT | EESK_BIT);
		DelayNMsec(1);
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT);
		DelayNMsec(1);
		readdata = SSWAP(ioaddr->eeprom_ctrl) & EEDO_BIT;
		if (readdata == 0) break;
	}
	sromSizeBits = i;

	ioaddr->eeprom_ctrl = SSWAP(EECS_BIT);
	DelayNMsec(1);
	readdata = ioaddr->eeprom_ctrl;

	for (i=15;i>=0;i--)
	{
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT | EESK_BIT);
		DelayNMsec(1);
		readdata = ioaddr->eeprom_ctrl;
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT);
		DelayNMsec(1);
	}
	ioaddr->eeprom_ctrl = 0;
	DelayNMsec(1);
}

USHORT srom_read_w(SCB_CRS *ioaddr, int sromaddr)
{
	int i;
	USHORT cmd,data,readdata;

	if (!sromSizeBits)
		getSromSize(ioaddr);

	cmd = (6 << sromSizeBits) | sromaddr;

	ioaddr->eeprom_ctrl = SSWAP(EECS_BIT);
	DelayNMsec(1);
		
	for (i=sromSizeBits+2;i>=0;i--)
	{
		data = (cmd & (1 << i)) ? EEDI_BIT : 0;
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT | data);
		DelayNMsec(1);
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT | EESK_BIT | data);
		DelayNMsec(1);
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT | data);
		DelayNMsec(1);
	}
	ioaddr->eeprom_ctrl = SSWAP(EECS_BIT);
	DelayNMsec(1);

	for (i=15;i>=0;i--)
	{
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT | EESK_BIT);
		DelayNMsec(1);
		readdata = (readdata << 1) | ((SSWAP(ioaddr->eeprom_ctrl) & EEDO_BIT) ? 1 : 0);
		ioaddr->eeprom_ctrl = SSWAP(EECS_BIT);
		DelayNMsec(1);
	}
	ioaddr->eeprom_ctrl = 0;
	DelayNMsec(1);

	return readdata;
}
/***************************************************************************
* Function:	EthClose(HANDLE pHandle)
*
* Purpose:	Close and deinit FCC channel
*
* Returns:	0 on success, non-0 on failure
****************************************************************************/
static DSTATUS EthClose(LAN_INTF *p)
{
VULONG		ba;

	EnterCriticalSection();
	DisableIRQ(p->esd8xxISR);
	RemoveHandler((HANDLE)p->esd8xxISR);
	p->IOaddr->port = LSWAP(2);
    DelayNmsec(100); 
    pciConfigInLong(p->p_dev.bus,p->p_dev.device,p->p_dev.function, PCI_CFG_COMMAND,(UINT32*)&ba);
    ba &= ~4;
	pciConfigOutLong(p->p_dev.bus,p->p_dev.device,p->p_dev.function,PCI_CFG_COMMAND,ba);
	DelayNmsec(50); 
	p->IOaddr->port = 0;	/* issue reset */
    DelayNmsec(500); 
	ExitCriticalSection();
	return STATUS_OK;
}

/***************************************************************************
* Function:	EthIoctl(DecCtrl *decp,UINT32 nFuncCode, void *pArgs)
*
* Purpose:	Set IO characteristics of the FCC channel
*
* Returns:	DSTATUS
****************************************************************************/
static DSTATUS EthIoctl(LAN_INTF *p, UINT32 nFuncCode, void *pArgs)
{
    switch(nFuncCode){
      case TURN_ON_LOOPBACK:
      case TURN_OFF_LOOPBACK:
      default:
	return NOT_IMPLEMENTED;
    }
    return STATUS_OK;
}

/***************************************************************************
* Function:	EthIsFreeBuffer(HANDLE pHandle, BDESC *pBD)
*
* Purpose:	Return non-zero if the indicated buffer descriptor is free.
*		VisionWARE calls this function to determine whether or not the
*		buffer that has been associated with this buffer descriptor
*		(via a call to EedriverAssignBuffer) is available for reuse.
*
* Returns:	0 if buffer is still in use, non-zero otherwise
***************************************************************************/
static int EthIsFreeBuffer(LAN_INTF *p, void *pBD)
{
    return 1;
}

/****************************************************************************
* Function:	EthOpen()
*
* Purpose:	Open and init the LAN channel
*
* Returns:	HANDLE ptr or 0 if error
****************************************************************************/
static HANDLE EthOpen(LAN_INTF *intfp)
{
int	i;
char	*cp;

	// Assumes PCI Auto Config
    cp = (char *)&intfp->demarc;
    memset(cp,0,sizeof(*intfp) - sizeof(intfp->func_ptrs));
    intfp->PciHandle = (UINT32)-1;
    return EthernetInit(intfp) ? (HANDLE)intfp : 0;

}

#define PRO100B_PCI_VENDOR_ID   0x8086  /* PCI vendor ID */
#define PRO100B_PCI_DEVICE_ID   0x1030  /* PCI device ID */
#define PRO100S_PCI_VENDOR_ID   0x8086  /* PCI vendor ID */
#define PRO100S_PCI_DEVICE_ID	0x1229  /* PCI device ID */
#define PRO100S_PCI_VENDOR_ID   0x8086  /* PCI vendor ID */
#define BRH_PCI_DEVICE_ID		0x1209  /* PCI device ID */
#define DEERISLAND_PCI_DEVICE_ID 0x1209
#define DEERISLAND_PCI_VENDOR_ID 0x8086

static int EthernetInit(LAN_INTF *p)
{
VULONG		ba;
VULONG		i,irq;
void		*irqp;
char		*uctmp,*MemPtr;

#ifndef PCI_AUTO_ON
	return 0;
#endif
#if 0
    if((pciFindDevice(PRO100B_PCI_VENDOR_ID,PRO100B_PCI_DEVICE_ID,0,
		&p->p_dev.bus, &p->p_dev.device, &p->p_dev.function) != OK)
    && (pciFindDevice(PRO100S_PCI_VENDOR_ID,PRO100S_PCI_DEVICE_ID,0,
		&p->p_dev.bus, &p->p_dev.device, &p->p_dev.function) != OK) 
	   && (pciFindDevice(PRO100S_PCI_VENDOR_ID,BRH_PCI_DEVICE_ID,0,
		   &p->p_dev.bus, &p->p_dev.device, &p->p_dev.function) != OK) ){
	return 0;
    }
#endif
    if (IS_EGRESS_NPU)
       {
       if (pciFindDevice(DEERISLAND_PCI_VENDOR_ID,
                         DEERISLAND_PCI_DEVICE_ID,
                         0,&p->p_dev.bus, &p->p_dev.device,
                         &p->p_dev.function) != OK)
          return 0;
       }
    else
       {
       if (pciFindDevice(DEERISLAND_PCI_VENDOR_ID,
                        DEERISLAND_PCI_DEVICE_ID,
                        1,&p->p_dev.bus, &p->p_dev.device,
                        &p->p_dev.function) != OK)
          return 0;
       }
                                        
    pciConfigInLong(p->p_dev.bus,p->p_dev.device,p->p_dev.function,
			PCI_CFG_COMMAND,(UINT32*)&ba);
    if(!(ba & 4)){	/* Check the CMD reg to see if device is MASTER enabled.   */
	return 0;	/* If not, autoconfig isn't working -- fix it */
    }
#ifndef BOARD_USE_PCI_MEMORY_SPACE
    pciConfigInLong(p->p_dev.bus,p->p_dev.device,p->p_dev.function,
		PCI_CFG_BASE_ADDRESS_1,(UINT32*)&ba);
    p->IOaddr = (void*)BRD_PCIIO_PCI2CPU(ba & ~1,0);
#else
    pciConfigInLong(p->p_dev.bus,p->p_dev.device,p->p_dev.function,
		PCI_CFG_BASE_ADDRESS_0,(UINT32*)&ba);
    p->IOaddr = (void*)(ba & ~0xf);
#endif

    p->num_lan_tcb = LAN_TCB;
    p->num_lan_pkb = LAN_PKB;

    MemPtr = malloc((p->num_lan_tcb+1) * sizeof(TCB));
    MemPtr = (char*)ALIGN(MemPtr,32);
    p->ni_tcb = (TXB*)MemPtr;

    MemPtr = malloc((p->num_lan_pkb+1) * PKB_SIZE);
    MemPtr = (char*)ALIGN(MemPtr,32);
    p->PacketsPtr = (PKB*)MemPtr;
    memset(p->PacketsPtr,0,p->num_lan_pkb * PKB_SIZE);

#if 0
    for(uctmp = (void*)&MyMAC,i = 0; i < 6; ++i){
	p->if_enet_addr.byte[i] = *uctmp++;
	p->if_broad_addr.byte[i] = 0xff;
    }
#else
    for (i = 0; i < 6; i++) {		/* Get the Ethernet Address. */
	USHORT	stmp;

	stmp = srom_read_w(p->IOaddr, (ULONG) (i >> 1));
	p->if_enet_addr.byte[i] = (char) stmp;
	p->if_broad_addr.byte[i] = 0xff;
	i++;
	p->if_enet_addr.byte[i] = (char) (stmp >> 8);
	p->if_broad_addr.byte[i] = 0xff;
    }
    memcpy((void*)&MyMAC,(void*)&p->if_enet_addr, 6);
#endif

    data_init(p);
    pkb_init(p);
    p->IOaddr->port = LSWAP(2);	/* issue selective reset */
    DelayNmsec(1);
    p->IOaddr->port = 0;	/* issue reset */
    DelayNmsec(1);
    if(!diagnose(p)){
	return 0;
    }

    if(!configure(p) || !set_ind_addr(p)){
	ack_cmd(p);
	return 0;
    }

    if(!recv_init(p, 5 /* p->num_lan_pkb >> 1 */ )){
	return 0;
    }

#if (defined(PCI_AUTO_ON) && defined(PCI_MULTIPLE_IRQS))
    pciConfigInLong(p->p_dev.bus,p->p_dev.device,p->p_dev.function,0x3c,(UINT32*)&ba);
    irq = (ETHERNETIRQ & ~0xff) | (int)(ba & 0x000000ff);
#else
    irq = ETHERNETIRQ;
#endif
	p->esd8xxISR  = InstallHandler(irq,(HANDLER)EthIsr,(void*)p);
	EnableIRQ(p->esd8xxISR);
    ru_start(p);
    p->good_open = 1;
    return 1;
}

/***************************************************************************
* Function:	EthWrite(volatile PPC8260_FCC *,
*				UINT8 *buffer1, unsigned int Length1,
*				UINT8 *buffer2, unsigned int Length2,
*				UINT8 *buffer3, unsigned int Length3);
*
* Purpose:	Send a packet composed of one, two or three buffers.  Set up all
*			buffer descriptors before initiating the send.
*
* Returns:	DSTATUS
****************************************************************************/
static DSTATUS EthWrite(LAN_INTF *p, UINT8 *pB1, UINT32 Len1,
		UINT8 *pB2, UINT32 Len2, UINT8 *pB3, UINT32 Len3)
{
TXB	*txp;
TBD	*tbdp;
UCHAR	tbd_count = 0;
int	i, len;

    if(p->good_open != 1){
	printf("EthWrite: device open must have failed!\n");
	return NOT_IMPLEMENTED;
    }
    EnterCriticalSection();
    if((txp = p->tcb_free)){
	p->tcb_free = (TXB *)txp->tcb.next;
	--p->tcb_count;
    }else{
	ExitCriticalSection();
	return NOT_IMPLEMENTED;
    }
    memset((void*)&txp->tcb,0,sizeof(TCB));
    memset((void*)txp->tbd_array,0,LAN_TBD*sizeof(TBD));
    txp->tcb.status = 0;
    txp->tcb.next = NULL;
    txp->tcb.cmd = SSWAP(CMD_ACT_TRANSMIT | FLEXIBLE | CMD_EL);
    txp->tcb.count = 0 /* 14 */ ;
    txp->tcb.bd_addr = (void*)CPU2DEV_ADDR(txp->tbd_array);
    txp->tcb.tx_thresh = 1;
    tbdp = &txp->tbd_array[0];
    if(Len1){
	EedriverAssignBuffer(pB1,(UINT32)&tbdp->baddr);
	txp->tbd_array[tbd_count].baddr = (void*)CPU2DEV_ADDR(pB1);
	txp->tbd_array[tbd_count++].size = SSWAP(Len1);
	++tbdp;
	if(Len2){
	    EedriverAssignBuffer(pB2,(UINT32)&tbdp->baddr);
	    txp->tbd_array[tbd_count].baddr = (void*)CPU2DEV_ADDR(pB2);
	    txp->tbd_array[tbd_count++].size = SSWAP(Len2);
	    ++tbdp;
	    if(Len3){
		EedriverAssignBuffer(pB3,(UINT32)&tbdp->baddr);
		txp->tbd_array[tbd_count].baddr = (void*)CPU2DEV_ADDR(pB3);
		txp->tbd_array[tbd_count++].size = SSWAP(Len3);
	    }
	}
    }else{
	ExitCriticalSection();
	return NOT_IMPLEMENTED;
    }
    txp->tcb.tbd_num = tbd_count;
    add_dynamic_cb(p, txp);
    ExitCriticalSection();
    return STATUS_OK;
}

/****************************************************************************
* Function:	ack_cmd(LAN_INTF *p);
*
* Note:		ack_cmd() should be used till the isr is not enabled.
*			ack_cmd() works as an interrupt handler.
*
* Purpose:	Once CPU issues CA to 82559, 559 executes following sequence:
*			1. Accept the START_CU command by clearing the scb.cuc
*			2. Execute the scb.general_addr.
*			3. Update the scb.status.
*			4. Generate an interrupt.
*
*			CPU should then:
*			1. Check that scb.cuc & scb.ruc are zero
*			2. Set scb.ack
*
* Returns:	void
****************************************************************************/
static void ack_cmd(LAN_INTF *p)
{
    wait_scb_cmd_zero(p->IOaddr);	/* scb.cmd = 0 is the last step  */
    p->IOaddr->stat_ack = p->IOaddr->stat_ack;
}

/****************************************************************************
* Function:	diagnose(LAN_INTF *p)
*
* Purpose:	send a "diagnose" command
*
* Returns:	TRUE if OK, FALSE if command fails
****************************************************************************/
static ULONG diagnose(LAN_INTF *p)
{
	volatile DIAG	*diagp;

    diagp = &p->diag;
    diagp->status = SSWAP(0);
    diagp->cmd = SSWAP(CMD_EL|CMD_S|CMD_I|CMD_ACT_DIAGNOSE);
    diagp->lnk = NULL;

    p->IOaddr->gen_ptr = CPU2DEV_ADDR(diagp);
    p->IOaddr->cr_uc = (RUC_NOP | CUC_START);

    eat_time(DELAY_CONSTANT);
    ack_cmd(p);

    while(!(diagp->status & SSWAP(STATUS_C)))
	;
    return !(diagp->status & SSWAP(STATUS_F));
}

/****************************************************************************
* Function:	configure(LAN_INTF *p)
*
* Purpose:	send a "configure" command
*
* Returns:	TRUE if OK, FALSE if command fails
****************************************************************************/
static ULONG configure(LAN_INTF *p)
{
CFG	*cfgp;

    cfgp = &p->config;
    cfgp->cmd_blk.status = SSWAP(0);
    cfgp->cmd_blk.cmd = SSWAP(CMD_EL|CMD_ACT_CONFIGURE);
    cfgp->cmd_blk.lnk = 0;

/**********************************************************/
/* For the values of each of the configure bytes, see:    */
/* Intel 82559 Software Developers Manual (6-49)          */
/**********************************************************/

    cfgp->byte_count = 0x16;		/* byte count */
    cfgp->b1 = 0x08;		/* 8 rx fifo - 8 */
				/* tx fif0 - 0 */
    cfgp->b2 = 0x0;		/* adaptive IFS */
    cfgp->b3 = 0x0;		/* MWI enable */
				/* Type enable */
				/* Read Align enable */
				/* Terminate on CL */
    cfgp->b4 = 0x0;		/* RX DMA max byte count */
    cfgp->b5 = 0x0;		/* TX DMA max byte count */
				/* TX DMA max byte count enable */
    cfgp->b6 = 0x3a;		/* 3a Don't save bad frames */
				/* Discard overruns */
				/* Standard Stat counters */
				/* CU Idle Interrupt enabled */
				/* Non-Direct DMA Mode */
    cfgp->b7 = 0x3;		/* 3 1 underrun retry */
				/* discard short frames */
    cfgp->b8 = 0x1;		/* 1 CSMA disable */
				/* bit 0 set to 1 */
    cfgp->b9 = 0x0;		/* unused */

    cfgp->b10 = 0x2E;		/* Normal mode */
				/* preamble length 7 bytes */
				/* no SA insertion */
    cfgp->b11 = 0x0;		/* unused */
    cfgp->b12 = 0x60;		/* interframe spacing 6 * 16 */
    cfgp->b13 = 0x0;		/* ARP filter low */
    cfgp->b14 = 0x0;		/* ARP filter high */
    cfgp->b15 = 0xc8;		/* Disable test mode */
				/* consider U/L bit */
				/* wait after WIN disabled */
				/* Don't disable broadcast */
				/* Not promiscuous */
    cfgp->b16 = 0x0;		/* FC Delay lsb */
    cfgp->b17 = 0x40;		/* FC Delay msb */
    cfgp->b18 = 0xf2;		/* Flow control thresh disabled */
				/* long rx not OK */
				/* RCV CRC Transfer disabled */
				/* padding enabled */
				/* ensable stripping */
    cfgp->b19 = 0x0;
    cfgp->b20 = 0x3f;		/* NO multi IA */
				/* FC location bit 31 */
    cfgp->b21 = 0x5;		/* disbale Multicast All */

    p->IOaddr->gen_ptr = CPU2DEV_ADDR(cfgp);
    p->IOaddr->cr_uc = (RUC_NOP | CUC_START);
    eat_time(DELAY_CONSTANT);
    ack_cmd(p);
    while(!(cfgp->cmd_blk.status & SSWAP(STATUS_C)))
		;
    return (cfgp->cmd_blk.status & SSWAP(STATUS_OKAY)) == SSWAP(STATUS_OKAY);
}

/****************************************************************************
* Function:	set_ind_addr(LAN_INTF *p)
*
* Purpose:	send a "set individual address" command
*
* Returns:	TRUE if OK, FALSE if command fails
****************************************************************************/
static ULONG set_ind_addr(LAN_INTF *p)
{
IAS	*iap;

    iap = &p->ia_struct;
    iap->cmd_blk.status = 0;
    iap->cmd_blk.cmd = SSWAP(CMD_EL|CMD_ACT_IA_SETUP);
    iap->cmd_blk.lnk = 0;

    iap->byte_1 = p->if_enet_addr.byte[0];	/* Individual Addr byte 1 */
    iap->byte_2 = p->if_enet_addr.byte[1];	/* Individual Addr byte 1 */
    iap->byte_3 = p->if_enet_addr.byte[2];	/* Individual Addr byte 1 */
    iap->byte_4 = p->if_enet_addr.byte[3];	/* Individual Addr byte 1 */
    iap->byte_5 = p->if_enet_addr.byte[4];	/* Individual Addr byte 1 */
    iap->byte_6 = p->if_enet_addr.byte[5];	/* Individual Addr byte 1 */

    p->IOaddr->gen_ptr = CPU2DEV_ADDR(iap);
    p->IOaddr->cr_uc = (RUC_NOP | CUC_START);
    eat_time(DELAY_CONSTANT);

    ack_cmd(p);
    while(!(iap->cmd_blk.status & SSWAP(STATUS_C)))
	;
    return (iap->cmd_blk.status & SSWAP(STATUS_OKAY)) == SSWAP(STATUS_OKAY);
}

/****************************************************************************
* Function:	data_init(LAN_INTF *p)
*
* Purpose:	data_init: Initialize the data area of the NI driver
*
* Returns:	TRUE if OK, FALSE if command fails
****************************************************************************/
static void data_init(LAN_INTF *p)
{
    p->begin_cbl = p->end_cbl = NULL;
    p->last_unk_cb = p->beg_saved_blks = p->end_saved_blks = NULL;

/*---------------------------------------------------------------------*/
/* Set Pointers to global buffer areas                                 */
/*                                                                     */
/* ***** NOTE *****                                                    */
/* These pointers MUST be at LEAST WORD Aligned !!!!!!                 */
/* DWORD is even better                                                */
/*---------------------------------------------------------------------*/
    p->pkb_free = NULL;
    p->pkb_count = 0;
    p->rfa_head = p->rfa_tail = p->last_unk_rfd = NULL;
    p->ru_stopped = FALSE;
    p->restarted_ru = 0;
}

static int EthIsr(LAN_INTF *p)
{
int	ret_val = 0;
UCHAR	int_stat;

    wait_scb_cmd_zero(p->IOaddr);
    int_stat = p->IOaddr->stat_ack;
    while(int_stat & (INT_FR | INT_RNR | INT_CX | INT_CNA)) {
	++ret_val;
	p->IOaddr->stat_ack = int_stat;
	wait_scb_cmd_zero(p->IOaddr);
	if(int_stat & INT_FR){
	    ++p->num_fr_events;
	    remove_filled_frames(p);
	}
	if((int_stat & INT_RNR) && start_ru(p)){
	    ++p->num_rnr_events;
	    p->IOaddr->cr_uc = RUC_START;
	    wait_scb_cmd_zero(p->IOaddr);
	}
	if((int_stat & INT_CNA) && cu_idle_processing(p)){
	    ++p->num_cna_events;
	    p->IOaddr->cr_uc = CUC_START;
	    wait_scb_cmd_zero(p->IOaddr);
	}
	int_stat = p->IOaddr->stat_ack;
    }
#ifdef TGT_PPMC8260
/* KLUDGE */
    *((volatile ULONG*)0x30000410) = *((volatile ULONG*)0x30000410);
#endif

    return (ret_val != 0);
}

static free_pkb(LAN_INTF *p, PKB *pkb)
{
    if(p->pkb_count){
	pkb->rfd.next = p->pkb_free;
	p->pkb_free = pkb;
    }else{
	pkb->rfd.next = NULL;
	p->pkb_free = pkb;
    }
    ++p->pkb_count;
}

/***********************************************************************/
/*    pkb_init: Initialize the driver's packet pool                    */
/*                                                                     */
/*     NOTE(S): The packet pool is a singly-linked list                */
/*                                                                     */
/***********************************************************************/
static void pkb_init(LAN_INTF *p)
{
ULONG	i;
PKB	*ptr;
UCHAR	*nextp;
TXB	*tcbp;

    EnterCriticalSection();
    ptr = p->PacketsPtr;
    p->pkb_count = 0;
    for(i = 0;i < p->num_lan_pkb;++i){
	free_pkb(p,ptr);
	nextp = (UCHAR *)((ULONG)ptr + PKB_SIZE);
	ptr = (PKB*)nextp;
    }
    p->tcb_free = 0;
    p->tcb_count = 0;
    for(i = 0;i < p->num_lan_tcb;++i){
	tcbp = &p->ni_tcb[i];
	tcbp->tcb.next = p->tcb_free;
	p->tcb_free = tcbp;
	++p->tcb_count;
    }
    ExitCriticalSection();
}

/***********************************************************************/
/*     get_pkb: Internal routine to fetch a packet from the NI's pool  */
/*                                                                     */
/*     RETURNS: Address of packet, or NULL if none available           */
/*     NOTE(S): Must be called with interrupts disabled.               */
/*                                                                     */
/***********************************************************************/
static PKB *get_pkb(LAN_INTF *p)
{
PKB	*ptr = NULL;

    if(p->pkb_count > 0){
	ptr = p->pkb_free;
	p->pkb_free = ptr->rfd.next;
	--p->pkb_count;
	if(p->pkb_count == 0){
	    p->pkb_free = NULL;
	}
    }
    return ptr;
}

/***********************************************************************/
/* add_dynamic_cb: Add a command block dynamically in the list         */
/*                                                                     */
/*      INPUTS: blk = command block to be added                        */
/*                                                                     */
/*     RETURNS: Always returns 0                                       */
/*     NOTE(S): this routine gets called from ni_send.                 */
/*                                                                     */
/***********************************************************************/
/* Following comment explains the strategy of dynamic addition.        */
/*                                                                     */
/* A dynamic linking process is defined as: Addition of elements in a  */
/* list when another process may or may not be working on the list.    */
/*                                                                     */
/* In this case, CBL is being executed by 82559, and command blocks    */
/* are added by CPU.  Command block bits  cb.c indicate the            */
/* status of that command block.                                       */
/*                                                                     */
/* Any command block goes through following sequence:                  */
/*                                                                     */
/* 1. C = 0  set initially by the CPU.                                 */
/* 3. C = 1  set when CU finishes executing.                           */
/*                                                                     */
/* So CPU can append another block, if last block of the list is in    */
/* state 1. The problem occurs when last block changes from state 1 to */
/* 2 or 3, after addition of new block. Depending on whether 82559 has */
/* seen the EL bit in last block, it will not or it will execute the   */
/* new block. Hence status of this new block is unknown till next CNA  */
/* interrupt from 82559.                                               */
/*                                                                     */
/* So the algorithm for dynamic linking is:                            */
/*                                                                     */
/* 1. check the bits C                                                 */
/* 2. if completed                                                     */
/*     do not link. save for next time                                 */
/* 3. if not (completed)                                               */
/*     linked successfully                                             */
/* 4. if completed                                                     */
/*     link doubtful                                                   */
/*                                                                     */
/* begin_cbl, and end_cbl indicate the start and end of CBL according  */
/* to CPU. At a time there can be only one last_unk_cb.                */
/*                                                                     */
/***********************************************************************/
static ULONG add_dynamic_cb(LAN_INTF *p, TXB *txp)
{
ULONG imask;

    EnterCriticalSection();
    txp->tcb.next = 0;
    if(p->begin_cbl){

#if 0
	if(p->end_cbl->tcb.status & STATUS_C){
#endif
	    if(p->num_saved_blks){
		p->end_saved_blks->tcb.cmd &= ~SSWAP(CMD_EL);
		p->end_saved_blks->tcb.next = (TXB*)CPU2DEV_ADDR(txp);
		p->end_saved_blks = txp;
		++p->num_saved_blks;
	    }else{
		p->beg_saved_blks = txp;
		p->end_saved_blks = txp;
		p->num_saved_blks = 1;
	    }
	    ExitCriticalSection();
	    return 0;

#if 0
	}
	p->end_cbl->tcb.next = txp;
	p->end_cbl->tcb.cmd &= ~CMD_EL;
	if(p->end_cbl->tcb.status & STATUS_C){
	    p->last_unk_cb = txp;
	}
	p->end_cbl = txp;
#endif

    }else{
	p->begin_cbl = txp;
	p->end_cbl = txp;
	wait_scb_cmd_zero(p->IOaddr);
	p->IOaddr->gen_ptr = CPU2DEV_ADDR(p->begin_cbl);
	p->IOaddr->cr_uc = CUC_START;
    }
    ExitCriticalSection();
    return 0;
}

/***********************************************************************/
/* cu_idle_processing:                                                 */
/*                                                                     */
/*     RETURNS: TRUE if new commands added to the CU, FALSE otherwise  */
/*     NOTE(S): cu_idle_processing removes cbs which are done and adds */
/*              cbs which were saved for this INTR.  last_unk_cb may   */
/*              or may not have been executed.  So in case last_unk_cb */
/*              is not done, it's merged with the saved blks. A new    */
/*              CBL is formed. And CUC is prepared for START of CU.    */
/*                                                                     */
/***********************************************************************/
static ULONG cu_idle_processing(LAN_INTF * p)
{
TXB	*txp;
ULONG	i;

    EnterCriticalSection();

    ++p->num_cu_idles;

    if(p->last_unk_cb){
	if(!(p->last_unk_cb->tcb.status & SSWAP(STATUS_C))){
	    if(p->IOaddr->cr_us & CUS_ACTIVE){
		++p->num_cus_active_during_cu_idle; /* what the... ?? */
		ExitCriticalSection();
		return FALSE;
	    }
	}
    }
    txp = p->begin_cbl;
    while(txp && txp->tcb.status & SSWAP(STATUS_C)){
	++p->total_cmds_completed;
	for(i = 0 ; i < txp->tcb.tbd_num; ++i){
	    EedriverFreeBuffer((void*)&txp->tbd_array[i]);
	}
	if(txp == p->end_cbl){
	    if(txp->tcb.next != NULL || !(txp->tcb.cmd & SSWAP(CMD_EL)) ){
		++p->impossible_happened;
	    }
	    p->end_cbl = NULL;
	}
	if(txp == p->last_unk_cb){
	    ++p->unknown_cmd_completed;
	    p->last_unk_cb = NULL;
	}
	p->begin_cbl =  (void*)DEV2CPU_ADDR(txp->tcb.next);
	txp->tcb.next = p->tcb_free;
	p->tcb_free = txp;
	++p->tcb_count;
	txp = p->begin_cbl;
    }
    /* Merge list of pending command blocks. */
    if((i = merge_cbl(p))){
	p->IOaddr->gen_ptr = CPU2DEV_ADDR(p->begin_cbl);
    }
    ExitCriticalSection();
    return i;
}

/***********************************************************************/
/*   merge_cbl: Merge all possible remaining command blocks            */
/*                                                                     */
/*     RETURNS: Pointer to new command block list                      */
/*        NOTE: Must be called with interrupts disabled                */
/*                                                                     */
/***********************************************************************/
static int merge_cbl(LAN_INTF *p)
{
    if(!p->last_unk_cb && !p->num_saved_blks)
	return 0;

    p->begin_cbl = p->end_cbl = 0;

    if(p->last_unk_cb){
	p->last_unk_cb->tcb.cmd &= ~SSWAP(CMD_EL);
	p->begin_cbl = p->end_cbl = p->last_unk_cb;
	p->last_unk_cb = 0;
    }
    if(p->num_saved_blks){
	if(!p->begin_cbl){
	    p->begin_cbl = p->beg_saved_blks;
	}else{
	    p->end_cbl->tcb.next = (void*)CPU2DEV_ADDR(p->beg_saved_blks);
	}
    }
    p->end_cbl = p->end_saved_blks;
    p->beg_saved_blks = p->end_saved_blks = 0;
    p->num_saved_blks = 0;
    return 1;
}

/***********************************************************************/
/*   recv_init: Initialize the receiver frame area                     */
/*                                                                     */
/*      INPUTS: count = Number of RFDs (Receive Frame Descriptors)     */
/*                      needed in the RFA                              */
/*     RETURNS: TRUE if succeeds, FALSE if fails                       */
/*       NOTES: This initializes all structures and pointers related   */
/*              to the RU and builds a linked list in the RFA.         */
/*                                                                     */
/***********************************************************************/
static ULONG recv_init(LAN_INTF *p, ULONG count)
{
PKB	*pkb;
ULONG	imask, i;

    pkb = 0;
    EnterCriticalSection();
    p->rfa_head = p->rfa_tail = NULL;
    if(count > p->pkb_count){
	p->rfa_count = 0;
	ExitCriticalSection();
	return FALSE;
    }
    pkb = get_pkb(p);
    pkb->rfd.status	= SSWAP(0);
    pkb->rfd.cmd	= SSWAP(0x8000);	/* EL = 1 */
    pkb->rfd.next	= 0;
    pkb->rfd.rbdptr	= (void *)-1UL;
    pkb->rfd.count	= 0;
    pkb->rfd.size	= SSWAP(MAX_BYTES);
    p->rfa_count = 1;
    p->rfa_head = p->rfa_tail = pkb;
    for(i = 1;i < count;++i){
	pkb = get_pkb(p);
	pkb->rfd.status	= 0;
	pkb->rfd.cmd	= 0;
	pkb->rfd.rbdptr	= (void *)-1UL;
	pkb->rfd.count	= 0;
	pkb->rfd.size	= SSWAP(MAX_BYTES);
	pkb->rfd.next	= (void*)CPU2DEV_ADDR(p->rfa_head);
	p->rfa_head = pkb;
	++p->rfa_count;
    }
    ExitCriticalSection();
    return TRUE;
}

/***********************************************************************/
/* add_dynamic_rfd: Dynamically add a Receive Frame Descriptor to the  */
/*                  RFA                                                */
/*                                                                     */
/*      INPUTS: blk = pointer to RFD to be added to the RFA            */
/*        NOTE: For an explanation of dynamic addition refer to        */
/*              add_dynamic_cb().                                      */
/*              Must be called with interrupts disabled.               */
/*                                                                     */
/***********************************************************************/
static void add_dynamic_rfd(LAN_INTF *p, PKB *blk)
{
    blk->rfd.status	= 0;
    blk->rfd.cmd	= SSWAP(0x8000);
    blk->rfd.next	= NULL;
    blk->rfd.rbdptr	= (void *)-1UL;
    blk->rfd.count	= 0;
    blk->rfd.size	= SSWAP(MAX_BYTES);

    if(p->rfa_count){
	if(p->rfa_tail->rfd.status & SSWAP(0x8000)) {	/* C */
	    free_pkb(p,blk);
	    return;
	}
	p->rfa_tail->rfd.next = (void*)CPU2DEV_ADDR(blk);
	p->rfa_tail->rfd.cmd = 0;
	if(p->rfa_tail->rfd.status & SSWAP(0x8000))	/* 1->2 */
	    p->last_unk_rfd = blk;
	else
	    p->rfa_tail = blk;
	p->rfa_count++;
    } else {
	p->rfa_head = blk;
	p->rfa_tail = blk;
	p->rfa_count = 1;
    }
}

/***********************************************************************/
/* remove_filled_frames: Remove filled frames from RFA                 */
/*                                                                     */
/***********************************************************************/
static void remove_filled_frames(LAN_INTF *p)
{
ULONG	len;
PKB	*pkb, *pkb_a, *next_rfd;
//ULONG	return_pkb;

    EnterCriticalSection();
    for(pkb = p->rfa_head; pkb && (pkb->rfd.status & SSWAP(0x8000)); pkb = next_rfd){
	next_rfd = (void*)DEV2CPU_ADDR(pkb->rfd.next);
	p->rfa_head = next_rfd;
	--p->rfa_count;
	if(p->last_unk_rfd == pkb){
	    p->last_unk_rfd = NULL;
	}
	if(pkb == p->rfa_tail){
	    p->rfa_head = p->rfa_tail = NULL;
	    p->rfa_count = 0;
	}
//	return_pkb = 0;
	if(pkb->rfd.count & SSWAP(0x4000)){	/* If the frame is filled with data    */
	    len = SSWAP(pkb->rfd.count) & 0x3fff;
	    p->inoctets += len;
	    if(pkb->rfd.status & SSWAP(0x2)) ++p->innucastpkts;
	    else ++p->inucastpkts;
	    ++p->num_rx_passed_up;
	    ProcessPacket((UCHAR*)&pkb->da,len);
	}else{
//	    return_pkb = 1;
	}
//	if(return_pkb){
	    add_dynamic_rfd(p, pkb);
//	}else if((pkb_a = get_pkb(p)) != NULL){
//	    add_dynamic_rfd(p, pkb_a);
//	}
    }
    if(!p->rfa_count || !p->rfa_head) {
	p->rfa_tail = p->rfa_head;
	p->rfa_count = 0;
    }
    ExitCriticalSection();
}

static void append_rfa(LAN_INTF *p, PKB *pkb)
{
PKB	*prev_tail;

    pkb->rfd.status	= 0;
    pkb->rfd.cmd	= SSWAP(0x8000);
    pkb->rfd.next	= NULL;
    pkb->rfd.rbdptr	= (void *)-1UL;
    pkb->rfd.count	= 0;
    pkb->rfd.size	= SSWAP(MAX_BYTES);
    if(p->rfa_count == 0){
	p->rfa_head = p->rfa_tail = pkb;
    }else{
	prev_tail = p->rfa_tail;
	prev_tail->rfd.next = (void*)CPU2DEV_ADDR(pkb);
	prev_tail->rfd.cmd = 0;
	p->rfa_tail = pkb;
    }
    ++p->rfa_count;
}

/***********************************************************************/
/* start_ru: Start the RU, which is stopped due to No RFDs             */
/*                                                                     */
/*     RETURNS:                                                        */
/*              FALSE                                                  */
/*                                                                     */
/***********************************************************************/
static ULONG start_ru(LAN_INTF *p)
{
ULONG	needed, imask;
PKB	*pkb;

    EnterCriticalSection();
    p->ru_stopped = TRUE;
    if( (p->rfa_count < min_rfds) && (p->pkb_count > (min_rfds - p->rfa_count))) {
	needed = min_rfds - p->rfa_count;
	if(p->rfa_count == 0){
	    recv_init(p, min_rfds);
	}else{
	    for (; needed; --needed) {
		append_rfa(p,get_pkb(p));
	    }
	}
	restart_ru(p);	/* ru_stopped = FALSE */
	ExitCriticalSection();
	return FALSE;	/* CA needed after this */
    } else if(p->rfa_count >= min_rfds) {
	restart_ru(p);
	ExitCriticalSection();
	return FALSE;
    }
    ExitCriticalSection();
    return FALSE;
}

/***********************************************************************/
/*  restart_ru: Restart the Receiver Unit                              */
/*                                                                     */
/*     NOTE(S): Must be called at highest Interrupt Level              */
/*                                                                     */
/***********************************************************************/
static void restart_ru(LAN_INTF *p)
{
PKB	*pkb;

    EnterCriticalSection();
    p->restarted_ru++;
    wait_scb_cmd_zero(p->IOaddr);	/* this can be called from ret_pkb */
    if((pkb = p->last_unk_rfd)){
	p->last_unk_rfd = NULL;
	if(p->rfa_count == 0){
	    pkb->rfd.cmd = SSWAP(0x8000);
	    pkb->rfd.next = NULL;
	    p->rfa_head = p->rfa_tail = pkb;
	}else{
	    pkb->rfd.cmd = 0;
	    pkb->rfd.next = (void*)CPU2DEV_ADDR(p->rfa_head);
	    p->rfa_head = pkb;
	}
	p->rfa_count++;
    }
    p->IOaddr->gen_ptr = CPU2DEV_ADDR(p->rfa_head);
    p->IOaddr->cr_uc = (CUC_NOP | RUC_START);
    p->ru_stopped = FALSE;
    ExitCriticalSection();
}

/***********************************************************************/
/*    ru_start: Start the Receiver Unit                                */
/*                                                                     */
/*     NOTE(S): Should be called only after SysSetVector()             */
/*                                                                     */
/***********************************************************************/
static void ru_start(LAN_INTF *p)
{
    EnterCriticalSection();
    p->IOaddr->gen_ptr = CPU2DEV_ADDR(p->rfa_head);
    p->IOaddr->cr_uc = (CUC_NOP | RUC_START);
    wait_scb_cmd_zero(p->IOaddr);
    ExitCriticalSection();
}

/***********************************************************************/
/*    eat_time: Delay for a while                                      */
/*                                                                     */
/*      INPUTS: count = Number of times to execute spin loop           */
/*                                                                     */
/***********************************************************************/
static void eat_time(volatile ULONG count)
{
	while (count--);
}

