/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 BY Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/* ixd28100Autoneg.c - IXD2810 Auto-negotiation module

modification history
--------------------
01a,08oct02,eme  created from autoneg.c 1/1/2001 by Nancy Wang
*/

/*
DESCRIPTION
This module handles the auto-negotiation for the IXF1110 MAC
device.

*/
#include "vm.h"		   
#include "cmd.h"
#include "mac.h"
#include "IXP2800diag.h"
#include "ixd2810Autoneg.h"

/*******************************************************************************
 * doAutoneg - Starting poing of auto-negotiation.
*				Determine copper or fiber media and
*				call other functions correspondingly.
 *
 * RETURNS:  OK or ERROR if failure occurs.
 */
STATUS doAutoneg
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port /* port number. From 0 to 9. */
	) {

	/* sanity check */
	if (port >= N_PORTS)
		return ERROR;

	/* Initialize link/speed/duplex status */
	pDev->ports[port].link = DOWN;
	pDev->ports[port].duplex = HALF;
	pDev->ports[port].speed = SPEED_10;
	
	/* Fiber */
	fiberAutoNeg(pDev, port);	/* fiber auto-negotiation */
	/* Fiber mode only supports giga speed and full duplex */
	if (pDev->ports[port].link == UP) {	/* link up */
		if ((pDev->ports[port].speed == SPEED_1000) && 
		(pDev->ports[port].duplex == FULL)) {
			configMac(pDev, port);  /* Configure MAC */
			printMsg(pDev, port, 1);   /* Print message */
		}
		else {
			disablePort(pDev,port);	     /* Disable port */
			printMsg(pDev, port, 2);   /* Print message */
		}
	} /* end if */
	else {				/* link down */
		disablePort(pDev, port);
		printMsg(pDev, port, 3);
	}

	return OK;
}

/*******************************************************************************
 * configMac - 	Configure specified MAC port. 
 *				(link/speed/duplex)
 *
 * RETURNS: N/A
 */
void configMac
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port /* port number. From 0 to 9. */
	) {
	uint32_t data;
	
	/* sanity check */
	if (port >= N_PORTS)
		return;

	/* Set link up*/
	IXD2810_REG_GET(pDev->vbase + LINK_LED_ENABLE, data);
	data |= LINK_UP(port);
	IXD2810_REG_SET(pDev->vbase + LINK_LED_ENABLE, data);

	/* Enable port */
	IXD2810_REG_GET(pDev->vbase + PORT_ENABLE, data);
	data |= PORT_EN(port);
	IXD2810_REG_SET(pDev->vbase + PORT_ENABLE, data);
}

/*******************************************************************************
 * fiberAutoNeg - Fiber auto-negotiation routine.
 *
 * RETURNS: N/A
 */
void fiberAutoNeg
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port /* port number. From 0 to 9. */
	) 
	{
	uint32_t txData, rxData, divData;
	int delay;
	uint32_t vbase;

	/* get a copy of the device virtual address */
	vbase = pDev->vbase;

	/* Start auto-negotiation */
	IXD2810_REG_GET(vbase + BLOCK_REG_ADRS(port, DIV_CONFIG_WORD), divData);
	divData |= DC_Autoneg_Enable;
	IXD2810_REG_SET(vbase + BLOCK_REG_ADRS(port, DIV_CONFIG_WORD), divData);

	/* Poll Rx Config Word to determine when AN is complete or timeout */
	for(delay = 0; delay < AN_TIMEOUT; delay++) {
		DelayNUsec(10000); /* 10 msec delay */
		IXD2810_REG_GET(vbase + BLOCK_REG_ADRS(port, RX_CONFIG_WORD), rxData);
		if (rxData & RC_An_Complete)
			break;
	}

	if (delay < AN_TIMEOUT) {
		/* Speed is fixed at giga */
		pDev->ports[port].speed = SPEED_1000;

		/* Get port status from Rx Config Word */
		if (rxData & RC_Half_Duplex) 
			pDev->ports[port].duplex = HALF;	
		if (rxData & RC_Full_Duplex)
			pDev->ports[port].duplex = FULL;

		/* Link is up when autoneg is enabled and finished */
		pDev->ports[port].link = UP;
	}
	else { /* timeout has occurred */
		pDev->ports[port].link = DOWN;
	}
}

/*******************************************************************************
 * disablePort - Disable MAC port and set link down.
 *
 * RETURNS:  N/A
 */
void disablePort
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port /* port number. From 0 to 9. */
	) 
	{

	uint32_t data;

	/* sanity check */
	if (port >= N_PORTS)
		return;

	/* bits 9:0 for port enable. 0=disable */
	IXD2810_REG_GET(pDev->vbase + PORT_ENABLE, data);
	data &= (ENABLE_BOUNDARY - PORT_EN(port));
	IXD2810_REG_SET(pDev->vbase + PORT_ENABLE, data);

	/* bits 9:0 for link up. 0=down */
	IXD2810_REG_GET(pDev->vbase + LINK_LED_ENABLE, data);
	data &= (ENABLE_BOUNDARY - LINK_UP(port));
	IXD2810_REG_SET(pDev->vbase + LINK_LED_ENABLE, data);
}


/*******************************************************************************
 *	enablePort - Enable MAC port and set link up.
 *
 *	RETURNS: N/A
 */
void enablePort
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port /* port number. From 0 to 9. */
	) 
	{

	uint32_t data;

	/* sanity check */
	if (port >= N_PORTS)
		return;

	/* bits 9:0 for port enable. 0=disable */
	IXD2810_REG_GET(pDev->vbase + PORT_ENABLE, data);
	data |= PORT_EN(port);
	IXD2810_REG_SET(pDev->vbase + PORT_ENABLE, data);

	/* bits 9:0 for link up. 0=down */
	IXD2810_REG_GET(pDev->vbase + LINK_LED_ENABLE, data);
	data |= LINK_UP(port);
	IXD2810_REG_SET(pDev->vbase + LINK_LED_ENABLE, data);
}



/*******************************************************************************
 * printMsg - Print message for current mode.
 *
 * RETURNS:  N/A
 */
void printMsg
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port,  /* port number. From 0 to 9. */
	int mode /* modes: 1=supported link 2=non-supported link 3=no link */
	) 
	{
	
	char *speedStr[] = {"10M", "100M", "1000M"};
	char *duplexStr[] = {"Half", "Full"};
	char *linkStr[] = {"Down", "Up"};
	int 	iIndex1, iIndex2;

	int devId = 0;  // TBD Make this MAC #? 

	/* sanity check */
	if (port >= N_PORTS)
		return;

	/* Get string index */
	iIndex1 = pDev->ports[port].speed;
	iIndex2 = pDev->ports[port].duplex;

	if (mode == 1) 
		{	/* supported link mode */
		printf("  MAC%d port%d is configured to %s speed and %s duplex.\n",
	        	devId, port, speedStr[iIndex1], duplexStr[iIndex2]);
		}
	else 
		{			
		printf("  MAC%d port%d is disabled", devId, port);
	
		if (mode == 2) 	/* parallel detection for forced 100M LP */
	      printf(" because half duplex is not supported.\n");
	   else 		/* no link mode */
	     printf(" because no link is established.\n");
	   }
	}

