/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 BY Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/* ixd2810Int.c - Interrupt control handler

modification history
--------------------
01a, 09oct02,eme	created
*/
#include "vm.h"		   
#include "cmd.h"
#include "mac.h"
#include "IXP2800diag.h"
#include "ixd2810Int.h"
#include "ixd2810Autoneg.h"

/* local variables */
LOCAL int autoFlag[N_DEVICES][N_PORTS];

/* forward prototypes */
LOCAL void getModState(IXD2810_DEV *pDev,	uint32_t port);
LOCAL void ixd2810Intr(IXD2810_DEV *pDev);
LOCAL void ixd2810RxLossIsr(IXD2810_DEV *pDev, uint32_t irq, uint32_t status);
LOCAL void ixd2810TxFaultIsr(IXD2810_DEV *pDev, uint32_t irq, uint32_t status);
LOCAL void ixd2810ModDefIsr(IXD2810_DEV *pDev);

/*******************************************************************************
 * ixd2810IntDevInit - Initialize the interrupt handler for the MAC device.
 *
 * RETURNS:  ERROR in case an error occurs, otherwise OK.
 */
#define IXD2810TEST	0x55AA
#define IXD2810IRQ (PIC_NUM_IRQ + 13)
HANDLE IXD2810ISR;


STATUS ixd2810IntDevInit
	(
	IXD2810_DEV *pDev
	) 
	{
	uint32_t status;
	int taskId;
    
	/* Install the Handler */
	IXD2810ISR = InstallHandler (IXD2810IRQ, (HANDLER)ixd2810Intr, (void *)IXD2810TEST);
        
	/* enable interrupt */
	EnableIRQ (IXD2810ISR);

	/* Enable interrupt in GBIC control register */
	IXD2810_REG_SET(pDev->vbase + GBIC_CONTROL, GC_INT_ENABLE);

	/* Read GBIC status value */
	IXD2810_REG_GET(pDev->vbase + GBIC_STATUS, pDev->gbicStatus);

	return OK;
	}

/*******************************************************************************
 * ixd2810IntDevRelease - Removes the interrupt handler.
 *
 * Uninitialize any configuration previous done by the ixd2810IntDevInit(),
 * releases any resources acquired during the intialization process.
 *
 * RETURNS:  ERROR in case an error occurs, otherwise OK.
 */
STATUS ixd2810IntDevRelease
	(
	IXD2810_DEV *pDev
	) 
	{
                                  
	/* Disable interrupt in GBIC control register */	
	IXD2810_REG_SET(pDev->vbase + GBIC_CONTROL, GC_INT_DISABLE);	

	DisableIRQ (IXD2810ISR);

	return OK;
	}

/*******************************************************************************
 * ixd2810Intr - IXD2810 Interrupt Service Routine
 *
 * This routine get called to handle interrupts. Call individual ISRs
 * based on GBIC value. 
 *
 * RETURNS:  N/A
 */
void ixd2810Intr
	(
	IXD2810_DEV *pDev
	) 
	{
	uint32_t gstatus;
	uint32_t irq;
    BOOL irqServiced = FALSE;
    
	pDev->intCnt++;

	/* Read GBIC status register  */
	IXD2810_REG_GET(pDev->vbase + GBIC_STATUS, gstatus);

	/* Find which bit(s) changed */
	irq = gstatus ^ pDev->gbicStatus;

	if (irq & GS_Rx_Loss)
        {
		ixd2810RxLossIsr(pDev, irq, gstatus);
        irqServiced = TRUE;
        }
	if (irq & GS_Tx_Fault)
        {
		ixd2810TxFaultIsr(pDev, irq, gstatus);
        irqServiced = TRUE;
        }
	if (irq & GS_Mod_Def)
        {
		ixd2810ModDefIsr(pDev);
        irqServiced = TRUE;
        }

    if (irqServiced == FALSE)
        {
        printf("Unable to service IRQ %d\n", irq);
        }
	}


/*******************************************************************************
 * ixd2810RxLossIsr - ISR to serve the IXD2810 MAC Rx Loss interrupt
 *
 * Interrupt Service Routine for Rx_Fault_Int. 
 * Display Rx Fault on appropriate device and port.
 *
 * RETURNS:  N/A
 */
void ixd2810RxLossIsr
	(
	IXD2810_DEV *pDev, 	/* pointer to servicing device */
	uint32_t irq,   		/* changed bits in GBIC status */
	uint32_t status 		/* current GBIC status */
	)
	{
	uint32_t	rxBitIndex = GS_RX_STARTING_BIT;  /* GBIC_STATUS bits 29:20 */
	int 	port;
	
	for (port = 0; port < N_PORTS; port++, rxBitIndex <<= 1) 
		{
		if (irq & rxBitIndex) 
			{  /* bit changed for this port */
		    if (status & rxBitIndex) 
		    	{
				disablePort(pDev, port);	
				printf("Rx loss appeared on MAC%d port%d.\n", 
						pDev->devId, port);
				} 
			else 
				{
				enablePort(pDev, port);
				printf("Rx loss disappeared on MAC%d port%d.\n ", 
						pDev->devId, port);
				}
			}
		}
	}

/*******************************************************************************
 * ixd2810TxFaultIsr - ISR to serve the IXD2810 Tx Fault interrupt
 *
 * Interrupt Service Routine for Tx_Fault_Int. 
 * Display Tx Fault on appropriate device and port.
 *
 * RETURNS:  N/A
 */
void ixd2810TxFaultIsr
	(
	IXD2810_DEV *pDev, 	/* pointer to servicing device */
	uint32_t irq,   		/* changed bits in GBIC status */
	uint32_t status 		/* current GBIC status */
	)
	{
	uint32_t	txBitIndex = GS_TX_STARTING_BIT; /* GBIC_STATUS bits 19:10 */
	int 	port;
	
	for (port = 0; port < N_PORTS; port++, txBitIndex <<= 1) 
		{
		if (irq & txBitIndex) 
			{  /* bit changed for this port */
		    if (status & txBitIndex) 
		    	{
                printf("Tx fault appeared on MAC%d port%d.\n",
                          pDev->devId, port);
				} 
			else 
				{
	            printf("Tx fault disappeared on MAC%d port%d.\n ",
                          pDev->devId, port);
				}
			}
		}
	}

/*******************************************************************************
 * ixd2810ModDefIsr - ISR to serve the IXD2810 MAC Mod Def interrupt
 *
 * RETURNS:  N/A
 */
void ixd2810ModDefIsr
	(
	IXD2810_DEV *pDev 	/* pointer to servicing device */
	)
	{
	uint32_t 	modBitIndex = GS_MOD_STARTING_BIT;		
	uint32_t status;
	uint32_t irq;
	int 	port;
	
		/* Read GBIC status register  */
		IXD2810_REG_GET(pDev->vbase + GBIC_STATUS, status); /*  bits 9:0 */

		/* Find which bit(s) changed */
		irq = status ^ pDev->gbicStatus;

		for (port = 0; port < N_PORTS; port++, modBitIndex <<= 1) 
			{
			if (irq & modBitIndex)   /* bit changed for this port */
				{
				/* if bit=1, module is not present */
				if (status & modBitIndex) 
            		{
					disablePort(pDev, port);	
					printf("Module removed from MAC%d port%d.\n", 
			  			pDev->devId, port);
					autoFlag[pDev->devId][port] = 0;
					}
				else 
		    		{		/* if bit=0, module is present */
					printf("Module inserted into MAC%d port%d.\n",
						pDev->devId, port);
					autoFlag[pDev->devId][port] = 1;		
					}
				} /* end if */
			} /* end for */
	
			/* auto-negotiation each port */
			for (port = 0; port < N_PORTS; port++) 
				{
				if (autoFlag[pDev->devId][port] == 1) 
					{
					/* Start auto-negotiation */
					printf("Start autoneg!\n");
#ifndef DEBUG
					doAutoneg(pDev, port);	
#endif
					/* Clear auto-negotiation flag */
					autoFlag[pDev->devId][port] = 0;
					break;
					}
				}
		}


