/* IXP2000I2C.c
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2003 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 *
 *  system: IXP2800
 *  subsystem: DIAG
 *  author:
 *  revisions:
 * 
 * 
 * --------------------------------------------------------------------------
 modification history
--------------------
01a,05jun02, jjd  created from atmel example code
*/

/*
DESCRIPTION

I2C File - Support for I2C via GPIO pins

INCLUDE FILES:  vm.h
				board.h
				ixp2800gpio.h
				timer.h
				ixp2000i2c.h
*/

/* includes */
#include "vm.h"
#include "board.h"
#include "ixp2000gpio.h"
#include "timer.h"
#include "ixp2000i2c.h"

/* globals */
volatile static int timed_out;

/**********************************************************
*
* i2ctimeout - used to set a timeout variable
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
static void i2ctimeout()
    {
    timed_out = TRUE;
    }

/*****************************************************************************
*
* WriteSDA - write i2c data bit
*
* Since SDA has a pullup, to set SDA to 1 set the GPIO as an input.  To set 
* SDA to a 0, set as output and write to the GPIO pin output clear register
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
static void WriteSDA
    (
     unsigned char state
    ) 
    {

    if (state)      
        *(volatile UINT32 *)GPIO_PDCR = (1 << I2C_SDA); /* input ... pullup will pull high or slave will drive low */
    else 
        {
        *(volatile UINT32 *)GPIO_PDSR = (1 << I2C_SDA); /* set as output */
        *(volatile UINT32 *)GPIO_POCR = (1 << I2C_SDA); /* drive low */
        }

    }

/*****************************************************************************
*
* WriteSCL - write i2c clock bit
*
* Since SCL has a pullup, to set SCL to 1 set the GPIO as an input.  To set 
* SCL to a 0, set as output and write to the GPIO pin output clear register
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
static void WriteSCL
    (
     unsigned char state
    )
    {

    if (state)
        *(volatile UINT32 *)GPIO_PDCR = (1 << I2C_SCL); /* input ... pullup will pull high or slave will drive low */
    else
    {
        *(volatile UINT32 *)GPIO_PDSR = (1 << I2C_SCL); /* output */
        *(volatile UINT32 *)GPIO_POCR = (1 << I2C_SCL); /* drive low */
    }

    }

/*****************************************************************************
*
* SetSCLHigh - set the SCL signal high
*
* Since SCL can be extended low by a slow slave, wait for SCL to be read as
* high before returning.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
static unsigned char SetSCLHigh(void)
    {

    WriteSCL(1); // release SCL

    /* setup a timeout (atmel used 32us) */
    timed_out = FALSE; // will be set after approximately 34 us

    InstallOneshotTimer(1,i2ctimeout,0); /* timeout in 1ms */

    // wait till SCL goes to a 1
    while (!(*(volatile UINT32 *)GPIO_PLR & (1 << I2C_SCL)) && !timed_out);

    return(timed_out);

    }

/*****************************************************************************
*
* BitDelay - delay for a bit time
*
* Delay a sufficient amount for a slave to read an output bit.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
static void BitDelay(void)
    {
     /* change to call board specific delay */
    return;
    }

/*****************************************************************************
*
* SendStopBit - output a stop bit 
*
* A stop bit is a rising edge on SDA while SCL is high.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
static void SendStopBit(void)
    {
    
    WriteSDA(0);
    BitDelay();
    SetSCLHigh();
    BitDelay();
    WriteSDA(1);
    BitDelay();
    }

/*****************************************************************************
*
* SendStartBit - output a start bit
*
* A start bit is a falling edge on SDA wihle SCL is high.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
static void SendStartBit(void)
    {
    
    WriteSDA(1);
    BitDelay();
    SetSCLHigh();
    BitDelay();
    WriteSDA(0);
    BitDelay();
    WriteSCL(0);
    BitDelay();
    }

/*****************************************************************************
*
* SendByte - output a byte serially
*
* Output a byte serially, either MSB first or LSB first selectable by 'order'
* parameter.  Wait for ACK signal from slave, return nonzero when there is no
* ACK received in time. 
*
* RETURNS: unsigned char (0 or nonzero)
*
* ERRNO: N/A
*/
static unsigned char SendByte
    (
     unsigned char byte, 
     unsigned char order
    )
    {
    unsigned char i;
    unsigned char error;
    
    for (i = 0; i < 8; i++) 
        {
        if (order) 
            {
            WriteSDA(byte & 0x80); // if > 0 SDA will be a 1
            byte = byte << 1; // send each bit, MSB first for address
            }
        else 
            {
            WriteSDA(byte & 0x01); // if > 0 SDA will be a 1
            byte = byte >> 1; // send each bit, LSB first for data
            }
        
        BitDelay();
        SetSCLHigh();
        BitDelay();
        WriteSCL(0);
        BitDelay();
        }
    
    // now for an ack
    
    // master generates clock pulse for ACK
    WriteSDA(1); // release SDA ... listen for ACK
    BitDelay();
    SetSCLHigh(); // ACK should be stable ... data not allowed to change when SCL is high
 
    // SDA at 0 ?
    error = (*(volatile UINT32 *)GPIO_PLR & (1 << I2C_SDA)); // ack didn't happen if bit 0 = 1

    WriteSCL(0);
    BitDelay();
 
    return(error);

    }

/*****************************************************************************
*
* GetByte - read a byte serially
*
* Read a byte, parameter 'lastbyte' indicates when no ACK should be sent to 
* slave because it is the last byte to be read.
*
* RETURNS: unsigned char (byte read from slave)
*
* ERRNO: N/A
*/
static unsigned char GetByte
    (
     unsigned char lastbyte
    )
    {

    // lastbyte == 1 for last byte
 
    unsigned char i, bit;
    unsigned char result = 0;

    *(volatile UINT32 *)GPIO_PDCR = (1 << I2C_SDA); /* set as input */

    for (i = 0;i < 8;i++) 
        { // each bit at a time, LSB first
        SetSCLHigh();
        BitDelay();
        bit = (*(volatile UINT32 *)GPIO_PLR & (1 << I2C_SDA));
        result = (bit << (i)) | result;
        WriteSCL(0);
        BitDelay();
        }
                       
    // send ACK
    WriteSDA(lastbyte); // no ack on last byte ... lastbyte = 1 for the last byte
    BitDelay();
    SetSCLHigh();
    BitDelay();
    WriteSCL(0);
    BitDelay();
    WriteSDA(1);
    BitDelay();
    
    return(result);

    }

/*****************************************************************************
*
* i2c_read - read a byte from i2c device
*
* Read an addressed byte from the selected i2c device.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
void i2c_read
    (
     unsigned int device,
     unsigned int byteaddr, 
     char *buf
    )
    {

    I2C_ENABLE();

    BitDelay();

    SendStartBit();
    SendByte(device+WRITE,MSB_FIRST);
    SendByte(byteaddr,MSB_FIRST);

    SendStartBit();
    SendByte(device+READ,MSB_FIRST);

    *buf = GetByte(1);

    SendStopBit();

    I2C_DISABLE();

    }

/*****************************************************************************
*
* i2c_seq_read - read multilple bytes from i2c device
*
* Read an addressed byte range from the selected i2c device.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
void i2c_seq_read
    (
     unsigned int device,
     unsigned int byteaddr, 
     char *buf,
   	 unsigned int count
    )
    {
    int i;

    I2C_ENABLE();

    BitDelay();

    SendStartBit();
    SendByte(device+WRITE,MSB_FIRST);
    SendByte(byteaddr,MSB_FIRST);

    SendStartBit();
    SendByte(device+READ,MSB_FIRST);

    for (i=0;i<(count-1);i++)
      buf[i] = GetByte(0);

    buf[count-1] = GetByte(1);

    SendStopBit();

    I2C_DISABLE();

    }

 
/*****************************************************************************
*
* i2c_write - write a byte from i2c device
*
* Write an addressed byte to the selected i2c device.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
void i2c_write
    (
     unsigned int device,
     unsigned int byteaddr,
     unsigned char data
    )
    {
    I2C_ENABLE();

    BitDelay();

    SendStartBit();
    SendByte(device+WRITE,MSB_FIRST);
    SendByte(byteaddr,MSB_FIRST);

    SendByte(data,MSB_FIRST);

    SendStopBit();

    I2C_DISABLE();
    }

/*****************************************************************************
*
* i2c_page_write - write a page to i2c device
*
* Write a 'page' of bytes to i2c device.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
void i2c_page_write
    (
     unsigned int device,
     unsigned int byteaddr,
     unsigned char *data,
     unsigned int numbytes
    )
    {
    int i;

    I2C_ENABLE();

    BitDelay();

    SendStartBit();
    SendByte(device+WRITE,MSB_FIRST);
    SendByte(byteaddr,MSB_FIRST);

    for (i = 0; i < numbytes; i++)
        {
        SendByte(data[i],MSB_FIRST);
        }

    SendStopBit();

    I2C_DISABLE();
    }
    
