/* ixp2000pci.c - Intel IXP2000 PCI */

/* Copyright 1984-1999 Wind River Systems, Inc., 2002-2003 Intel Corporation */

/*
 * $Header: /IXDP2800_BoardSupport_3.5_FCS_310D/VisionWARE2.1/PROJECTS/IXP28X0/Board/IXMB2800/ixp2000pci.c 1     3/05/04 2:54p Ajpasqui $
 * $Log: /IXDP2800_BoardSupport_3.5_FCS_310D/VisionWARE2.1/PROJECTS/IXP28X0/Board/IXMB2800/ixp2000pci.c $
 * 
 * 1     3/05/04 2:54p Ajpasqui
 * 
 * 3     9/17/03 11:24a Ajpasqui
 * Modified the Master ABort msg string
 * 
 * 2     8/20/03 5:07p Daly
 * added vss tags
 *
modification history
--------------------
01a,29Apr02, jjd  created from 01a brhpci.c
01b,08aug02, jjd  correct SRAM size allocation
01c,24feb03, jjd  fixed PCI init for Egress/Ingress
01d,25feb03, jjd  fixed PCI config cycles
01e,06mar03, jjd  don't allow NPU to see themselves on bus
01f,11mar03, ajp  revert previous change
01g,19mar03, jjd  fixed up PCI init, different maps for Egress/Ingress again
02a,30may03, jjd  A-step PCI workaround
02b,20jun03, jjd  set PCI_RCOMP for A2
02c,25jun03, jjd  change PCI_RCOMP to midpoint value for A2
03a,30aug03, jjd  added modification history, now managed by vss log
03b,17sep03, ajp  modified the master abort msg
*/

/*
DESCRIPTION

IXP2000 PCI Configuration File

INCLUDE FILES:	vm.h
				pci.h
				ixp2000pci.h	
*/

/* includes */
#include "vm.h"
#include "pci.h"
#include "ixp2000pci.h"
#include "ixp2x00.h"

/* error checking */
#if (EGRESS_PCI_SRAM_SIZE < 0x40000)
#error Must allocate more space to SRAM
#endif
#if (EGRESS_PCI_DRAM_SIZE < 0x100000)
#error Must allocate more space to DRAM
#endif

/* defines */
#define IXP2000_PCI
#define PCI_CFG0_BASE		0xda000000
#define PCI_CFG1_BASE           0xdb000000
#define PCI_INT_LINE_NUM	29
#define IXP2000_DEV		(PCI_MAX_DEV + 1)
                 
typedef union 					/* CFG_DATA_REG */
{
    unsigned long	ul;			/* Unsigned Long */
    unsigned short	us[2];		/* Unsigned Short */
    unsigned char	uc[4];		/* Unsigned Char */
}CFG_DATA_REG;

/* Forward Declarations */
static void vCfgWrite (UINT32 Offset, UINT32 Data, int Size);
static UINT32 vCfgRead (UINT32 Offset, int Size);
static void PciWorkaround(void);

/***************************************************************************
*
* PciEncode - Form a PCI-encoded address
*
* Form a PCI-encoded address for a particular bus/device/function 
*
* RETURNS: UINT32 PCI-encoded address for the device 
*
* ERRNO:
*/

UINT32 PciEncode
    (
    unsigned int BusNumber, 		/* input: Bus Number */
    unsigned int DeviceNumber, 		/* input: Device Number */
    unsigned int Function		/* input: Function Number */
    )
    {
    UINT32 address;		        /* PCI-Encoded Address */

    if (BusNumber == 0)
        {
	/*
	 * local bus segment so need a type 0 config cycle
         * XXX fix this so we don't go beyond our available bus 0 resources
         * (change the '8' to some #define)
	 */
	address = PCI_CFG0_BASE;
        address |= ((Function & 0x7) << 8);
	if (IS_EGRESS_NPU) {
        switch (DeviceNumber)
            {
            /* deer island implementation (from egress POV) */
            case 0: address |= (1 << 20); break; /* bridge */
            case 1: address |= (1 << 21); break; /* SLAVE */
            case 2: address |= (1 << 22); break; /* bridge */
            case IXP2000_DEV: address = IXP2000_PCI_BASE; break;
            default: address |= (1 << 16); break; /* needs to have one-hot or else it hangs */
            }
	}
	else {
        switch (DeviceNumber)
            {
            /* deer island implementation (from egress POV) */
            case 0: address |= (1 << 20); break; /* bridge */
            case 2: address |= (1 << 22); break; /* bridge */
            case 3: address |= (1 << 23); break; /* MASTER */
            case IXP2000_DEV: address = IXP2000_PCI_BASE; break;
            default: address |= (1 << 16); break; /* needs to have one-hot or else it hangs */
            }
	}

	}
    else
        {
	/* 	-- bus !=0 --
	 *
	 * not the local bus segment so need a type 1 config cycle
	 * 
	 */
	
	address = PCI_CFG1_BASE;
	address |= ((BusNumber & 0xFF) << 16);	/* bits 23..16 = bus number */
	address |= ((DeviceNumber & 0x1F) << 11);  	/* bits 15..11 = slot number */
	address |= ((Function & 0x07) << 8);  	/* bits 10..8  = function num */
	}
	
    return address;

    }


/***************************************************************************
*
* vCSRWrite - IXP2000 PCI CSR write
*
* Write to IXP2000 PCI CSR space. 
*
* RETURNS: N/A
*
* ERRNO:
*/

static void vCSRWrite
    (
    UINT32 Offset,  /* input: Register offset to write */
    UINT32 Data,    /* input: Data to write */
    int Size        /* input: Size to write */
    )
    {

    CFG_DATA_REG *cfg_ptr = (CFG_DATA_REG *)(IXP2000_PCI_CSR_BASE | (Offset & ~3));
    
    /* Determine the size of the write and write the data */
    switch (Size) 
        {
        case 4:
            cfg_ptr->ul = Data;
	        break;
        case 2:
            cfg_ptr->us[(Offset&3)>>1] = Data;
	        break;
        default:
            cfg_ptr->uc[Offset&3] = Data;
        }

    return;

    }


/***************************************************************************
*
* vCSRRead - IXP2000 PCI CSR read
*
* Read to IXP2000 PCI CSR space. 
*
* RETURNS: UINT32 - Register value
*
* ERRNO:
*/

static UINT32 vCSRRead
    (
    UINT32 Offset,  /* input: Register offset to read */
    int Size        /* input: Size to read */
    )
    {
    
    CFG_DATA_REG *cfg_ptr = (CFG_DATA_REG *)(IXP2000_PCI_CSR_BASE | (Offset & ~3));
    UINT32 regData;

    /* Determine the size of the read and read the data */
    switch (Size) 
        {
        case 4:
            regData = cfg_ptr->ul;
            break;
        case 2:
            regData = cfg_ptr->us[(Offset&3)>>1];
            break;
        default:
            regData = cfg_ptr->uc[Offset&3];
        }

    return (regData);

    }

/***************************************************************************
*
* vCfgWrite - IXP2000 PCI configuration write
*
* Write to IXP2000 PCI configuration space. 
*
* RETURNS: N/A
*
* ERRNO:
*/

static void vCfgWrite
    (
    UINT32 Offset,  /* input: Register offset to write */
    UINT32 Data,    /* input: Data to write */
    int Size        /* input: Size to write */
    )
    {

    CFG_DATA_REG *cfg_ptr = (CFG_DATA_REG *)(IXP2000_PCI_BASE | (Offset & ~3));
    
    /* Determine the size of the write and write the data */
    switch (Size) 
        {
        case 4:
            cfg_ptr->ul = Data;
	        break;
        case 2:
            cfg_ptr->us[(Offset&3)>>1] = Data;
	        break;
        default:
            cfg_ptr->uc[Offset&3] = Data;
        }

    return;

    }


/***************************************************************************
*
* vCfgRead - IXP2000 PCI configuration read
*
* Read to IXP2000 PCI configuration space. 
*
* RETURNS: UINT32 - Register value
*
* ERRNO:
*/

static UINT32 vCfgRead
    (
    UINT32 Offset,  /* input: Register offset to read */
    int Size        /* input: Size to read */
    )
    {
    
    CFG_DATA_REG *cfg_ptr = (CFG_DATA_REG *)(IXP2000_PCI_BASE | (Offset & ~3));
    UINT32 regData;

    /* Determine the size of the read and read the data */
    switch (Size) 
        {
        case 4:
            regData = cfg_ptr->ul;
            break;
        case 2:
            regData = cfg_ptr->us[(Offset&3)>>1];
            break;
        default:
            regData = cfg_ptr->uc[Offset&3];
        }

    return (regData);

    }

/***************************************************************************
*
* PciCfgWrite - PCI configuration write
*
* Write to PCI configuration space. 
*
* RETURNS: N/A
*
* ERRNO:
*/

void PciCfgWrite
    (
    UINT32 PciDevice,   /* input: Pci Device */
    UINT32 Offset,      /* input: Register Offset */
    UINT32 Data,        /* input: Data to write to register */
    int Size            /* input: Size to write */
    )
    {

    CFG_DATA_REG *cfg_ptr = (CFG_DATA_REG *)(PciDevice | (Offset & ~3));

    /* Disable Interrupts */
    EnterCriticalSection();

    /* Determine the size of the write and write the data */
    switch (Size)
        {
        case 4:
            cfg_ptr->ul = Data;
            break;
        case 2:
#ifdef __LITTLE_ENDIAN__
            cfg_ptr->us[(Offset&3)>>1] = Data;
#else
            cfg_ptr->us[((Offset&3)^3)>>1] = Data;
#endif
            break;
        default:
#ifdef __LITTLE_ENDIAN__
            cfg_ptr->uc[Offset &3] = Data;
#else
            cfg_ptr->uc[(Offset&3) ^ 3] = Data;
#endif
        }

    if ( vCSRRead(PCI_CONTROL,REG_32BIT) & (PCI_CONTROL_Dct | PCI_CONTROL_Rma) )
        {
        vCSRWrite(PCI_CONTROL,(PCI_CONTROL_Dct | PCI_CONTROL_Rma),REG_32BIT);
        vCfgWrite(IXP2000_PCI_CFG_STATUS,0x3000,REG_16BIT);
        }

    /* Enable Interrupts */
    ExitCriticalSection();

    return;

    }


/***************************************************************************
*
* PciCfgRead - Read from PCI config space
*
* Read data from PCI config space of given size
* pdata
*
* RETURNS: UINT32 - value read from PCI config space, ERROR if HOST BRIDGE
*
* ERRNO:
*/

UINT32 PciCfgRead
    (
	UINT32 PciDevice,   /* input: Pci Device */
	UINT32 Offset,      /* input: Register Offset */
	int Size            /* input: Size to write */
    )
    {
    
    CFG_DATA_REG *cfg_ptr = (CFG_DATA_REG *)(PciDevice | (Offset & ~3));
    volatile UINT32 regData;

    /* Disable Interrupts */
    EnterCriticalSection();

    switch (Size) 
        {
        case 4:
            regData = cfg_ptr->ul;
            break;
        case 2:
#ifdef __LITTLE_ENDIAN__
            regData = cfg_ptr->us[(Offset&3)>>1];
#else
            regData = cfg_ptr->us[((Offset & 3)^3)>>1];
#endif
            break;
        default:
#ifdef __LITTLE_ENDIAN__
            regData = cfg_ptr->uc[Offset&3];
#else
            regData = cfg_ptr->uc[(Offset&3) ^ 3];
#endif
        }

    if ( vCSRRead(PCI_CONTROL,REG_32BIT) & (PCI_CONTROL_Dct | PCI_CONTROL_Rma) )
        {
        vCSRWrite(PCI_CONTROL,(PCI_CONTROL_Dct | PCI_CONTROL_Rma),REG_32BIT);
        vCfgWrite(IXP2000_PCI_CFG_STATUS,0x3000,REG_16BIT);
        regData = -1;
        }

    /* Enable interrupts */
    ExitCriticalSection();

    return (regData);

    }


/***************************************************************************
*
* pciConfigInLong - Read Long from PCI configuration space
*
* Read one long from the PCI configuration space.  Data read is placed in
* pdata
*
* RETURNS: OK
*
* ERRNO:
*/

int pciConfigInLong(
    int bus,			/* input: PCI Bus */
    int device,			/* input: PCI Device */
    int function,		/* input: PCI Function */
    int address,		/* input: Register Address */
    UINT32 *pdata		/* input: User data pointer */
    )
    {

    ULONG	configaddr;	/* Configuration access address */
    UINT32	tmp;        /* Scratch Variable */

    /* Get the configuration address */
    configaddr = PciEncode (bus,device,function);

    /* Read the data from the config space */
    tmp = PciCfgRead (configaddr,address,REG_32BIT);

    /* Save the data read */
    *pdata = (int)tmp;
    return OK;

    }


/***************************************************************************
*
* pciConfigInWord - Read Word from PCI configuration space
*
* Read one word from the PCI configuration space.  Data read is placed in
* pdata
*
* RETURNS: OK
*
* ERRNO:
*/

int pciConfigInWord
    (
	int bus,			/* input: PCI Bus */
	int device,			/* input: PCI Device */
	int function,		/* input: PCI Function */
	int address,		/* input: Register Address */
	UINT16 *pdata		/* input: User data pointer */
    )
    {

    ULONG	configaddr;	// Configuration access address
    UINT32	tmp;

    /* Get the configuration address */
    configaddr = PciEncode (bus,device,function);

    /* Read the data from the config space */
    tmp = PciCfgRead (configaddr,address,REG_16BIT);

    /* Save the data read */
    *pdata = (short)tmp;
    return OK;

    }


/***************************************************************************
*
* pciConfigInByte - Read Byte from PCI configuration space
*
* Read one byte from the PCI configuration space.  Data read is placed in
* pdata
*
* RETURNS: OK
*
* ERRNO:
*/

int pciConfigInByte
    (
	int bus,			/* input: PCI Bus */
	int device,			/* input: PCI Device */
	int function,		/* input: PCI Function */
	int address,		/* input: Register Address */
	UINT8 *pdata		/* input: User data pointer */
    )
    {

    ULONG	configaddr;	// Configuration access address
    UINT32	tmp;

    /* Get the configuration address */
    configaddr = PciEncode (bus,device,function);

    /* Read the data from the config space */
    tmp = PciCfgRead (configaddr,address,REG_8BIT);

    /* Save the data read */
    *pdata = (char)tmp;

    return OK;

    }


/***************************************************************************
*
* pciConfigOutLong - Write Long into PCI configuration space
*
* Writing one long into the PCI configuration space.
*
* RETURNS: OK
*
* ERRNO:
*/

int pciConfigOutLong
    (
	int bus,			/* input: PCI Bus */
	int device,			/* input: PCI Device */
	int function,		/* input: PCI Function */
	int address,		/* input: Register Address */
    UINT32 data			/* input: Data to write out */
    )
    {

    ULONG	configaddr;	/* Configuration access address */
    UINT32	tmp;

    /* Get the configuration address */
    configaddr = PciEncode (bus,device,function);

    /* Save the data */
    tmp = data;

    /* Write the data to the config space */
    PciCfgWrite (configaddr,address,tmp,REG_32BIT);

    return OK;

    }


/***************************************************************************
*
* pciConfigOutWord - Write Word into PCI configuration space
*
* Writing one word into the PCI configuration space.
*
* RETURNS: OK
*
* ERRNO:
*/

int pciConfigOutWord
    (
	int bus,			/* input: PCI Bus */
	int device,			/* input: PCI Device */
	int function,		/* input: PCI Function */
	int address,		/* input: Register Address */
    UINT16 data			/* input: Data to write out */
    )
    {

    ULONG	configaddr;	/* Configuration access address */
    UINT32	tmp;

    /* Get the configuration address */
    configaddr = PciEncode (bus,device,function);

    /* Save the data */
    tmp = data;

    /* Write the data to the config space */
    PciCfgWrite (configaddr,address,tmp,REG_16BIT);

    return OK;

    }


/***************************************************************************
*
* pciConfigOutByte - Write Byte into PCI configuration space
*
* Writing one byte into the PCI configuration space.
*
* RETURNS: OK
*
* ERRNO:
*/

int pciConfigOutByte
    (
	int bus,			/* input: PCI Bus */
	int device,			/* input: PCI Device */
	int function,		/* input: PCI Function */
	int address,		/* input: Register Address */
    UINT8 data			/* input: Data to write out */
    )
    {

    ULONG	configaddr;	/* Configuration access address */
    UINT32	tmp;

    /* Get the configuration address */
    configaddr = PciEncode (bus,device,function);

    /* Save the data */
    tmp = data;

    /* Write the data to the config space */
    PciCfgWrite (configaddr,address,tmp,REG_8BIT);

    return OK;

    }


/***************************************************************************
*
* PciPrimaryPCIInit - Init primary PCI bus
*
* Initialize the Primary PCI Bus
*
* RETURNS: OK
*
* ERRNO:
*/

UINT32 PciPrimaryPCIInit (void)
    {                 

#define BusNo	0
#define DevNo	IXP2000_DEV
#define FuncNo	0							

    UINT16  tmp16;  /* Scratch Pad Variable 16 - bit */
    UINT32  idx;    /* Scratch Counter */

    if (IS_EGRESS_NPU)
        {
        *(volatile unsigned int *)RESET_0 |= (RESET_0_AssertPciRst);
        *(volatile unsigned int *)RESET_0 &= ~(RESET_0_AssertPciRst);
        }
#if 1
	if ((*(volatile UINT *)PRODUCT_ID & PRODUCT_ID_MajRevMask) == PRODUCT_ID_Astep)
		PciWorkaround();
#endif
    /* Write the PCI Command Register */
    vCfgWrite (IXP2000_PCI_CFG_COMMAND, 0, REG_16BIT);

    /* allocate some SRAM to PCI */
    /* XXX fix this */
    vCSRWrite (SRAM_BASE_ADDR_MASK, (((IS_EGRESS_NPU ? EGRESS_PCI_SRAM_SIZE :
                                     INGRESS_PCI_SRAM_SIZE) - 1) & ~0x3FFFF), REG_32BIT);
    /* allocate some DRAM to PCI */
    vCSRWrite (DRAM_BASE_ADDR_MASK, (((IS_EGRESS_NPU ? EGRESS_PCI_DRAM_SIZE :
                                     INGRESS_PCI_DRAM_SIZE) - 1) & ~0xFFFFF), REG_32BIT);

    /* Configure the IXP PCI */
    vCfgWrite (IXP2000_PCI_CFG_LATENCY_TIMER, PCI_LAT_TIMER, REG_8BIT);
    vCfgWrite (IXP2000_PCI_CFG_DEV_INT_LINE, PCI_INT_LINE_NUM, REG_8BIT);
    vCfgWrite (IXP2000_PCI_CFG_STATUS, 0xfb00, REG_16BIT);

    *(volatile UINT *)RESET_0 |= RESET_0_InitComplete;
    (void)*(volatile UINT *)RESET_0; /* flush if if necessary */

    /* setup the PCI_ADDR_EXT register */
    vCSRWrite (PCI_ADDR_EXT, (PCI_ADDR_EXT_PioAddr(PCI_PP_IO_START) |
                             PCI_ADDR_EXT_PmemAddr(PCI_PP_MEM_START)),
                             REG_32BIT);

    /* configure DRAM base address */
    if (IS_EGRESS_NPU)
        {
        vCfgWrite (IXP2000_PCI_CFG_BASE_ADDRESS_2, 0, REG_32BIT);
        vCfgWrite (IXP2000_PCI_CFG_BASE_ADDRESS_0, (0 + EGRESS_PCI_DRAM_SIZE), REG_32BIT);
	vCfgWrite (IXP2000_PCI_CFG_BASE_ADDRESS_1, (0 + EGRESS_PCI_DRAM_SIZE+0x1000000), REG_32BIT);
        vCfgWrite (IXP2000_PCI_CFG_COMMAND, (PCI_MST_ENBL | PCI_MEM_ACCESS_ENBL), REG_16BIT);
        DelayNMsec(100);
        }

    vCSRWrite(PCI_XSCALE_INT_EN, PCI_XSCALE_INT_EN_PCIBEnable, REG_32BIT);

    return (OK);

    }


UINT32 PciSecondaryPCIInit (void)
    {

    return;

    }

extern int DownloadUcode(unsigned short, unsigned short, unsigned int *);
extern void rom_outstring(char *);

#include "pciconfiglib.h"
#include "pciworkaround.uh"

/*
 * PciWorkaround - Get around PCI bug in A-step
 *
 * The workaround involves some microcode (found in pciworkaround.uc) that does
 * a sanity check on the PCI bus operation.  XScale monitors progress and when 
 * both NPUs pass, operation continues in normal fashion.  Once the PCI bus is
 * determined to be working at this point, it continues to work properly.  Only 
 * out of RESET is there a potential problem.  This ugly hack goes away with
 * B-step.  It is also specific to Deer Island.  Other platforms will have to 
 * modify this if they don't have the same type of dual NPU system as Deer 
 * Island.
 *
 * RETURNS: N/A
 */
static void PciWorkaround(void)
	{

	int status;
	int complete;
	rom_outstring("Starting PciWorkaround\n");
	/* check for empty fifos, this code is on the old laptop */
	if ((*(volatile UINT *)PRODUCT_ID & PRODUCT_ID_MinRevMask) == 2)
		vCfgWrite(IXP2000_PCI_RCOMP_OVERRIDE, 0x100505, REG_32BIT);
    vCfgWrite (IXP2000_PCI_CFG_STATUS, 0xfb00, REG_16BIT);

	do 
		{
		complete = 0;
		*(volatile UINT *)RESET_0 |= RESET_0_Pci;
		DelayNUsec(5);
		*(volatile UINT *)RESET_0 &= ~RESET_0_Pci;
		/* assign temporary addresses */
		rom_outstring("Assign addresses\n");
    	if (IS_EGRESS_NPU)
        	{
        	vCfgWrite (IXP2000_PCI_CFG_BASE_ADDRESS_2, 0, REG_32BIT);
        	vCfgWrite (IXP2000_PCI_CFG_BASE_ADDRESS_0, (0 + EGRESS_PCI_DRAM_SIZE), REG_32BIT);
			vCfgWrite (IXP2000_PCI_CFG_BASE_ADDRESS_1, (0 + EGRESS_PCI_DRAM_SIZE+0x1000000), REG_32BIT);
        	vCfgWrite (IXP2000_PCI_CFG_COMMAND, (PCI_MST_ENBL | PCI_MEM_ACCESS_ENBL), REG_16BIT);
        	}
		else
			{
			vCfgWrite (IXP2000_PCI_CFG_BASE_ADDRESS_2, 0x10000000, REG_32BIT);
			vCfgWrite (IXP2000_PCI_CFG_BASE_ADDRESS_0, (0x10000000 + INGRESS_PCI_DRAM_SIZE), REG_32BIT);
			vCfgWrite (IXP2000_PCI_CFG_BASE_ADDRESS_1, (0x10000000 + INGRESS_PCI_DRAM_SIZE + 0x1000000), REG_32BIT);
			vCfgWrite (IXP2000_PCI_CFG_COMMAND, (PCI_MST_ENBL | PCI_MEM_ACCESS_ENBL), REG_16BIT);
			}
	
    	*(volatile UINT *)RESET_0 |= RESET_0_InitComplete;
    	(void)*(volatile UINT *)RESET_0; /* flush if if necessary */
	
		vCSRWrite(MAILBOX_0, REG_32BIT, 0);
		vCSRWrite(MAILBOX_1, REG_32BIT, 0);

		/* run the microcode */
		rom_outstring("Run microcode\n");
		if (status = DownLoadUcode(0,sizeof(pciworkaround),pciworkaround))
			{
			rom_outstring("DownloadUcode failed....");
			status += 0x30;
			rom_outstring((char *)&status);
			}
		rom_outstring("waiting for start...");
		while ((vCSRRead(MAILBOX_0,REG_32BIT) != 0xA5A5A5A5));
		rom_outstring("done\n");

		/* check for errors */
    	if ( vCSRRead(PCI_CONTROL,REG_32BIT) & PCI_CONTROL_Rma )
        	{
			rom_outstring("Notice: got Master Abort!\n");
        	vCSRWrite(PCI_CONTROL,(PCI_CONTROL_Dct | PCI_CONTROL_Rma),REG_32BIT);
        	vCfgWrite(IXP2000_PCI_CFG_STATUS,0x3000,REG_16BIT);
			}
		else if (vCSRRead(MAILBOX_1,REG_32BIT) == 0xDEADBEEF) /* check for completeness */
			{
			rom_outstring("no errors!\n");
			complete = 1;
			vCSRWrite(MAILBOX_0, 0x5A5A5A5A, REG_32BIT);
			(void)vCSRRead(MAILBOX_0,REG_32BIT);
			}
		else if (vCSRRead(MAILBOX_1,REG_32BIT) == 0xFEEDFACE)
			{
			rom_outstring("errors!\n");
			complete = 0;
			}

		/* if we are EGRESS, we must wait for Ingress to be done */
		if (complete && IS_EGRESS_NPU)
			{
			rom_outstring("waiting for Ingress to complete...");
			while (*(volatile UINT *)0xf40fe050 != 0x5A5A5A5A)
				{
				if ( vCSRRead(PCI_CONTROL,REG_32BIT) & (PCI_CONTROL_Dct | PCI_CONTROL_Rma) )
					{
        			vCSRWrite(PCI_CONTROL,(PCI_CONTROL_Dct | PCI_CONTROL_Rma),REG_32BIT);
        			vCfgWrite(IXP2000_PCI_CFG_STATUS,0x3000,REG_16BIT);
					}
				}	
			pciConfigOutWord(0,1,0,PCI_CFG_COMMAND,0); /* clear MasterEnable and MemEnable on Slave */
			rom_outstring("done\n");
			}

	} while (!complete);

	if (!IS_EGRESS_NPU)
		{
		rom_outstring("waiting for CMD to clear...");
		while ((vCfgRead(IXP2000_PCI_CFG_COMMAND, REG_16BIT)));
		rom_outstring("done\n");
		}
	*(volatile UINT *)RESET_0 &= ~RESET_0_InitComplete;
	(void)*(volatile UINT *)RESET_0; /* flush if if necessary */
	
	}