/***************************************************************************
*
* Module:	serial16550.c   
*
* Description:  Serial driver for the classic PC UART
*
* Project:	HSI VisionWare -- vWare
*
* Functions:
*
* Notes:	This is a interrupt driven driver. The read function is not implemented
*			for poll mode
*
* History:	See Visual SourceSafe History
*
* Disclaimer:	Copyright 1999-2000 Wind River Systems, Inc. All Rights Reserved
****************************************************************************/

//  Includes
#include "vm.h"  
#include "esdriver.h"
#include "serial16550.h"
#include "ixp2x00.h"


// Typedefs
/****************************************************************************
 * Here we "superclass" the tESDRIVER structure so that we can add some
 * driver-specific fields to the end of the structure.
 ***************************************************************************/
typedef struct
{
    tESDRIVER	Esdriver;
    U550		*pReg;
    UCHAR		lcr_reg;
    UINT32		nBaudRate;
    UINT32		Clock16550;
    UINT32		Interrupt16550;
    HANDLE		hISR;
} tDRIVER;

//  External References

//  Forward References
static HANDLE	Open(tDRIVER *pInitStruct);
static DSTATUS	Write(tDRIVER *pHandle, UINT8 *cBuffer, unsigned intnBytes);
static DSTATUS	Ioctl(tDRIVER *pHandle, UINT32 nFuncCode, void *pArgs);
static DSTATUS	Close(tDRIVER *pHandle);
static int Isr(tDRIVER *pHandle);

//  Global Declarations
tDRIVER Esdriver =
{
    {
		(ESDRIVEROpen)	Open,
		(ESDRIVERClose)	Close,
		(ESDRIVERWrite)	Write,
		(ESDRIVERIoctl)	Ioctl
    },
	(volatile U550 *) SERIALADR,
	0,
	BAUDRATE,		// declared in board.h
	SERIALCLK,
	SERIALIRQ
};

//  Local Declarations (statics)

//  Local Prototypes

/***************************************************************************
* Function:	SetBaud(tDRIVER *pI)
*
* Purpose:	Set the UART baud rate
*
* Returns:	void
****************************************************************************/
static void setBaud(tDRIVER *pI)
{
	USHORT	div;
	int pllMult;

	pllMult = (*(volatile int *)(STRAP_OPTIONS) & STRAP_OPTIONS_CfgPllMask) >> 8;

	if (pllMult == 0x3f)
		pI->Clock16550 = 41666666;
	else
		pI->Clock16550 = 50000000;
#ifndef IXP2x00
    div = (pI->Clock16550 * 10)/pI->nBaudRate/16;
    div = (div + 5)/10;
#else
    div = (pI->Clock16550)/pI->nBaudRate/16;
#endif
	EnterCriticalSection();

    pI->pReg->LCR = pI->lcr_reg|DLA_BIT;
    pI->pReg->DLLO = div & 0xff;
    pI->pReg->DLHI = (div >> 8);
    pI->pReg->LCR = pI->lcr_reg;

	ExitCriticalSection();
}



/***************************************************************************
* Function:	Open(tDRIVER *pInitStruct)
*
* Purpose:	Open and init the SMC channel
*
* Returns:	HANDLE ptr
****************************************************************************/
static HANDLE Open(tDRIVER *pI)
{
	volatile U550	*pReg = pI->pReg;

	/************************************************************************
	 * Register a handler for the receive interrupt.  The third parameter that
	 * gets passed to InstallHandler is the parameter that will be passed to
	 * the registered ISR handler whenever it gets called.
	 ***********************************************************************/
	pI->hISR = InstallHandler((UINT32)pI->Interrupt16550, (HANDLER)Isr, (void *)pI);

	/************************************************************************
	 * Initialize the UART.
	 ***********************************************************************/
    pI->lcr_reg = 3;
    pReg->IER = UUEN;
    pReg->FCR = (ENABLE_FIFO|CLEAR_RCV_FIFO|CLEAR_XMIT_FIFO);
#ifndef IXP2x00
    pReg->MCR = (MOD_DCTS|MOD_DDSR);
#endif
    setBaud(pI);
    pReg->IER = UUEN|RXINT;
#ifndef IXP2x00
    pReg->MCR = OUT2;
#endif
	/************************************************************************
	 * Enable the IRQ.
	 ***********************************************************************/
	EnableIRQ(pI->hISR);

	/************************************************************************
	 * The value that gets returned here is the parameter that will be passed
	 * to our Close, Write and IOCTL functions.
	 ***********************************************************************/
	return (HANDLE)&Esdriver;
}

/***************************************************************************
* Function:	Close(tDRIVER *pHandle)
*
* Purpose:	Close and deinit SMC channel
*
* Returns:	DSTATUS
****************************************************************************/
static DSTATUS Close(tDRIVER *pHandle)
{
	volatile U550	*pReg = pHandle->pReg;

	/************************************************************************
	 * Disable the IRQ.
	 ***********************************************************************/
	DisableIRQ(pHandle->hISR);

	/************************************************************************
	 * Disable the interrupts at the UART.
	 ***********************************************************************/
	pReg->IER = UUEN|0;

	/************************************************************************
	 * Unregister the handler.
	 ***********************************************************************/
	RemoveHandler(pHandle->hISR);

    return STATUS_OK;
}

/***************************************************************************
* Function:	Write(tDRIVER *pHandle, UINT8 *cBuffer, unsigned int nBytes)
*
* Purpose:	Write a buffer out the SMC channel
*
* Returns:	DSTATUS
****************************************************************************/
static DSTATUS Write(tDRIVER *pHandle, UINT8 *cBuffer, unsigned int nBytes)
{
	volatile U550	*pReg = pHandle->pReg;

    while(nBytes)
	{
		/********************************************************************
		 * Wait for the transmit hold register to empty.
		 *******************************************************************/
		while(!(pReg->LSTAT & LSR_THRE))
	    	;

		/********************************************************************
		 * Buffer a character for transmitting.
		 *******************************************************************/
		pReg->TXDATA = *cBuffer++;
		--nBytes;
    }

    return STATUS_OK;
}

/***************************************************************************
* Function:	Ioctl(volatile PPC860_SMC *pHandle, UINT32 nFuncCode, void *pArgs)
*
* Purpose:	Set IO characteristics of the SMC channel
*
* Returns:	DSTATUS
****************************************************************************/
static DSTATUS Ioctl(tDRIVER *pHandle, UINT32 nFuncCode, void *pArgs)
{

	switch(nFuncCode)
	{
	case IOCTL_SETBAUD:
		pHandle->nBaudRate = *(UINT32 *)pArgs;
		setBaud(pHandle);
		return STATUS_OK;

	case IOCTL_GETBAUD:
		*(UINT32 *)pArgs = pHandle->nBaudRate;
  		return STATUS_OK;

	}
	return NOT_IMPLEMENTED;
}

/***************************************************************************
* Function:	Isr(tDRIVER *pHandle)
*
* Purpose:	ISR routine for the 16550
*
* Returns:	None
****************************************************************************/
static int Isr(tDRIVER *pHandle)
{
	volatile U550	*pReg = pHandle->pReg;
	volatile unsigned char tmp;

	/************************************************************************
	 * If no interrupt is pending then return immediately with a zero.
	 ***********************************************************************/
	if (pReg->IIR & 1)
		return 0;

	/************************************************************************
	 * Clear the interrupt by reading a character from the receive buffer.
	 ***********************************************************************/
    tmp = pReg->LSTAT;
	tmp = pReg->IIR;
    tmp = pReg->RXDATA;

	/************************************************************************
	 * Pass the received character to the core.
	 ***********************************************************************/
    EsdriverProcessCharacters((UINT8*)&tmp,(UINT32)1);

	/************************************************************************
	 * Return 1 to indicate the IRQ was serviced.
	 ***********************************************************************/
	return 1;
}
