/* shellhide.c - Shell variable hiding routines  */

/* Copyright 2002 Wind River Systems, Inc.  */

/*
modification history
--------------------
01a,15May02,CJW  Created.
01b,23May02,RCS  Changed attributes to hidden
*/

/*
DESCRIPTION
This module provides a way to hide or display shell variables to prevent 
user modification or viewing.  It implements the command, "shellhide", 
which allows a user to select which variables should be displayed when
a "shell" command is issued.

INCLUDE FILES:  shell.h
*/

/* includes */

#include "vm.h"
#include "shell.h"
#include "cmd.h"

/* defines */

#define HIDE_ATTR_CHANGED   1
#define HIDE_ATTR_UNCHANGED 0

/* globals */

extern UINT8 ShortForm;

/* locals */

static void *pShellVars = 0;
static BOOL exitFlag;
static BOOL hiddenFlag = FALSE;

/***************************************************************************
*
* shellPrintName - print the name of a shell variable
*
* This routine prints the name of a shell variable specified by its 
* structure pointer and element number.
*
* RETURNS: N/A
*/
static void shellPrintName
    (
    tShellVar *pSV,      /* Pointer to the variable to be printed */ 
    int element          /* element number of the variable */
    )
    {
    char buffer[40];

    sprintf(buffer, pSV->Name, element + 1);
    printf("%32s = ", buffer);
    }

/***************************************************************************
*
* hideShellEntry - Allow the user to hide or display this shell variable
*
* If the <shortForm> parameter is TRUE, this routine will display the
* current hidden attribute for the shell variable element.  If it is FALSE,
* the user will be prompted to specify the hidden attribute of the variable
* element.
*
* RETURNS: HIDE_ATTR_CHANGED, if the hidden attribute has changed, otherwise
*          HIDE_ATTR_UNCHANGED
*/
static UINT8 hideShellEntry
    (
    tShellVar *pSV,         /* Pointer to the shell variable structure */
    int element,            /* Element number within the variable */
    BOOL shortForm          /* Set to FALSE to allow changes  */
    )
    {
    char c;                 /* Character input by the user */
    
    hiddenFlag = FALSE;
    shellPrintName(pSV, element);
    printf("%s > ", (pSV->isHidden & HIDE_ELE_MASK(element) ? "Hide" : 
                                                              "Display"));

    /* Just return if we only need to display the values  */
    
    if (shortForm) 
        {
        printf("\n");
        if (pSV->isHidden & HIDE_ELE_MASK(element))
            hiddenFlag = TRUE;
        return HIDE_ATTR_UNCHANGED;
        }
        
    /*
     *  Wait for user input, checking for 'h', 'd' or return
     */
         
    for (;;)
        {
        c = tolower(getchar());
        
        if (c == 'h')
            {
            printf("Hide\n");
            hiddenFlag = TRUE;
            
            /* If the value was previously displayed, change it to hidden  */
            
            if (!(pSV->isHidden & HIDE_ELE_MASK(element))) 
                {
                pSV->isHidden |= HIDE_ELE_MASK(element);
                return HIDE_ATTR_CHANGED;
                }
            else
                return HIDE_ATTR_UNCHANGED;
            }
            
        if (c == 'd')
            {
            printf("Display\n");
            if (pSV->isHidden & HIDE_ELE_MASK(element))
                {
                pSV->isHidden &= ~HIDE_ELE_MASK(element);
                return HIDE_ATTR_CHANGED;
                } 
            else
                return HIDE_ATTR_UNCHANGED;
            }
            
        if (c == '\r') 
            {
            printf("\n");
            if (pSV->isHidden & HIDE_ELE_MASK(element))
                hiddenFlag = TRUE;
            return HIDE_ATTR_UNCHANGED;
            }
        }
    }


/***************************************************************************
*
* shShellHide - Allow the user to display/modify hidden attribute of variables
*
* This routine is executed by issuing the "shellhide" command from the 
* visionWare I/O window.  It may also be invoked with the short form of the 
* command, ".shellhide".  Using ".shellhide" will display the hidden
* attribute of all the shell variables, without allowing any modification by
* the user.  The standard form, "shellhide" will sequentially display each
* variable and element, allowing the user to change the settings as desired.
*
* Linked variables will not be displayed if the parent variable is hidden.
*
* Any variables that the user specified as hidden will not be shown through
* the "shell" or ".shell" commands.
*
* RETURNS: N/A
*/
void shShellHide
    (
    void *pIn,                     /* Input parameters from command line */
    void *pOut                     /* Output parameters                  */
    )
    {
    int         i;                 /* Index into shell var array */
    int         j;                 /* Index into linked variables */
    int         max = SHELLVARS;   /* Count of shell variables */
    int         element;           /* Index into element array */
    UINT32      elements;          /* Count of elements for current variable */
    universal   u;                 /* Pointer to current element of variable */
    universal   save;              /* Pointer to first element of variable */
    tShellVar * shellVars;         /* Pointer to shell variable values */
    UINT8       changes = HIDE_ATTR_UNCHANGED; /* Flag indicating whether or */
                                               /* not hidden attribute has   */
                                               /* has changed.               */
                                               
    if (ShortForm == 0) 
        {
        printf("Each of the shell variables will be displayed below");
        printf("\n");
        printf("with it's current hide/display setting.  Type \"h\" to");
        printf("\n");
        printf("hide the value, \"d\" to display it, or return to use");
        printf("\n");
        printf("the current value.");
        printf("\n");
        printf("\n");
        printf("\n");
        printf("\n");
        }
    
    /*
     * Upon return from this call pShellVars points to the start of 
     * a list of 32-bit aligned shell variable values.              
     */
     
    GetShellVarPointer(0);      /* force a read of the shell vars */
    u.pVoid = pShellVars;

    /* 
     * ShortForm is set if the user preceeds the command name
     * with a period.  In this instance, we use ShortForm so
     * that typing ".shellhide" will just display all the
     * settings, without allowing the user to change them.
     */

    exitFlag = ShortForm;     

    shellVars = (tShellVar *)__SHELLVAR_START;
    for (i = 0; i < max; i = j)
        {
            
        /* Get pointer to value of current shell variable */
        
        u.pVoid = GetShellVarPointer(*shellVars[i].pNumber);

        save.uvalue = u.uvalue;
        elements = shellVars[i].Elements;
        for (element = 0; element < elements; element++)
            {
            
            /* Display the entry to the user */
            
            changes |= hideShellEntry(&shellVars[i], element, ShortForm);
            
            /*
             * Point u to the next shellvar value, so that linked vars 
             * can be processed.                                       
             */
             
            u.uvalue = save.uvalue + (elements * shellVars[i].Size);
            
            /* Linked variables are always consecutive, so set j to i+1 */
            
            for (j = i + 1; j < max; j++)
                {
        
                /*
                 * If the variable isn't linked, then break out of   
                 * this loop so that we can process the next element 
                 * of the current variable (specified by i)          
                 */
                
                if (!(shellVars[j].Flags & SHELLFLAG_LINKED))
                    break;

                /* 
                 * If the last value was hidden, don't display the linked 
                 * variables 
                 */
                
                if (!hiddenFlag)
                    changes |= hideShellEntry(&shellVars[j], 
                                              element, 
                                              ShortForm);
                        
                /* Advanced to the next linked variable */
                
                u.uvalue += shellVars[j].Size * elements;
                }
            }   
        }
    
    /*
     * If any of the values have changed, write the new value to flash 
     * memory 
     */
    
    if (changes & HIDE_ATTR_CHANGED)
        WriteShellVars(pShellVars, u.uvalue - (UINT32)pShellVars);
    
}

    
Command(CMD_SHELLHIDE, "shellhide", shShellHide, HIDDEN, "", 
        "Hide/display shell settings");
CommandEnd();
