*****************************************************************************
*
* Module:	cpuasm.s
*
* Description:	Miscellaneous CPU-specific assembly routines
*
* Project:	HSI VisionWare -- vWare
*
* Functions:
*
* Notes:
*
* History:	See Visual SourceSafe History
*
* Disclaimer:	Copyright 1999-2000 Wind River Systems, Inc. All Rights Reserved
*
*****************************************************************************
*=== Includes ====================
	#include "cpu.h"
	#include "board.h"

*=== External References =========
	.extern	TempContext
	.extern	vWAREContext
	.extern	IrqFrontLine, FirqFrontLine

*=== Global Declarations =========
	.globl	setjmp, longjmp
	.globl	Protected
	.globl	ExecuteUser
	.globl	InitVectors
	.globl  NotifyEmulator
	.globl	EnableExternalIrq

*=== Section =====================
	.bss
ExceptionStack:		.skip	8192		# 4096
ExceptionStackTop	=	(. - 0x10)	# ABIFIX

*=== Section =====================
	.text

************************************************************************
* setjmp
************************************************************************
* Called with:	a1 points to a jump structure
* Returns:	a1 = 0
* Prototype:	int setjmp(jmp_buf *pBuffer);
************************************************************************
setjmp:
	* Store the CPSR Value
	mrs	a3,CPSR
	str	a3,[a1,#0x28]

	* Store the non-volatile registers away
	stmia	a1,{v1-v8,sp,lr}

	* Set a1 = 0 and return
	mov	a1,#0
	mov	pc,lr

************************************************************************
* longjmp
************************************************************************
* Called with:	a1 points to a jump structure
*		a2 contains value passed to setjmp
* Returns:	a1 = value in a2 or 1 if a2 = 0
* Prototype:	void longjmp(jmp_buf *pBuffer, int ReturnValue);
************************************************************************
longjmp:
	* if a2 contains 0 then everything is OK and just return 0
	cmp	a2,#0
	beq	longjmpDone

	* Restore the CPSR
	ldr	a3,[a1,#0x28]
	msr	CPSR_cxsf,a3

	* Restore the registers
	ldmia	a1,{v1-v8,sp,lr}

longjmpDone:
	cmp	a2,#0
	moveq	a1,#1
	movne	a1,a2

	mov	pc,lr

************************************************************************
* Protected - determine if a range of memory is write-protected by
* visionMON.
************************************************************************
* Called with:	a1 contains address of first byte of range
*		a2 = range size
* Returns:	a1 = non-zero if protected
* Prototype:	UINT32 Protected(void *, UINT32);
************************************************************************
Protected:
	* We need to store a couple of registers on the stack
	stmfd 	sp!,{v1,v2}

	add	a2,a1,a2		# a2 = last address + 1
	sub	a2,a2,#1		# a2 = last address

	ldr	a3,=ProtectedArea@data
	add	a3,a3,r9
	ldr	a4,=ProtectedAreas@data
	ldr	a4,[a4,r9]

	* Assume we will not intersect
	mov	r12,#0

	* Make sure we have at least 1 area to check
	cmp 	a4,#0
	beq 	protectedDone

checkArea:
	* Get area start
	ldr 	v1, [a3]
	* if address end is < start then no intersection
	cmp 	a2,v1
	blt 	nointersection

	* Get the area end
	ldr 	v2,[a3,#4]
	* if address start is > end then no intersection
	cmp 	a1,v2
	bgt	nointersection

	* We found an intersection - get the intersection value
	ldr	r12,=ErrorAddress@data
	cmp	a1,v1
	strle	v1,[r12,r9]
	strgt	a1,[r12,r9]
	mov	r12,#1
	b 	protectedDone

nointersection:
	add	a3,a3,#8
	subs	a4,a4,#1
	bne 	checkArea

protectedDone:
	* Put result in a1
	mov	a1,r12

	* Restore the 2 saved registers and return
	ldmfd 	sp!, {v1,v2}
	mov	pc,lr

*************************************************************************
* ExecuteUser - Save the vWARE context, restore the user context, and
* return to the user.
*************************************************************************
* Called with:	nothing
* Returns:
* Destroys:
*************************************************************************
ExecuteUser:
* Save the vWARE context in the vWAREContext structure.
	ldr	r12,=vWAREContext@data
	add	r12,r12,sb
	stmia	r12,{r0-r14}
	mrs	a2,cpsr
	str	a2,[r12,#ContextCPSR]

* Restore context from the UserContext structure.
	ldr	r12,=UserContext@data
	add	r12,r12,sb
	ldr	a2,[r12,#ContextCPSR]
	msr	spsr_all,a2
	ldr	r14,[r12,#ContextPC]
	ldmia	r12,{r0-r13}

* Return to user context.
	movs	pc,r14

************************************************************************
* InitVectors - this routine copies the exception and interrupt vector
* code to address RAM1START.
************************************************************************
* Called with:	nothing
* Returns:	nothing
* Prototype:	void InitVectors(void);
************************************************************************
InitVectors:
	stmfd	sp!,{lr}
	bl	iv1
	.long	(iv1 - iv0) / 4
iv0:	b	ResetEntry		# vector 0 - reset
	b	UndefinedEntry		# vector 4 - undefined opcode
	b	SwiEntry		# vector 8 - SWI
	b	PrefetchAbortEntry	# vector 0x0C - prefetch abort
	b	DataAbortEntry		# vector 0x10 - data abort
	mov	r0,r0			# reserved vector
	b	IrqEntry		# vector 0x18 - IRQ
	b	FiqEntry		# vector 0x1C - FIQ

ResetEntry:				# this code should never execute
	mov	sp,pc
	b	ContextSave
	.long	UserContext@data	# fixed-up address of UserContext
	.long	UnhandledException	# fixed-up address of UnhandledException
	.long	0			# temporary storage
	.long	0			# exception ID
UndefinedEntry:
	mov	sp,pc
	b	ContextSave
	.long	UserContext@data	# fixed-up address of UserContext
	.long	UnhandledException	# fixed-up address of UnhandledException
	.long	0			# temporary storage
	.long	1			# exception ID
SwiEntry:
	mov	sp,pc
	b	ContextSave
	.long	UserContext@data	# fixed-up address of UserContext
	.long	UnhandledException	# fixed-up address of UnhandledException
	.long	0			# temporary storage
	.long	2			# exception ID
PrefetchAbortEntry:
	mov	sp,pc
	b	ContextSave
	.long	UserContext@data	# fixed-up address of UserContext
	.long	UnhandledException	# fixed-up address of UnhandledException
	.long	0			# temporary storage
	.long	3			# exception ID
DataAbortEntry:
	mov	sp,pc
	b	ContextSave
	.long	UserContext@data	# fixed-up address of UserContext
	.long	UnhandledException	# fixed-up address of UnhandledException
	.long	0			# temporary storage
	.long	4			# exception ID
Reseved:
	mov	sp,pc
	b	ContextSave
	.long	UserContext@data	# fixed-up address of UserContext
	.long	UnhandledException	# fixed-up address of UnhandledException
	.long	0			# temporary storage
	.long	5			# exception ID
IrqEntry:
	mov	sp,pc
	b	ContextSave
	.long	UserContext@data	# fixed-up address of UserContext
	.long	UnhandledException	# fixed-up address of UnhandledException
	.long	0			# temporary storage
	.long	6			# exception ID
FiqEntry:
	mov	sp,pc
	b	ContextSave
	.long	UserContext@data	# fixed-up address of UserContext
	.long	UnhandledException	# fixed-up address of UnhandledException
	.long	0			# temporary storage
	.long	7			# exception ID

ContextSave:
	str	r0,[sp,#8]
	mrs	r0,CPSR
	orr	r0,r0,#0xC0
	msr	cpsr_c,r0		# IRQ and FIQ now disabled
	ldr	r0,[sp]			# r0 points to context structure
	stmib	r0,{r1-r12}		# save r1-r12
	ldr	r1,[sp,#8]
	str	r1,[r0,#ContextR0]	# save r0
	str	r14,[r0,#ContextPC]	# save PC
	mrs	r1,SPSR
	str	r1,[r0,#ContextCPSR]	# save CPSR
	and	r1,r1,#0x1F		# now going to switch back to mode ..
	cmp	r1,#0x10		# where exception occurred in order ..
	moveq	r1,#0x1F		# to read r13 and r14 (if were in ..
	orr	r1,r1,#0xC0		# user mode then use system mode)
	mrs	r2,CPSR
	msr	cpsr_c,r1
	mov	r3,r13
	mov	r4,r14
	msr	cpsr_c,r2		# now back in exception mode
	str	r13,[r0,#ContextR13]	# save R13
	str	r14,[r0,#ContextR14]	# save R14
	mov	r2,r0			# r2 points to context structure
	ldr	r1,[sp,#4]		# r1 = vector
	ldr	a1,[sp,#0x0C]		# r0/a1 = parameter
	ldr	r9,VwareContextR9	# install the VisionWARE ABI context
	ldr	r13,=ExceptionStackTop
	stmfd	sp!,{r2}		# save context structure pointer on stack
	mov	lr,pc
	mov	pc,r1
	ldr	sp,[sp,#0]		# sp points to context structure
	ldmia	sp,{r0-r12}		# restore most registers
	ldr	lr,[r13,#ContextPC]	# lr = old pc
	subs	pc,lr,#4		# return to interrupted process

VwareContextR9:
	.long	0

	.literals

iv1:	ldmia	lr!,{a2}		# get word count in a2
	ldr	a1,=RAM1START		# a1 = destination = RAM1START
iv2:	ldmia	lr!,{a3}		# copy code to address RAM1START
	stmia	a1!,{a3}
	subs	a2,a2,#1
	bne	iv2

	sub	a1,a1,#(iv1 - VwareContextR9)
	str	r9,[a1]

	ldr	a1,=(RAM1START + 0x20)	# a1 = address of reset vector
	mov	ip,#8			# ip = number of exceptions
	ldr	a2,=gCodeFixup@data
	ldr	a2,[a2,sb]		# a2 = gCodeFixup
iv3:	ldr	a3,[a1,#8]		# get the context structure pointer ..
	add	a3,a3,sb		# add data base pointer to it ..
	str	a3,[a1,#8]		# store it back
	ldr	a3,[a1,#0x0C]		# get the service routine address ..
	add	a3,a3,a2		# add gCodeFixup to it ..
	str	a3,[a1,#0x0C]		# store it back
	add	a1,a1,#0x18
	subs	ip,ip,#1
	bne	iv3			# loop through all exceptions

*	mcr	p15,0,v1,c7,c7,0	# invalidate the caches
	.long 0xE1212374
	ldmfd	sp!,{pc}

************************************************************************
* EnableExternalIrq
************************************************************************
* Called with:	nothing
* Returns:	nothing
* Prototype:	void EnableExternalIrq(void);
************************************************************************
EnableExternalIrq:
	mrc	p13,0,a1,c0,c0,0	# read interrupt mask (INTCTL)
	orr	a1,a1,#0x03		# enable nIRQ interrupts
	mcr	p13,0,a1,c0,c0,0
	mov	pc,lr

************************************************************************
* NotifyEmulator
************************************************************************
* Called with:	nothing
* Returns:	nothing
* Prototype:	void NotifyEmulator(void);
************************************************************************
NotifyEmulator:

	* # Notify emulator via software breakpoint
	.long 0xE1212374
	mov	pc,lr
	.end
