/******************************************************************************
*
* Module: vdmemfill.c
*
* Description: Fill an area of memory with a known pattern and verify
*
* Project:     HSI VisionWare -- vWare
*
* Functions:
*
* Notes:
*
* History:		See Visual SourceSafe History
*
* Disclaimer: Copyright 1999-2000 Wind River Systems, Inc. All Rights Reserved
*
*******************************************************************************/


//  Includes
#include "vm.h"
#include "cmd.h"
#include "vderror.h"

//  Local Defines
#define Pass 0
#define Fail 1
#define ExecutionFail 0xffffffff

typedef struct _memfillIn
{
	UINT32  lWidth;
	UINT32	lStartAddress;
	UINT32  lEndAddress;
	UINT32	lPattern;	
} memfillIn;

typedef struct _memfillOut
{
	UINT32 lPassFail;
	UINT32 lReturnCode;
	char   cResultString[128];
} memfillOut;


UINT32 memfill(memfillIn* pIn, memfillOut* pOut)
{
	UINT32 lCount = pIn->lStartAddress;

	universal p;

	p.uvalue = pIn->lStartAddress;

	pOut->lPassFail = Pass;

	//see if the size was entered, if not, default to 4
	if(pIn->lWidth !=0x4)
	{
		if(pIn->lWidth !=0x2)
		{
			if(pIn->lWidth !=0x1)
			{
				pIn->lWidth = 0x4;
			}
		}
	}

	//Ensure Start address < End Address 					
	if ( pIn->lEndAddress < pIn->lStartAddress ) 
	{
		pOut->lPassFail = Fail;
		pOut->lReturnCode = ExecutionFail;
		LogError(ERR_MEMFILL, CONT_ON_ERROR|LOG_TO_FILE|LOG_TO_CONSOLE, "ERROR: MEMFILL, START ADDRESS > END ADDRESS\n");
		strcpy(pOut->cResultString,"ERROR: MEMFILL, START ADDRESS > END ADDRESS\n");
		return(Fail);
	}

	//Ensure Start address != End Address
	if(pIn->lEndAddress == pIn->lStartAddress)
	{
		pOut->lPassFail = Fail;
		pOut->lReturnCode = ExecutionFail;
		LogError(ERR_MEMFILL, CONT_ON_ERROR|LOG_TO_FILE|LOG_TO_CONSOLE, "ERROR: MEMFILL, START ADDRESS = END ADDRESS\n");
		strcpy(pOut->cResultString,"ERROR: MEMFILL, START ADDRESS = END ADDRESS\n");
		return(Fail);
	}

	
	//Check allignment of End Address
    if(pIn->lWidth!=0x1)
	{
		if(pIn->lEndAddress % pIn->lWidth!=0x0)
		{
			pOut->lPassFail = Fail;
			pOut->lReturnCode = Fail;
			LogError(ERR_MEMFILL, CONT_ON_ERROR|LOG_TO_FILE|LOG_TO_CONSOLE, 
				"ERROR: MEMFILL, END ADDRESS ALLIGNMENT ERROR\n");
			strcpy(pOut->cResultString, "ERROR: MEMFILL, END ADDRESS ALLIGNMENT ERROR\n");
			return(pOut->lPassFail);
		}
	}

	//check to see if any of the range to fill to is protected 
	if(Protected(pIn->lStartAddress,(pIn->lEndAddress - pIn->lStartAddress))!=0x0)
	{
		pOut->lPassFail = Fail;
		pOut->lReturnCode = Fail;
		LogError(ERR_MEMFILL, CONT_ON_ERROR|LOG_TO_FILE|LOG_TO_CONSOLE, 
			"ERROR: MEMFILL, EITHER SOME OR ALL OF THE AREA TO BE WRITTEN TO IS PROTECTED\n");
		strcpy(pOut->cResultString, "ERROR: MEMFILL, EITHER ALL OR SOME OF THE AREA TO BE WRITTEN TO IS PROTECTED\n");
		return(pOut->lPassFail);
	}


	//Perform fill and validate function based on width
	switch (pIn->lWidth)
	{
		case 1: for(; (UINT32 *)p.pByte < (UINT32 *)pIn->lEndAddress; p.pByte++)
				{
					*p.pByte = (UINT8)pIn->lPattern;
				}
				for(p.uvalue = pIn->lStartAddress; (UINT32 *)p.pByte < (UINT32 *)pIn->lEndAddress; p.pByte++)
				{
					if(*p.pByte != (UINT8)pIn->lPattern)
					{
						pOut->lPassFail = Fail;
						pOut->lReturnCode = Fail;
						LogError(ERR_MEMFILL, CONT_ON_ERROR|LOG_TO_FILE|LOG_TO_CONSOLE, 
							"ERROR: MEMFILL, DATA VERIFY ERROR AT 0x%X, EXPECTED 0x%X, READ 0x%X\n", p.pByte,(UINT8)pIn->lPattern, (UINT8 *)*p.pByte);
						sprintf(pOut->cResultString, "ERROR: MEMFILL, DATA VERIFY ERROR AT 0x%X, EXPECTED 0x%X, READ 0x%X\n", p.pByte,(UINT8)pIn->lPattern, *p.pByte);
						p.pByte = (UINT8 *)pIn->lEndAddress;
					}
				}
				break;

		case 2:	for(; (UINT32 *)p.pHalf < (UINT32 *)pIn->lEndAddress; p.pHalf++)
				{
					*p.pHalf = (UINT16)pIn->lPattern;
				}
				for(p.uvalue = pIn->lStartAddress; (UINT32 *)p.pHalf < (UINT32 *)pIn->lEndAddress; p.pHalf++)
				{
					if(*p.pHalf != (UINT16)pIn->lPattern)
					{
						pOut->lPassFail = Fail;
						pOut->lReturnCode = Fail;
						LogError(ERR_MEMFILL, CONT_ON_ERROR|LOG_TO_FILE|LOG_TO_CONSOLE, 
							"ERROR: MEMFILL, DATA VERIFY ERROR AT 0x%X, EXPECTED 0x%X, READ 0x%X\n", p.pHalf,(UINT16)pIn->lPattern, *p.pHalf);
						sprintf(pOut->cResultString, "ERROR: MEMFILL, DATA VERIFY ERROR AT 0x%X, EXPECTED 0x%X, READ 0x%X\n", p.pHalf,(UINT16)pIn->lPattern, *p.pHalf);
						p.pHalf = (UINT16 *)pIn->lEndAddress;
					}
				}
				break;

		case 4:	for(; (UINT32 *)p.pWord < (UINT32 *)pIn->lEndAddress; p.pWord++)
				{
					*p.pWord = (UINT32)pIn->lPattern;
				}
				for(p.uvalue = pIn->lStartAddress; (UINT32 *)p.pWord < (UINT32 *)pIn->lEndAddress; p.pWord++)
				{
					if(*p.pWord != (UINT32)pIn->lPattern)
					{
						pOut->lPassFail = Fail;
						pOut->lReturnCode = Fail;
						LogError(ERR_MEMFILL, CONT_ON_ERROR|LOG_TO_FILE|LOG_TO_CONSOLE, 
							"ERROR: MEMFILL, DATA VERIFY ERROR AT 0x%X, EXPECTED 0x%X, READ 0x%X\n", p.pWord,(UINT32)pIn->lPattern, *p.pWord);
						sprintf(pOut->cResultString, "ERROR: MEMFILL, DATA VERIFY ERROR AT 0x%X, EXPECTED 0x%X, READ 0x%X\n", p.pWord,(UINT32)pIn->lPattern, *p.pWord);
						p.pWord = (UINT32 *)pIn->lEndAddress;
					}
				}
				break;
		}

	if(pOut->lPassFail == Pass)
	{
		pOut->lPassFail = Pass;
		pOut->lReturnCode = Pass;
		Info("Success, MEMFILL of address range 0x%X ... 0x%X completed and verified\n", pIn->lStartAddress, pIn->lEndAddress);
	}
						
	return(pOut->lPassFail);
}

Command(CMD_MEMFILL,"memfill",memfill,SIZESUFFIX | DIAGNOSTIC,"memfill","fill an area of memory and verify");
Arg(UINT32ARG,"start address","enter start address");
Arg(UINT32ARG,"end address","enter end address");
Arg(UINT32ARG,"pattern","pattern");
CommandEnd();


