
#include "vm.h"
#include "cmd.h"

#include "vdxscale_mmudiag.h"

/***************************************************************************
* Function: ParseCoarsePageTable(tMMUAtributesForMemoryPage *pOut)
*
* Purpose:  Parse the MMU coarse page table entry. The FirstLevelDescriptor
*			must be filled out, and for a coarse page table. Fills in data for
*			PhysicalAddress, SecondLevelDescriptor, StartOfPage, EndOfPage,
*			Attributes, and xcbbits.
*
* Returns:  Fault Code -- 0 is no fault
****************************************************************************/
UINT32 ParseCoarsePageTable(tMMUAtributesForMemoryPage *pOut)
{
	unsigned char format;
	UINT32  SecondLevelTableIndex;
	UINT32	*AddressOfSecondLevelDescriptor;
	UINT32	PageTableBaseAddress, PageIndex;

	if((pOut->FirstLevelDescriptor & 0x03) != 0x01)
	{
		// this is not a coarse page table
		pOut->Fault = FAULT_IN_COARSE_PAGE_TABLE_DESCRIPTOR;
		return(FAULT_IN_COARSE_PAGE_TABLE_DESCRIPTOR);
	}
	SecondLevelTableIndex = (( pOut->VirtualAddress & 0x000FF000 ) >> 10 );
	PageTableBaseAddress = ( pOut->FirstLevelDescriptor & 0xFFFFFC00 );
	AddressOfSecondLevelDescriptor = (UINT32 *)( PageTableBaseAddress | SecondLevelTableIndex );
	pOut->SecondLevelDescriptor = *AddressOfSecondLevelDescriptor;

	format = ( pOut->SecondLevelDescriptor & 0x03 );
	switch ( format )
	{
		case 0: // fault
			pOut->Fault = FAULT_IN_SECOND_LEVEL_DESCRIPTOR;
			return(FAULT_IN_SECOND_LEVEL_DESCRIPTOR);
			break;
		case 1: // Large Page
			PageIndex = ( pOut->VirtualAddress & 0x0000FFFF );
			pOut->PhysicalAddress = ( pOut->SecondLevelDescriptor & 0xFFFF0000) | PageIndex;
			pOut->Attributes = ( pOut->SecondLevelDescriptor & 0x0000FFFF );
			pOut->StartOfPage = ( pOut->SecondLevelDescriptor & 0xFFFF0000);
			pOut->EndOfPage = ( pOut->SecondLevelDescriptor | 0x0000FFFF);
			pOut->xcbbits = (pOut->SecondLevelDescriptor & 0x00001000) >> 10;
			pOut->xcbbits |= (pOut->SecondLevelDescriptor & 0x0000000C) >> 2;
			break;
		case 2: // Small Page
			PageIndex = ( pOut->VirtualAddress & 0x00000FFF );
			pOut->PhysicalAddress = ( pOut->SecondLevelDescriptor & 0xFFFFF000) | PageIndex;
			pOut->Attributes = ( pOut->SecondLevelDescriptor & 0x00000FFF );
			pOut->StartOfPage = ( pOut->SecondLevelDescriptor & 0xFFFFF000);
			pOut->EndOfPage = ( pOut->SecondLevelDescriptor | 0x00000FFF);
			// there is no x bit for a small page, always = 0
			pOut->xcbbits = (pOut->SecondLevelDescriptor & 0x0000000C) >> 2;
			break;
		case 3: // Extended Small Page
			PageIndex = ( pOut->VirtualAddress & 0x00000FFF );
			pOut->PhysicalAddress = ( pOut->SecondLevelDescriptor & 0xFFFFF000) | PageIndex;
			pOut->Attributes = ( pOut->SecondLevelDescriptor & 0x00000FFF );
			pOut->StartOfPage = ( pOut->SecondLevelDescriptor & 0xFFFF0000);
			pOut->EndOfPage = ( pOut->SecondLevelDescriptor | 0x0000FFFF);
			pOut->xcbbits = (pOut->SecondLevelDescriptor & 0x00000040) >> 4;
			pOut->xcbbits |= (pOut->SecondLevelDescriptor & 0x0000000C) >> 2;
			break;
		default: // There was an Error in This code if it got here
			pOut->Fault = FAULT_IN_PARSING_SECOND_LEVEL_DESCRIPTOR;
			return(FAULT_IN_PARSING_SECOND_LEVEL_DESCRIPTOR);
			break;
	}
	return(0);
}
/***************************************************************************
* Function: ParseSectionTable(tMMUAtributesForMemoryPage *pOut)
*
* Purpose:  Parse the MMU coarse page table entry. The FirstLevelDescriptor
*			must be filled out, and for a section table. Fills in data for
*			PhysicalAddress, SecondLevelDescriptor, StartOfPage, EndOfPage,
*			Attributes, and xcbbits.
*
* Returns:  Fault Code -- 0 is no fault
****************************************************************************/
UINT32 ParseSectionTable(tMMUAtributesForMemoryPage *pOut)
{
	UINT32	SectionIndex;

	if((pOut->FirstLevelDescriptor & 0x03) != 0x02)
	{
		// this is not a section table
		pOut->Fault = FAULT_IN_SECTION_DESCRIPTOR;
		return(FAULT_IN_SECTION_DESCRIPTOR);
	}
	pOut->SecondLevelDescriptor = 0; // there is no second level descriptor
	SectionIndex = ( pOut->VirtualAddress & 0x000FFFFF);
	pOut->PhysicalAddress = ( pOut->FirstLevelDescriptor & 0xFFF00000) | SectionIndex;
	pOut->Attributes = (pOut->FirstLevelDescriptor & 0x00001FFF);	
	pOut->StartOfPage = ( pOut->FirstLevelDescriptor & 0xFFF00000);
	pOut->EndOfPage = ( pOut->FirstLevelDescriptor | 0x000FFFFF);
	pOut->xcbbits = (pOut->FirstLevelDescriptor & 0x00001000) >> 10;
	pOut->xcbbits |= (pOut->FirstLevelDescriptor & 0x0000000C) >> 2;
	return(0);
}
/***************************************************************************
* Function: ParseFinePageTable(tMMUAtributesForMemoryPage *pOut)
*
* Purpose:  Parse the MMU coarse page table entry. The FirstLevelDescriptor
*			must be filled out, and for a fine page table. Fills in data for
*			PhysicalAddress, SecondLevelDescriptor, StartOfPage, EndOfPage,
*			Attributes, and xcbbits.
*
* Returns:  Fault Code -- 0 is no fault
****************************************************************************/
UINT32 ParseFinePageTable(tMMUAtributesForMemoryPage *pOut)
{
	unsigned char format;
	UINT32  SecondLevelTableIndex;
	UINT32	*AddressOfSecondLevelDescriptor;
	UINT32	PageTableBaseAddress, PageIndex;

	if((pOut->FirstLevelDescriptor & 0x03) != 0x03)
	{
		// this is not a fine page table
		pOut->Fault = FAULT_IN_FINE_PAGE_TABLE_DESCRIPTOR;
		return(FAULT_IN_FINE_PAGE_TABLE_DESCRIPTOR);
	}
	
	SecondLevelTableIndex = (( pOut->VirtualAddress & 0x000FF000 ) >> 10 );
	PageTableBaseAddress = ( pOut->FirstLevelDescriptor & 0xFFFFF000 );
//	2 bits are always going to be zero (10:11)
	AddressOfSecondLevelDescriptor = (UINT32 *)( PageTableBaseAddress | SecondLevelTableIndex );
	pOut->SecondLevelDescriptor = *AddressOfSecondLevelDescriptor;

	format = ( pOut->SecondLevelDescriptor & 0x03 );
	switch ( format )
	{
		case 0: // fault
			pOut->Fault = FAULT_IN_SECOND_LEVEL_DESCRIPTOR;
			return(FAULT_IN_SECOND_LEVEL_DESCRIPTOR);
			break;
		case 1: // Large Page
			PageIndex = ( pOut->VirtualAddress & 0x0000FFFF );
			pOut->PhysicalAddress = ( pOut->SecondLevelDescriptor & 0xFFFF0000) | PageIndex;
			pOut->Attributes = ( pOut->SecondLevelDescriptor & 0x0000FFFF );
			pOut->StartOfPage = ( pOut->SecondLevelDescriptor & 0xFFFF0000);
			pOut->EndOfPage = ( pOut->SecondLevelDescriptor | 0x0000FFFF);
			pOut->xcbbits = (pOut->SecondLevelDescriptor & 0x00001000) >> 10;
			pOut->xcbbits |= (pOut->SecondLevelDescriptor & 0x0000000C) >> 2;
			break;
		case 2: // Small Page
			PageIndex = ( pOut->VirtualAddress & 0x00000FFF );
			pOut->PhysicalAddress = ( pOut->SecondLevelDescriptor & 0xFFFFF000) | PageIndex;
			pOut->Attributes = ( pOut->SecondLevelDescriptor & 0x00000FFF );
			pOut->StartOfPage = ( pOut->SecondLevelDescriptor & 0xFFFFF000);
			pOut->EndOfPage = ( pOut->SecondLevelDescriptor | 0x00000FFF);
			// there is no x bit for a small page, always = 0
			pOut->xcbbits = (pOut->SecondLevelDescriptor & 0x0000000C) >> 2;
			break;
		case 3: // Tiny Page
			PageIndex = ( pOut->VirtualAddress & 0x000003FF );
			pOut->PhysicalAddress = ( pOut->SecondLevelDescriptor & 0xFFFFFC00) | PageIndex;
			pOut->Attributes = ( pOut->SecondLevelDescriptor & 0x000003FF );
			pOut->StartOfPage = ( pOut->SecondLevelDescriptor & 0xFFFFFC00);
			pOut->EndOfPage = ( pOut->SecondLevelDescriptor | 0x000003FF);
			pOut->xcbbits = (pOut->SecondLevelDescriptor & 0x00000040) >> 4;
			pOut->xcbbits |= (pOut->SecondLevelDescriptor & 0x0000000C) >> 2;
			break;
		default: // There was an Error in This code if it got here
			pOut->Fault = FAULT_IN_PARSING_SECOND_LEVEL_DESCRIPTOR;
			return(FAULT_IN_PARSING_SECOND_LEVEL_DESCRIPTOR);
			break;
	}
	return(0);
}
/***************************************************************************
* Function: GetAuxControlReg(void)
*
* Purpose:  Get auxilliary control register value.
*
* Returns:  auxilliary control register value.
****************************************************************************/
UINT32 GetAuxControlReg(void)
{
	asm ("		mrc   p15, 0, r0, c1, c0, 1	");
}
/***************************************************************************
* Function: GetGlobalMemorySettings(tMMUGlobalAtributes *pOut)
*
* Purpose:  Fill out TTBase, MMUEnabled, DataCacheEnabled, and InstructionCacheEnabled.
*
* Returns:  Fault Code -- 0 is no fault
****************************************************************************/
UINT32 GetGlobalMemorySettings(tMMUGlobalAtributes *pOut)
{
	pOut->Fault = 0;

	pOut->TTBase = GetTTBase();
	pOut->MMUEnabled = IsMMUEnabled();
	pOut->DataCacheEnabled = IsDataCacheEnabled();
	// pOut->InstructionCacheEnabled = !!!; // complete during instruction cache phase.
	pOut->AuxControlRegValue = GetAuxControlReg();
	return(0);
}
/***************************************************************************
* Function: GetMMUDescriptors(tMMUGlobalAtributes *pIn, tMMUAtributesForMemoryPage *pOut)
*
* Purpose:  Get the FirstLevelDescriptor and SecondLevelDescriptor for Address,
*			TTBase and VirtualAddress must be filled out before calling!
*
* Returns:  Fault Code -- 0 is no fault
****************************************************************************/
UINT32 GetMMUDescriptors(tMMUGlobalAtributes *pIn, tMMUAtributesForMemoryPage *pOut)
{
	UINT32 TableIndex;
	UINT32 *AddressOfDescriptor;
	UINT32	PageTableBaseAddress, PageIndex;
	unsigned char format;

	pOut->Fault = 0;

	TableIndex = (( pOut->VirtualAddress & 0xFFF00000) >> 18);
	AddressOfDescriptor = (UINT32 *)( pIn->TTBase | TableIndex);
	pOut->FirstLevelDescriptor = *AddressOfDescriptor;

	TableIndex = (( pOut->VirtualAddress & 0x000FF000 ) >> 10 );
	format = (unsigned char)( pOut->FirstLevelDescriptor & 0x03 );
	switch ( format )
	{
		case 0: // fault
			pOut->Fault = FAULT_IN_FIRST_LEVEL_DESCRIPTOR;
			return(FAULT_IN_FIRST_LEVEL_DESCRIPTOR);
			break;
		case 1: // Coarse Page Table
			PageTableBaseAddress = ( pOut->FirstLevelDescriptor & 0xFFFFFC00 );
			AddressOfDescriptor = (UINT32 *)( PageTableBaseAddress | TableIndex );
			pOut->SecondLevelDescriptor = *AddressOfDescriptor;
			break;
		case 2: // Section Base Address
			pOut->SecondLevelDescriptor = 0; // there is no second level descriptor
			break;
		case 3: // Fine Page Table
			PageTableBaseAddress = ( pOut->FirstLevelDescriptor & 0xFFFFF000 );
			//	2 bits are always going to be zero (10:11)
			AddressOfDescriptor = (UINT32 *)( PageTableBaseAddress | TableIndex );
			pOut->SecondLevelDescriptor = *AddressOfDescriptor;
			break;
		default: // There was an Error in This code if it got here
			pOut->Fault = FAULT_IN_PARSING_FIRST_LEVEL_DESCRIPTOR;
			return(FAULT_IN_PARSING_FIRST_LEVEL_DESCRIPTOR);
			break;
	}
	return(0);
}
/***************************************************************************
* Function: ParseMMUDescriptors(tMMUAtributesForMemoryPage *pOut)
*
* Purpose:  Parse the FirstLevelDescriptor and SecondLevelDescriptor if Fault = 0
*				filling	out the	rest of the tMMUAtributesForMemoryPage table.
*
* Returns:  Fault Code -- 0 is no fault
****************************************************************************/
UINT32 ParseMMUDescriptors(tMMUAtributesForMemoryPage *pOut)
{
	unsigned char buffer[84];
	unsigned char *pc;
	unsigned char format;

	if(pOut->Fault != 0)
	{
		// thre was some sort of fault
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);
		pc += sprintf((char *) buffer, "Fault in ParseMMUDescriptors");
		more("%s\n", buffer);
		return(pOut->Fault);
	}
	format = (unsigned char)( pOut->FirstLevelDescriptor & 0x03 );
	switch ( format )
	{
		case 0: // fault
			pOut->Fault = FAULT_IN_FIRST_LEVEL_DESCRIPTOR;
			return(FAULT_IN_FIRST_LEVEL_DESCRIPTOR);
			break;
		case 1: // Coarse Page Table
			//PrintSettingsForCoarsePageTable(VirtualAddress,FirstLevelDescriptor);
			ParseCoarsePageTable(pOut);
			break;
		case 2: // Section Base Address
			//PrintSettingsForSectionTable(VirtualAddress,FirstLevelDescriptor);
			ParseSectionTable(pOut);
			break;
		case 3: // Fine Page Table
			//PrintSettingsForFinePageTable(VirtualAddress,FirstLevelDescriptor);
			ParseFinePageTable(pOut);
			break;
		default: // There was an Error in This code if it got here
			pOut->Fault = FAULT_IN_PARSING_FIRST_LEVEL_DESCRIPTOR;
			return(FAULT_IN_PARSING_FIRST_LEVEL_DESCRIPTOR);
			break;
	}
	return(0);
}
/***************************************************************************
* Function: GetMemorySettingForAddress(UINT32 Address,tMMUGlobalAtributes GlobalMemorySettings,
*										 tMMUAtributesForMemoryPage *pOut)
*
* Purpose:  Fill out the tMMUAtributesForMemoryPage table for Address.
*
* Returns:  Fault Code -- 0 is no fault
****************************************************************************/
UINT32 GetMemorySettingForAddress(UINT32 Address, tMMUGlobalAtributes *GlobalMemorySettings,
								  tMMUAtributesForMemoryPage *pOut)
{
	UINT32 ReturnValue;

	pOut->VirtualAddress = Address;
	GetMMUDescriptors(GlobalMemorySettings, pOut);
	ReturnValue = ParseMMUDescriptors(pOut);
	return(ReturnValue);
}

#if 0
/***************************************************************************
* Function: FindIntersectionOfRanges(tUINT32Range range1, tUINT32Range range2, tUINT32Range out)
*
* Purpose:  Find the intersection of 2 memory ranges.
*			If there is no intesection out.Start = out.End = 0
*
* Returns:	0 for no intersection, 1 if there is an intersection
****************************************************************************/
UINT32 FindIntersectionOfRanges(tUINT32Range *range1, tUINT32Range *range2, tUINT32Range *out)
{
	UINT32 TempForSwap;
	
	// just in case start and end are not in order test and swap here.
	if(range1->Start > range1->End)
	{
		TempForSwap = range1->Start;
		range1->Start = range1->End;
		range1->End = TempForSwap;
	}
	if(range2->Start > range2->End)
	{
		TempForSwap = range2->Start;
		range2->Start = range2->End;
		range2->End = TempForSwap;
	}
	// start and end are now in order
	// start and end of range 1 are before range 2
	if(range1->End <= range2->Start)
	{
		out->Start = 0;
		out->End = 0;
		return(0);
	}
	// start and end of range 1 are after range 2
	if(range1->Start >= range2->End)
	{
		out->Start = 0;
		out->End = 0;
		return(0);
	}
	// start and end of range 1 are within range 2
	if((range1->Start >= range2->Start) & (range1->End <= range2->End))
	{
		out->Start = range1->Start;
		out->End = range1->End;
		return(1);
	}
	// range 1 starts before range 2, and ends after range 2
	if((range1->Start <= range2->Start) & (range1->End >= range2->End))
	{
		out->Start = range2->Start;
		out->End = range2->End;
		return(1);
	}
	// range 1 starts before range 2, and ends in range 2
	if((range1->Start <= range2->Start) & (range1->End <= range2->End))
	{
		out->Start = range2->Start;
		out->End = range1->End;
		return(1);
	}
	// range 1 starts in range 2, and ends after range 2
	if((range1->Start >= range2->Start) & (range1->End >= range2->End))
	{
		out->Start = range1->Start;
		out->End = range2->End;
		return(1);
	}
}
#endif
/***************************************************************************
* Function: GetPhysicalMemory(tPhysicalMemoryList *pOut)
*
* Purpose:  Gets the physical memory configuration from Yavapai.
* 2 SDRAM segments must be contiguous, 2 flash segments don't have to be contiguous
* 1 segment is for Yavapai registers.
* This routine will return "NumberOfSegments" valid segments first in the arrays,
* remaining data in the arrays is undefined.
*
* Returns:	Nothing
****************************************************************************/
void GetPhysicalMemory(tPhysicalMemoryList *pOut, unsigned char Verbose)
{
	UINT32 i,temp;
	unsigned char buffer[84];
	unsigned char *pc;

	pOut->StartAddress[0] = YAVAPAI.MCU.SDBR;
	pOut->EndAddress[0]   = pOut->StartAddress[0] + ((YAVAPAI.MCU.SBR0 & SDRAM_SIZE_MASK) << 22) - 1;
	pOut->StartAddress[1] = pOut->EndAddress[0] + 1;
	pOut->EndAddress[1]   = pOut->StartAddress[0] + ((YAVAPAI.MCU.SBR1 & SDRAM_SIZE_MASK) << 22) - 1;
	
	pOut->StartAddress[2] = YAVAPAI.MCU.FEBR0;
	pOut->EndAddress[2] = pOut->StartAddress[2];
	temp = YAVAPAI.MCU.FBSR0;
	if(temp >= 8)
	{
		pOut->EndAddress[2] += (0x8000 << 8) - 1;
	}else if(temp == 0)
	{
		pOut->EndAddress[2] = 0;
	}else
	{
		pOut->EndAddress[2] += (0x8000 << temp) - 1;
	}
	pOut->StartAddress[3] = YAVAPAI.MCU.FEBR1;
	pOut->EndAddress[3] = pOut->StartAddress[3];
	temp = YAVAPAI.MCU.FBSR1;
	if(temp >= 8)
	{
		pOut->EndAddress[3] += (0x8000 << 8) - 1;
	}else if(temp == 0)
	{
		pOut->EndAddress[3] = 0;
	}else
	{
		pOut->EndAddress[3] += (0x8000 << temp) - 1;
	}
	if(Verbose)
	{
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);
		pc += sprintf((char *) buffer, "Physical Memory settings from the 80312\n");
		more("%s\n", buffer);
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);
		pc += sprintf((char *) buffer, "\tSDRAM0 range: %08X : ",pOut->StartAddress[0]);
		pc += sprintf((char *) pc, "%08X \n",pOut->EndAddress[0]);
		more("%s\n", buffer);
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);
		pc += sprintf((char *) buffer, "\tSDRAM1 range: %08X : ",pOut->StartAddress[1]);
		pc += sprintf((char *) pc, "%08X \n",pOut->EndAddress[1]);
		more("%s\n", buffer);
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);
		pc += sprintf((char *) buffer, "\tFLASH0 range: %08X : ",pOut->StartAddress[2]);
		pc += sprintf((char *) pc, "%08X \n",pOut->EndAddress[2]);
		more("%s\n", buffer);
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);
		pc += sprintf((char *) buffer, "\tFLASH1 range: %08X : ",pOut->StartAddress[3]);
		pc += sprintf((char *) pc, "%08X \n",pOut->EndAddress[3]);
		more("%s\n", buffer);
	}
	// don't check the last segment because it hasn't been filled in yet.
	for( i = pOut->NumberOfSegments = 0; i < (MAX_PHYSICAL_MEMORY_SEGMENTS - 1); i++)
	{
		// make a list of memory segments that start before they end, and count them.
		if(pOut->StartAddress[i] < pOut->EndAddress[i])
		{
			pOut->StartAddress[pOut->NumberOfSegments] = pOut->StartAddress[i];
			pOut->EndAddress[pOut->NumberOfSegments] = pOut->EndAddress[i];
			pOut->NumberOfSegments += 1;
		}
	}
	// fill in the last block for 80312 registers
	pOut->StartAddress[pOut->NumberOfSegments] = 0x400;
	pOut->EndAddress[pOut->NumberOfSegments] = 0x1FFF;
	if(Verbose)
	{
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);
		pc += sprintf((char *) buffer, "\t80312  range: %08X : ",pOut->StartAddress[pOut->NumberOfSegments]);
		pc += sprintf((char *) pc, "%08X \n",pOut->EndAddress[pOut->NumberOfSegments]);
		more("%s\n", buffer);
	}
	pOut->NumberOfSegments += 1;
}

/***************************************************************************
* Function: GetVirtualMemory(tPhysicalMemoryList PMemList, tVirtualMemoryList VMemList)
*
* Purpose:  Gets the virtual memory configuration based on the physical memory and MMU tables.
*			PMemList is input, VMemList is output
*			VMemList will be a list of arrays of start and end values for regions of 
*			Virtual Memory that are mapped to physical memory from PMemList. 
*
* Returns:	Nothing
****************************************************************************/
void GetVirtualMemory(tPhysicalMemoryList *PMemList, tVirtualMemoryList *VMemList)
{
	tMMUAtributesForMemoryPage PageTable;
	tMMUGlobalAtributes GlobalMemorySettings;
	UINT32 i, j, length, lasti, ValidVirtualAddresses = 0;

	GetGlobalMemorySettings(&GlobalMemorySettings);
	GetMemorySettingForAddress(0, &GlobalMemorySettings, &PageTable);
	lasti = 0;
	// check all virtual addresses, stop after rollover, would be good to use double value!!!
	for( i=0; i >= lasti; i += ((PageTable.EndOfPage - PageTable.StartOfPage) + 1))
	{
		lasti = i;
		GetMemorySettingForAddress(i, &GlobalMemorySettings, &PageTable);
		// against all valid physical addresses, making a table of virtual addresses
		// that map to valid physical addresses.
		for(j=0; j < PMemList->NumberOfSegments; j++)
		{
			// !!!could be more thorough in the case of partially overlapping segments.
			if((PageTable.StartOfPage >= PMemList->StartAddress[j]) & (PageTable.StartOfPage <= PMemList->EndAddress[j]))
//			if(FindIntersectionOfRanges(PhysicalRange, VirtualRange, IntersectingRange))
			{
				// if segment is contiguous with last, just update end address, otherwise make a new entry.
				if(((PageTable.VirtualAddress & (0xFFFFFFFF - length)) -1) == VMemList->EndAddress[ValidVirtualAddresses -1])
				{
					VMemList->EndAddress[ValidVirtualAddresses -1] = (PageTable.VirtualAddress | length);
				}else
				{
					length = (PageTable.EndOfPage - PageTable.StartOfPage);
					VMemList->StartAddress[ValidVirtualAddresses] = (PageTable.VirtualAddress & (0xFFFFFFFF - length));
					VMemList->EndAddress[ValidVirtualAddresses] = (PageTable.VirtualAddress | length);
					ValidVirtualAddresses++;
				}
				// don't need to check the other physical segments
				j = PMemList->NumberOfSegments;
			}
		}
	}
	VMemList->NumberOfSegments = ValidVirtualAddresses;
}
/***************************************************************************
* Function: PrintPageType(tMMUAtributesForMemoryPage *pOut)
*
* Purpose:  Print the type of table entry from structure.
*
* Returns:	Nothing
****************************************************************************/
void PrintPageType(tMMUAtributesForMemoryPage *pIn)
{
	unsigned char buffer[84];
	unsigned char *pc;
	unsigned char format1, format2;

	buffer[79] = '\0';
	memset(pc = buffer, ' ', 79);
	format1 = (unsigned char)pIn->FirstLevelDescriptor & 0x03;
	format2 = (unsigned char)pIn->SecondLevelDescriptor & 0x03;
	switch ( format1 )
	{
		case 0: // fault
			pc += sprintf((char *) buffer, "Invalid or unitialized first level descriptor.\n");	
			break;
		case 1: // Coarse Page Table
			pc += sprintf((char *) buffer, "Coarse Page Table, ");	
			switch ( format2 )
			{
				case 0: // fault
					pc += sprintf((char *) pc, "Invalid or unitialized second level descriptor.\n");	
					break;
				case 1: // Large Page Table
					pc += sprintf((char *) pc, "Large Page.\n");	
					break;
				case 2: // Small Page Table
					pc += sprintf((char *) pc, "Small Page.\n");	
					break;
				case 3: // Extended Small Page Table
					pc += sprintf((char *) pc, "Extended Small Page.\n");	
					break;
				default: // There was an Error in This code if it got here
					pc += sprintf((char *) pc, "Error in PrintPageType routine.\n");	
					break;
			}
			break;
		case 2: // Section Base Address
			pc += sprintf((char *) buffer, "Section Table.\n");	
			break;
		case 3: // Fine Page Table
			pc += sprintf((char *) buffer, "Fine Page Table, ");	
			switch ( format2 )
			{
				case 0: // fault
					pc += sprintf((char *) pc, "Invalid or unitialized second level descriptor.\n");	
					break;
				case 1: // Large Page Table
					pc += sprintf((char *) pc, "Large Page.\n");	
					break;
				case 2: // Small Page Table
					pc += sprintf((char *) pc, "Small Page.\n");	
					break;
				case 3: // Extended Small Page Table
					pc += sprintf((char *) pc, "Tiny Page.\n");	
					break;
				default: // There was an Error in This code if it got here
					pc += sprintf((char *) pc, "Error in PrintPageType routine.\n");	
					break;
			}
			break;
		default: // There was an Error in This code if it got here
			pc += sprintf((char *) buffer, "Error in PrintPageType routine.\n");	
			break;
	}
	more("%s\n", buffer);
}

typedef struct
{
	UINT32 Width;
	UINT32 Address;
	UINT32 Attributes;
} in_WrapChangeSettingsForMemorySection;
/***************************************************************************
* Function: ChangeSettingsForMemorySection(UINT32 VirtualAddress, UINT32 Attributes)
*
* Purpose:  Change the attributes for the section that includes virtual address
*			If virtual address is in a segment that is not a section print error. 
*
* Returns:  nothing
****************************************************************************/
void ChangeSettingsForMemorySection(UINT32 VirtualAddress, UINT32 Attributes)
{
	UINT32 TTBase, *AddressOfFirstLevelDescriptor;
	UINT32 FirstLevelDescriptor;
	UINT32 TableIndex;
	unsigned char buffer[84];
	unsigned char *pc;

	buffer[79] = '\0';
	memset(pc = buffer, ' ', 79);
	TableIndex = (( VirtualAddress & 0xFFF00000) >> 18);
	TTBase = GetTTBase();
	AddressOfFirstLevelDescriptor = (UINT32 *)( TTBase | TableIndex);
	FirstLevelDescriptor = *AddressOfFirstLevelDescriptor;
	if( (FirstLevelDescriptor & 0x02) == 0x02)
	{
		// this is really a section, so change the attributes, leaving the section bits
		Attributes &= 0x00001FFC; // don't allow a change to section base address
		FirstLevelDescriptor &= 0xFFFFE002;
		FirstLevelDescriptor |= Attributes;
		*AddressOfFirstLevelDescriptor = FirstLevelDescriptor;
		InvalidateDataTLBLine(VirtualAddress);
	}else
	{
		pc += sprintf((char *) buffer, "Can't change settings, this memory is not a section.\n" );
		more("%s\n", buffer);
	}
}
/***************************************************************************
* Function: WrapChangeSettingsForMemorySection(in_WrapChangeSettingsForMemorySection *pIn, void *pOut)
*
* Purpose:  Called from terminal command
*			calls ChangeSettingsForMemorySection(VirtualAddress, Attributes)
*
* Returns:  nothing
****************************************************************************/
void WrapChangeSettingsForMemorySection(in_WrapChangeSettingsForMemorySection *pIn, void *pOut)
{
	UINT32 VirtualAddress = pIn->Address;
	UINT32 Attributes = pIn->Attributes;

	ChangeSettingsForMemorySection(VirtualAddress, Attributes);
}

/***************************************************************************
* Function: PrintMemorySettings(tMMUAtributesForMemoryPage *pIn, tMMUGlobalAtributes *pGMS,
*								 unsigned char verbose)
*
* Purpose:  Print the MMU table entry settings from structure.
*
* Returns:	Nothing
****************************************************************************/
void PrintMemorySettings(tMMUAtributesForMemoryPage *pIn, tMMUGlobalAtributes *pGMS,
						 unsigned char verbose)
{
	unsigned char buffer[84];
	unsigned char *pc;


	if(pIn->Fault != 0)
	{
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);
		pc += sprintf((char *) buffer, "PrintMemorySettings Fault code:\t%08X\n",pIn->Fault);
		more("%s\n", buffer);
		return;
	}
	if(verbose)
	{
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);

		pc += sprintf((char *) buffer, "Memory is ");	
		if(!pIn->ECCEnabled)
		{
			pc += sprintf((char *) pc, "NOT ");
		}
		pc += sprintf((char *) pc, "protected by the Error Correction Controller.\n");
		more("%s\n", buffer);
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);

		pc += sprintf((char *) buffer, "MMU is ");	
		if(pGMS->MMUEnabled)
		{
			pc += sprintf((char *) pc, "Enabled.\n");
		}else
		{
			pc += sprintf((char *) pc, "Disabled.\n");
		}
		more("%s\n", buffer);
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);

		pc += sprintf((char *) buffer, "Data Cache is ");	
		if(pGMS->DataCacheEnabled)
		{
			pc += sprintf((char *) pc, "Enabled.\n");
		}else
		{
			pc += sprintf((char *) pc, "Disabled.\n");
		}
		more("%s\n", buffer);
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);

		PrintPageType(pIn);
		pc += sprintf((char *) buffer, "\tVirtual  Address: %08X\n",pIn->VirtualAddress);
		pc += sprintf((char *) pc, "\tPhysical Address: %08X\n",pIn->PhysicalAddress);
		more("%s\n", buffer);
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);
		pc += sprintf((char *) buffer, "\tStart of page: %08X\n",pIn->StartOfPage);
		pc += sprintf((char *) pc, "\tEnd of page:   %08X\n",pIn->EndOfPage);
		more("%s\n", buffer);
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);
		pc += sprintf((char *) buffer, "\tFirst  Level Descriptor: %08X\n",pIn->FirstLevelDescriptor);
		pc += sprintf((char *) pc, "\tSecond Level Descriptor: %08X\n",pIn->SecondLevelDescriptor);
		more("%s\n", buffer);
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);
		pc += sprintf((char *) buffer, "\tAttributes: %08X\n",pIn->Attributes);
		more("%s\n", buffer);
		PrintCacheMode(pIn->xcbbits, pGMS->AuxControlRegValue);
	}else
	{
		buffer[79] = '\0';
		memset(pc = buffer, ' ', 79);

		pc += sprintf((char *) buffer, "Virtual: %08X\t",pIn->VirtualAddress);
		pc += sprintf((char *) pc, "Physical: %08X\t",pIn->PhysicalAddress);
		pc += sprintf((char *) pc, "Length: %08X\n",((pIn->EndOfPage - pIn->StartOfPage) + 1));
		more("%s\n", buffer);
		
		PrintCacheMode(pIn->xcbbits, pGMS->AuxControlRegValue);
	}
}
typedef struct
{
	UINT32 Width;
	UINT32 Address;
} in_PrintSettingsForMemoryPage;
/***************************************************************************
* Function: PrintSettingsForMemoryPage(in_PrintSettingsForMemoryPage *pIn, void *pOut)
*
* Purpose:  Called from terminal command. Gets global memory settings
*			MMU settings for the address and prints the results
*
* Returns:  nothing
****************************************************************************/
void PrintSettingsForMemoryPage(in_PrintSettingsForMemoryPage *pIn, void *pOut)
{
	tMMUAtributesForMemoryPage PageTable;
	tMMUGlobalAtributes GlobalMemorySettings;
	unsigned char verbose = 1;

	GetGlobalMemorySettings(&GlobalMemorySettings);
	GetMemorySettingForAddress(pIn->Address, &GlobalMemorySettings, &PageTable);
	PrintMemorySettings(&PageTable, &GlobalMemorySettings, verbose);
}
typedef struct
{
	UINT32 Width;
	UINT32 Verbose;
} in_PrintSettingsForAllMemoryPages;
/***************************************************************************
* Function: PrintSettingsForAllMemoryPages(in_PrintSettingsForAllMemoryPages *pIn, void *pOut)
*
* Purpose:  Prints a list of all memory segments.
*			To be human readable the output only includes segments that change from
*			the previous in at least one of the following ways:
*			Virtual to Physical offset, cache bits, length of segment
*
* Returns:  nothing
****************************************************************************/
void PrintSettingsForAllMemoryPages(in_PrintSettingsForAllMemoryPages *pIn, void *pOut)
{
	tMMUAtributesForMemoryPage PageTable;
	tMMUGlobalAtributes GlobalMemorySettings;

//	unsigned char verbose = (unsigned char)pIn->Verbose;
	UINT32 Address, lastlength = 0, length, VirtPhysDelta, LastVirtPhysDelta = 0;
	UINT32 LastVirtPhysDeltaDir = 0, SectionsNonContiguous = 0;
	unsigned char lastxcb = 0xFF; // invalid combination

	GetGlobalMemorySettings(&GlobalMemorySettings);
	Address = 0x00;
	while(Address < 0xFFF00000)
	{
		GetMemorySettingForAddress(Address, &GlobalMemorySettings, &PageTable);
		length = (PageTable.EndOfPage - PageTable.StartOfPage);
		// check if memory is contiguous between sections, have to look at distance between
		// virtual and physical as well as the direction of the difference.
		if(PageTable.VirtualAddress > PageTable.PhysicalAddress)
		{
			VirtPhysDelta = (PageTable.VirtualAddress - PageTable.PhysicalAddress);
			if((LastVirtPhysDeltaDir == 0) | (VirtPhysDelta != LastVirtPhysDelta))
				SectionsNonContiguous = 1;
			else
				SectionsNonContiguous = 0;
		}else
		{
			VirtPhysDelta = (PageTable.PhysicalAddress - PageTable.VirtualAddress);
			if((LastVirtPhysDeltaDir == 1) | (VirtPhysDelta != LastVirtPhysDelta))
				SectionsNonContiguous = 1;
			else
				SectionsNonContiguous = 0;
		}
		if((lastlength != length) | (lastxcb != PageTable.xcbbits) | (SectionsNonContiguous))
		{
			PrintMemorySettings(&PageTable, &GlobalMemorySettings, (unsigned char)pIn->Verbose);
		}
		Address += (length + 1);
		lastlength = length;
		lastxcb = PageTable.xcbbits;
		LastVirtPhysDelta = VirtPhysDelta;
	}
}

Command(CMD_PRINT_ALL_MEM_SETTINGS, "pa", PrintSettingsForAllMemoryPages, SIZESUFFIX, "PrintAllPages", "Print All memory pages, short form.");
Arg(UINT32ARG, "Verbose", "");
CommandEnd();

Command(CMD_PRINT_MEM_SETTINGS, "pm", PrintSettingsForMemoryPage, SIZESUFFIX, "PrintMemPage", "Print Memory page settings");
Arg(UINT32ARG, "Address", "");
CommandEnd();

Command(CMD_CHANGE_MEM_SETTING, "cm", WrapChangeSettingsForMemorySection, SIZESUFFIX, "ChangeMemSetting", "Change Memory section settings");
Arg(UINT32ARG, "Address", "");
Arg(UINT32ARG, "Parameters", "");
CommandEnd();





