/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2002 Intel Corporation. All Rights Reserved.
*****************************************************************************/

/*****************************************************************64
*     Name:    OSAL C source code file for VxWorks
*  
*    Type:     C  source file
* 
*    Desc:     This file contains the C source code for OSAL
*              primitives. The OSAL API mapping code for VxWorks   
*              is bounded by #ifdef IX_VXWORKS flag.  
*
*    File:     ix_osal_VxWorks.c 
*
*    Sid:      ix_osal_VxWorks.c@@/main/1 - Tue Dec 10 10:33:12 2002
*
*    Prg:      shabbir ali 
*
*****************************************************************/

#ifdef IX_VXWORKS

/* Common Include files */
#include "ix_osal.h"          /* OSAL header file */


/*****************************************************************
*
*      Fun:   IX_DestroyBinarySem_VxWrks 
*
*      Desc:  Destroy an instance of Binary Semaphore. 
*
*      Ret:   IX_NO_ERROR       - Success
*             IX_UNKNOWN_HANDLE - Failure(invalid parameter)
*             IX_FAILED         - Failure(failed to destroy the sem)   
*
*      Notes:  
*
*      File:  ix_osal_VxWorks.c
*
*****************************************************************/

IX_Err_t IX_DestroyBinarySem_VxWrks (IX_Sem_t s)
{
    /* Delete the Semaphore */
    if( semDelete(s) == OK)
        return(IX_NO_ERROR);
    else
    {
        
        switch(errno)
        {
        case S_intLib_NOT_ISR_CALLABLE:
        case S_smObjLib_NO_OBJECT_DESTROY:
            return(IX_FAILED);
            break;
        case S_objLib_OBJ_ID_ERROR:
            return(IX_UNKNOWN_HANDLE);
            break;
        default:
            return(IX_FAILED);
            break;
        }
    }
    
}


/*****************************************************************
*
*      Fun:   IX_GiveBinarySem_VxWrks 
*
*      Desc:  Give/Unlock an instance of Binary Semaphore. 
*
*      Ret:   IX_NO_ERROR       - Success
*             IX_UNKNOWN_HANDLE - Failure(invalid parameter)
*             IX_FAILED         - Failure(semaphore is already in 
*                                 given state)   
*
*      Notes:  
*
*      File:  ix_osal_VxWorks.c
*
*****************************************************************/

IX_Err_t IX_GiveBinarySem_VxWrks (IX_Sem_t s)
{
    
    /* Give/Unlock the semaphore */
    if( semGive(s) == OK)
        return(IX_NO_ERROR);
    else
    {
        
        switch(errno)
        {
        case S_intLib_NOT_ISR_CALLABLE:
        case S_semLib_INVALID_OPERATION:
            return(IX_FAILED);
            break;
        case S_objLib_OBJ_ID_ERROR:
            return(IX_UNKNOWN_HANDLE);
            break;
        default:
            return(IX_FAILED);
            break;
        }
        
    }
    
}

/*****************************************************************
*
*      Fun:   IX_TakeBinarySem_VxWrks 
*
*      Desc:  Take/Lock an instance of Binary Semaphore. 
*
*      Ret:   IX_NO_ERROR       - Success
*             IX_UNKNOWN_HANDLE - Failure(invalid parameter)
*             IX_FAILED         - Failure(Semaphore already in 
*                                 lock state and block = IX_TRUE)   
*
*      Notes:  
*
*      File:  ix_osal_VxWorks.c
*
*****************************************************************/

IX_Err_t IX_TakeBinarySem_VxWrks (IX_Sem_t s, IX_Boolean_t block)
{
    /* Check if block is FALSE */
    if( block == IX_FALSE)
    {
        /* Non - Blocking Call */
        if(semTake(s,NO_WAIT) == OK)
            return(IX_NO_ERROR);
        
    }
    else
    {
        /* Blocking Call */
        if(semTake(s,WAIT_FOREVER) == OK)
            return(IX_NO_ERROR);
        
    }
    
    switch(errno)
    {
    case S_intLib_NOT_ISR_CALLABLE:
    case S_objLib_OBJ_UNAVAILABLE:
        return(IX_FAILED);
        break;
    case S_objLib_OBJ_ID_ERROR:
        return(IX_UNKNOWN_HANDLE);
        break;
    default:
        return(IX_FAILED);
        break;
    }
    
}

/*****************************************************************
*
*      Fun:   IX_Int_Lock_VxWrks
*
*      Desc:  Lock/Disable all the interrupts in the system. 
*
*      Ret:   IX_NO_ERROR       - Success
*             IX_UNKNOWN_ERROR  - Failure(failed to lock interrupts)   
*
*      Notes:  
*
*      File:  ix_osal_VxWorks.c
*
*****************************************************************/

IX_Err_t IX_Int_Lock_VxWrks (IX_U32 *lockId)
{
    
    IX_U32 regVal = 0;
    
    /* Disable Task Switching */
    if( intContext() == IX_FALSE)
        taskLock();
    
    /* Lock the Interrupts in the system */
    regVal = intLock();
    
    /* Set the Lock Id returned by intLock */
    *lockId = regVal;
    
    return(IX_NO_ERROR);
}

/*****************************************************************
*
*      Fun:   IX_Int_Unlock_VxWrks
*
*      Desc:  Unlock/Enable the interrupts in the system. 
*
*      Ret:   IX_NO_ERROR       - Success
*             IX_UNKNOWN_ERROR  - Failure(failed to unlock interrupts)   
*
*      Notes:  
*
*      File:  ix_osal_VxWorks.c
*
*****************************************************************/

IX_Err_t IX_Int_Unlock_VxWrks (IX_U32 lockId)
{
    
    /* Unlock/Enable the interrupts in the system */
    intUnlock(lockId);
    
    /* Enable Task Switching */
    if( intContext() == 	IX_FALSE)
        taskUnlock();
    
    return(IX_NO_ERROR);
}

/*****************************************************************
*
*      Fun:   IX_Thread_Spawn_VxWrks
*
*      Desc:  Spawn a new Task/Thread. 
*
*      Ret:   IX_NO_ERROR       - Success
*             IX_OUT_OF_MEMORY  - Failure(out of memory)
*             IX_BAD_PARAMS     - Failure(invalid parameter)
*             IX_FAILED         - Failure(failed to allocate more sem)   
*
*      Notes:  
*
*      File:  ix_osal_VxWorks.c
*
*****************************************************************/

IX_Err_t IX_Thread_Spawn_VxWrks(IX_Function_Ptr_t f, IX_Thread_Info_t *t,
                                IX_Thread_id *i)
{
    
    IX_Thread_id retVal = 0; /* return Value */
    IX_U16 lsbFuncPtr = 0x0000;
    IX_U16 msbFuncPtr = 0x0000;
    IX_U16 lsbArgPtr = 0x0000;
    IX_U16 msbArgPtr = 0x0000;
    
    /* Check if Function Pointer is NULL */
    if(f == NULL)
        return(IX_BAD_PARAM);
    
    /* Store the Function Pointer address */
    lsbFuncPtr = (IX_U32)f & 0x0000ffff;
    msbFuncPtr = ((IX_U32)f >> 16) & 0x0000ffff;
    
    /* Store the address of the Argument to Function Pointer */
    lsbArgPtr = (IX_U32)(t->arg) & 0x0000ffff;
    msbArgPtr = ((IX_U32)(t->arg) >> 16) & 0x0000ffff;
    
    /* Spawn the Task */
    retVal = taskSpawn(t->thread_name, t->priority,0,t->stackSizeInBytes,
							 (FUNCPTR)ixConvFuncVxWrks,lsbFuncPtr,msbFuncPtr,lsbArgPtr,
                             msbArgPtr,0,0,0,0,0,0);
    
    /* Check for failure conditions */
    if(retVal == -1)
    {
        switch(errno)
        {
        case S_intLib_NOT_ISR_CALLABLE:
        case S_smObjLib_NOT_INITIALIZED: 
        case S_memLib_BLOCK_ERROR:
            return(IX_FAILED);
            break;
        case S_objLib_OBJ_ID_ERROR:
            return(IX_BAD_PARAM);
            break;
        case S_memLib_NOT_ENOUGH_MEMORY:
            return(IX_OUT_OF_MEMORY);
            break;
        default:
            return(IX_FAILED);
            break;
        }
    }
    else
    {
        /* set the Thread ID */
        *i = retVal;
    }
    
    return(IX_NO_ERROR);
}

/*****************************************************************
*
*      Fun:   IX_Thread_Stop_VxWrks
*
*      Desc:  Stop a Task/Thread. 
*
*      Ret:   IX_NO_ERROR       - Success
*             IX_UNKNOWN_HANDLE - Failure(Unknown Thread ID)   
*             IX_FAILED         - Failure(failed to Stop Task/Thread)   
*
*      Notes:  
*
*      File:  ix_osal_VxWorks.c
*
*****************************************************************/

IX_Err_t IX_Thread_Stop_VxWrks (IX_Thread_id i)
{
    
    /* Stop the Task */
    if( taskDelete(i) == OK)
        return(IX_NO_ERROR);
    else
    {
        switch(errno)
        {
        case S_intLib_NOT_ISR_CALLABLE:
        case S_objLib_OBJ_DELETED:
        case S_objLib_OBJ_UNAVAILABLE:
            return(IX_FAILED);
            break;
        case S_objLib_OBJ_ID_ERROR:
            return(IX_UNKNOWN_HANDLE);
            break;
        default:
            return(IX_FAILED);
            break;
        }
    }
}


/*****************************************************************
*
*      Fun:   ixConvFuncVxWrks 
*
*      Desc:  This function is used only for VxWorks. This function
*					is used to map the FUNCPTR type used in VxWorks
*					with the Pointer to function returing void *, as 
*					defined in OSAL. 
*
*      Ret:   NONE 
*
*      Notes:  
*
*      File:  ix_osal_VxWorks.c
*
*****************************************************************/
FUNCPTR ixConvFuncVxWrks(int lsbFuncPtr, int msbFuncPtr,int lsbArgPtr,
                         int msbArgPtr, int arg4,int arg5, int arg6,int arg7, int arg8, int arg9)
{
    IX_Function_Ptr_t temp;
    IX_U32 arg = 0x00000000;
    IX_U32 funPtr = 0x00000000;
    
    /* Create the Function Pointer */
    funPtr = (msbFuncPtr << 16) | lsbFuncPtr;
    
    /* Create the Argument Pointer */
    arg = (msbArgPtr << 16) | lsbArgPtr;
    
    temp = (IX_Function_Ptr_t)funPtr;
    
    /* Call the actual entry point Function */
    (*temp)((void *)arg);
    
    return (IX_NO_ERROR);
}




#endif /* IX_VXWORKS */
/**********************************************************************

  ver       pat    init                  description
  ------------ -------- ---- ----------------------------------------------
  $SID$          ---    sali 1. initial release.
  
**********************************************************************/
