/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/* ixd2810.c - IXD2810 MAC device driver 

modification history
--------------------
01a,08oct02,eme		written.
01b,16jun03,tep added include of ixd2810Int.h, added "break" statement to 
case BOARDID_BLOCK: in ixd2810IoctlSet, removed bzero and strncpy commands
from create function and replaced with setting of pointer to string, changed
return(devid) to return(OK) in Create function, added LED init code to Start 
function, changed all dprintf to printf for debugging.
01c,16jun03,tep Changed BOARD_ID to IXD2810_BOARD_ID.
01d,24jun03,tep Added enabling of all ports in Start function.
01e,25jun03,tep Added use of macro ALL_PORTS in CSR access to MAC.
01f,25jun03,tep Added ixd2810Version for driver version info.
01g,26jun03,tep Fixed ioctlset and get by multiplying regoffset by 4.
01h,27jun03,tep Moved the RX and TX blocks to be read\write-able via ioctl,
these were read only before.
*/

/*
DESCRIPTION


*/
#include <ixd2810.h>
#include <ixd2810Int.h>
#include <ixd2810Mmap.h>
#include "string.h"

/* locals */
LOCAL	IXD2810_DEV ixd2810Dev[N_DEVICES];
LOCAL	uint32_t 	ixd2810BoardId; 



/* forward prototypes */
LOCAL STATUS ixd2810IoctlSet(IXD2810_DEV *pDev, block_t *stBlock);
LOCAL STATUS ixd2810IoctlGet(IXD2810_DEV *pDev, block_t *stBlock);
char *ixd2810Version(void);

/*******************************************************************************
 * ixd2810_Create - Create a device and returns its ID
 *
 * RETURNS: OK or ERROR if a failure occurs.
 */
uint32_t ixd2810_Create
    (
    uint32_t devId  /* device ID for creation */
    ) 
	{
    uint32_t data;
	IXD2810_DEV *pDev;

	printf(ixd2810Version());

	/* check for device availability */
	if (devId >= N_DEVICES) 
		{
		printf("Device number out of range.\n");
		(void) errnoSet (S_ixdLib_DEVICE_ID_NOT_FOUND);
		return ERROR;
		}
	
	/* get pointer to device structure */
	pDev = &ixd2810Dev[devId];

    /* sanity check */
    if (pDev->created == TRUE)
        {
        printf("Device %d already created\n", devId);
        (void) errnoSet (S_ixdLib_DEVICE_ALREADY_CREATED);
        return ERROR;
        }
        
	/* Get board ID */
    IXD2810_REG_GET(ixd2810BoardIdBase[devId] + IXD2810_BOARDID_OFFSET, data);
	ixd2810BoardId = data & IXD2810_BOARDID_MASK;

	if (ixd2810BoardId & ~IXD2810_BOARD_ID) 
		{
		printf("Valid board ID not found for dev %d!\n", devId);
		(void) errnoSet (S_ixdLib_BOARD_ID_NOT_FOUND);
		return ERROR;
		}

	printf("Device ID, %08x, found!\n", ixd2810BoardId);

	printf("Initializing the created IXF1110 MAC dev %d\n", devId);

	/* initialize the device local attributes */
	pDev->devHdr.name = taskNames[devId];
	pDev->devId = devId;
	pDev->intLevel = intLevelList[devId];
	pDev->vbase = ixd2810BoardIdBase[devId];
	pDev->syncSem = semBCreate(SEM_Q_FIFO, SEM_EMPTY);

	/* device status; current mode */
	pDev->created = TRUE;

	return OK;
	}

	
/*******************************************************************************
 * ixd2810_Start - Start the execution of the MAC
 *
 * RETURNS: OK or ERROR if a failure occurs.
 */
STATUS ixd2810_Start
	(
	uint32_t devId	/* device to execute */
	) 
 	{
	uint32_t p;
	IXD2810_DEV *pDev;

	/* sanity check */
	if (devId >= N_DEVICES) 
		{
		printf("Device number out of range.\n");
		(void) errnoSet (S_ixdLib_DEVICE_ID_NOT_FOUND);
		return ERROR;
		}

	/* initialize the device(s) found */
	pDev = &ixd2810Dev[devId];

    /* verify whether or not the device has been created */
    if (pDev->created == FALSE) 
        {
        printf("Device %d not yet created\n", devId);
        (void) errnoSet (S_ixdLib_DEVICE_NOT_CREATED);
        return ERROR;
        }
        
	/* do nothing if already running */
	if (pDev->isRunning) 
        {
		printf("Device %x already started.\n");
		(void) errnoSet (S_ixdLib_DEVICE_ALREADY_RUNNING);
		return ERROR;
        }

	/* Initialize LEDs */
    IXD2810_REG_SET(pDev->vbase + GLOBAL_ADRS(LED_CONTROL), 0x2);
    IXD2810_REG_SET(pDev->vbase + GLOBAL_ADRS(LINK_LED_ENABLE), ALL_PORTS);
	/* Enable all ports */
    IXD2810_REG_SET(pDev->vbase + GLOBAL_ADRS(PORT_ENABLE), ALL_PORTS);

	/* initialize port attributes */
	for(p = 0; p < N_PORTS; p++) 
		{
		printf("Start autoneg on MAC%d port %d.\n", devId, p);

#ifndef DEBUG
		/* Auto-negotation on all ports */		
		if (doAutoneg(pDev, p) == ERROR)
			{
			printf("ixdp2810Start(): Unable to perform auto-negotiation\n");
			return ERROR;
			}
#endif /* DEBUG */
		}

	/* install the interrupt handler  */
	if (ixd2810IntDevInit(pDev) == ERROR)
		{
		printf("ixdp2810Start(): Unable to initialize the interrupt\n");
		return ERROR;
		}

	/* update device execution mode */
	pDev->isRunning = TRUE;

	printf("Device %d is executing!\n", devId);

	return OK;
	}


/*******************************************************************************
 * ixd2810_Stop - Terminate the execution of the MAC
 *
 * RETURNS: OK or ERROR
 */
STATUS ixd2810_Stop
	(
	uint32_t devId	/* device to stop execution */
	)
	{
	IXD2810_DEV *pDev;

	/* sanity check */
	if (devId >= N_DEVICES) {
		printf("Device number out of range\n");
		(void) errnoSet (S_ixdLib_DEVICE_ID_NOT_FOUND);
		return ERROR;
	}

	/* get pointer to device structure */
	pDev = &ixd2810Dev[devId];

	/* device current mode */
	if (pDev->isRunning == FALSE) {
		printf("Unable to stop device %d since it's not running\n", devId);
		(void) errnoSet (S_ixdLib_DEVICE_NOT_RUNNING);
		return ERROR;
	}

	ixd2810IntDevRelease(pDev);

	semFlush(pDev->syncSem);
	semDelete(pDev->syncSem);

	pDev->isRunning = FALSE;
    pDev->created = FALSE;
    
	return OK;
	}

/*******************************************************************************
 * ixd2810_Ioctl - special device control.
 *
 * This routine handles specialized command requests.
 *
 * RETURNS: OK or ERROR if invalid request.
 */
STATUS ixd2810_Ioctl
	(
 	uint32_t devId,		/* device control */
	uint32_t cmd,		/* request code */
	block_t *stBlock	/* user-accessable device structure */
	)
	{
    IXD2810_DEV *pDev;
	STATUS status = OK;

	/* sanity check */
	if (devId >= N_DEVICES) 
		{
		printf("Device number out of range\n");
		(void) errnoSet (S_ixdLib_DEVICE_ID_NOT_FOUND);
		return ERROR;
		}

    pDev = &ixd2810Dev[devId];
    
	if (pDev->created == FALSE || pDev->isRunning == FALSE) 
		{
    	(void) errnoSet (S_ioLib_DEVICE_ERROR);
		return ERROR;
		}

	switch(cmd) 
		{
		case IXD2810_IOCTL_GET:
			status = ixd2810IoctlGet(pDev, stBlock);
			break;
		case IXD2810_IOCTL_SET:
			status = ixd2810IoctlSet(pDev, stBlock);
			break;
		default:
	    	(void) errnoSet (S_ioLib_UNKNOWN_REQUEST);
			status = ERROR;
			break;
		}

	return status;
	}

/*******************************************************************************
 * ixd2810IoctlGet - special device control.
 *
 * This routine handles the get ioctl command requests.
 * Retrieve the values of a contiguous registers list.
 *
 * RETURNS: OK or ERROR if invalid request.
 */
STATUS ixd2810IoctlGet
    (
    IXD2810_DEV *pDev,      /* pointer to device structure */
    block_t *stBlock		/* user-accessable device structure */
	)
	{
	int regOffset;
	STATUS status = OK;

	/* sanity check */
	if (stBlock->blockId > BOARDID_BLOCK || stBlock->blockId < PORT0_BLOCK)
		{
		printf("Block %d out of boundary.\n", stBlock->blockId);
		(void) errnoSet (S_ixdLib_BLOCK_NOT_FOUND);
		return ERROR;
		}

    /* access registers */
	/* other blocks on the memory map */
	switch(stBlock->blockId)
		{
		case PORT0_BLOCK:
		case PORT1_BLOCK:
		case PORT2_BLOCK:
		case PORT3_BLOCK:
		case PORT4_BLOCK:
		case PORT5_BLOCK:
		case PORT6_BLOCK:
		case PORT7_BLOCK:
		case PORT8_BLOCK:
		case PORT9_BLOCK:
		case GLOBAL_BLOCK:
		case TX_BLOCK:
		case RX_BLOCK:
		case SPI42_BLOCK:
		case SERDES_BLOCK:
			for(regOffset = 0; regOffset < stBlock->regNum; regOffset++) 
				{
				IXD2810_REG_GET(pDev->vbase +
			      BLOCK_REG_ADRS(stBlock->blockId, (stBlock->startingReg+regOffset)*4),
				  stBlock->data[regOffset]);
				}
			break;				
		case GBIC_BLOCK:
			for(regOffset = 0; regOffset < stBlock->regNum; regOffset++) 
				{
				IXD2810_REG_GET(pDev->vbase + GBIC_BASE + 
                                ((stBlock->startingReg+regOffset)*4),
								stBlock->data[regOffset]);
				}
			break;
		case BOARDID_BLOCK:
            IXD2810_REG_GET(pDev->vbase + IXD2810_BOARDID_OFFSET, stBlock->data[0]);
            break;
		default:
			printf("Unknown block specified: %s\n", 
				ixd2810BlockStr[stBlock->blockId]);
			(void) errnoSet (S_ixdLib_UNKNOWN_BLOCK);
			status = ERROR;
			break;
		}

	return status;
	}


/*******************************************************************************
 * ixd2810IoctlSet - special device control.
 *
 * This routine handles the ioctl get command requests.
 * Access and configure the list of registers with the value
 * given in the stBlock structure.
 *
 * RETURNS: OK or ERROR if invalid request.
 */
STATUS ixd2810IoctlSet
	(
    IXD2810_DEV *pDev,      /* pointer to device structure */
	block_t *stBlock		/* user-accessable device structure */
	)
	{
	int regOffset;
	STATUS status = OK;

	/* sanity check */
	if (stBlock->blockId > BOARDID_BLOCK || stBlock->blockId < PORT0_BLOCK)
		{
		printf("Block %d out of boundary.\n", stBlock->blockId);
		(void) errnoSet (S_ixdLib_BLOCK_NOT_FOUND);
		return ERROR;
		}

	/* access registers */
	/* blocks on the memory map */
	switch(stBlock->blockId)
		{
		case PORT0_BLOCK:
		case PORT1_BLOCK:
		case PORT2_BLOCK:
		case PORT3_BLOCK:
		case PORT4_BLOCK:
		case PORT5_BLOCK:
		case PORT6_BLOCK:
		case PORT7_BLOCK:
		case PORT8_BLOCK:
		case PORT9_BLOCK:
		case GLOBAL_BLOCK:
		case TX_BLOCK:
		case RX_BLOCK:
		case SPI42_BLOCK:
		case SERDES_BLOCK:
			for(regOffset = 0; regOffset < stBlock->regNum; regOffset++) 
				{
				IXD2810_REG_SET(pDev->vbase + 
				    BLOCK_REG_ADRS(stBlock->blockId, (stBlock->startingReg+regOffset)*4),
					stBlock->data[regOffset]);
				}
			break;
		case GBIC_BLOCK:
			for(regOffset = 0; regOffset < stBlock->regNum; regOffset++) 
				{
				IXD2810_REG_SET(pDev->vbase + GBIC_BASE + 
                                ((stBlock->startingReg+regOffset)*4), 
								stBlock->data[regOffset]);
				}
			break;
		case BOARDID_BLOCK:
            printf("Block %d [%s] is READ-ONLY\n", stBlock->blockId, 
                    ixd2810BlockStr[stBlock->blockId]);
			(void) errnoSet (S_ixdLib_READONLY_BLOCK);
			status = ERROR;
			break;
		default:
            printf("Block %d [%s] unknown\n", stBlock->blockId,
                    ixd2810BlockStr[stBlock->blockId]);
			(void) errnoSet (S_ixdLib_UNKNOWN_BLOCK);
			status = ERROR;
			break;
		}

	return status;
	}

/*******************************************************************************
 * ixd2810Version - 
 *
 * This routine assembles and returns a pointer to the version string.
 *
 * RETURNS: pointer to the driver version or build number.
 */
char versionString[80];

char *ixd2810Version
	(
	void
	)
	{
	int i;
	for(i=0;i<80;i++) versionString[i]=0;
	strcat(versionString,"Intel IXD2810 Media Card Device Driver Version ");
	strcat(versionString,IXD_VERSION);
	strcat(versionString,"\n");
	return versionString; 
    }

