/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/* ixd28100Autoneg.c - IXD2810 Auto-negotiation module

modification history
--------------------
01a,08oct02,eme  created from autoneg.c 1/1/2001 by Nancy Wang
01b, 18jun03,tep added include of ixd2810Mmap.h, fixed configMAC function by 
adding global_addr macro to reg accesses, changed dprintf to printf for 
debugging.
01c,25jun03,tep Removed printf statments.
*/

/*
DESCRIPTION
This module handles the auto-negotiation for the IXF1110 MAC
device.

*/
#include <ixd2810Autoneg.h>
#include <ixd2810Mmap.h>
#include <ixd2810Int.h>

/*******************************************************************************
 * doAutoneg - Starting poing of auto-negotiation.
*				Determine copper or fiber media and
*				call other functions correspondingly.
 *
 * RETURNS:  OK or ERROR if failure occurs.
 */
STATUS doAutoneg
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port /* port number. From 0 to 9. */
	) {

	/* sanity check */
	if (port >= N_PORTS)
		return ERROR;

	/* Initialize link/speed/duplex status */
	pDev->ports[port].link = DOWN;
	pDev->ports[port].duplex = HALF;
	pDev->ports[port].speed = SPEED_10;
	
	/* Fiber */
	fiberAutoNeg(pDev, port);	/* fiber auto-negotiation */
	/* Fiber mode only supports giga speed and full duplex */
	if (pDev->ports[port].link == UP) {	/* link up */
		if ((pDev->ports[port].speed == SPEED_1000) && 
		(pDev->ports[port].duplex == FULL)) {
			configMac(pDev, port);  /* Configure MAC */
			printMsg(pDev, port, 1);   /* Print message */
		}
		else {
			ixd2810DisablePort(pDev,port);	     /* Disable port */
			printMsg(pDev, port, 2);   /* Print message */
		}
	} /* end if */
	else {				/* link down */
		ixd2810DisablePort(pDev, port);
		printMsg(pDev, port, 3);
	}

	return OK;
}

/*******************************************************************************
 * configMac - 	Configure specified MAC port. 
 *				(link/speed/duplex)
 *
 * RETURNS: N/A
 */
void configMac
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port /* port number. From 0 to 9. */
	) {
	uint32_t data;
	
	/* sanity check */
	if (port >= N_PORTS)
		return;

	/* Set link up*/
	IXD2810_REG_GET(pDev->vbase + GLOBAL_ADRS(LINK_LED_ENABLE), data);
	data |= LINK_UP(port);
	IXD2810_REG_SET(pDev->vbase + GLOBAL_ADRS(LINK_LED_ENABLE), data);

	/* Enable port */
	IXD2810_REG_GET(pDev->vbase + GLOBAL_ADRS(PORT_ENABLE), data);
	data |= PORT_EN(port);
	IXD2810_REG_SET(pDev->vbase + GLOBAL_ADRS(PORT_ENABLE), data);
}

/*******************************************************************************
 * fiberAutoNeg - Fiber auto-negotiation routine.
 *
 * RETURNS: N/A
 */
void fiberAutoNeg
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port /* port number. From 0 to 9. */
	) 
	{
	uint32_t txData, rxData, divData;
	int delay;
	uint32_t vbase;

	/* get a copy of the device virtual address */
	vbase = pDev->vbase;

	/* Start auto-negotiation */
	IXD2810_REG_GET(vbase + BLOCK_REG_ADRS(port, DIV_CONFIG_WORD), divData);
	divData |= DC_Autoneg_Enable;
	IXD2810_REG_SET(vbase + BLOCK_REG_ADRS(port, DIV_CONFIG_WORD), divData);

	/* Poll Rx Config Word to determine when AN is complete or timeout */
	for(delay = 0; delay < AN_TIMEOUT; delay++) {
		delayUSec(10000); /* 10 msec delay */
		IXD2810_REG_GET(vbase + BLOCK_REG_ADRS(port, RX_CONFIG_WORD), rxData);
		if (rxData & RC_An_Complete)
			break;
	}

	if (delay < AN_TIMEOUT) {
		/* Speed is fixed at giga */
		pDev->ports[port].speed = SPEED_1000;

		/* Get port status from Rx Config Word */
		if (rxData & RC_Half_Duplex) 
			pDev->ports[port].duplex = HALF;	
		if (rxData & RC_Full_Duplex)
			pDev->ports[port].duplex = FULL;

		/* Link is up when autoneg is enabled and finished */
		pDev->ports[port].link = UP;
	}
	else { /* timeout has occurred */
		pDev->ports[port].link = DOWN;
	}
}

/*******************************************************************************
 * printMsg - Print message for current mode.
 *
 * RETURNS:  N/A
 */
void printMsg
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port,  /* port number. From 0 to 9. */
	int mode /* modes: 1=supported link 2=non-supported link 3=no link */
	) 
	{
	
	char *speedStr[] = {"10M", "100M", "1000M"};
	char *duplexStr[] = {"Half", "Full"};
	char *linkStr[] = {"Down", "Up"};
	int speed, duplex;

	/* sanity check */
	if (port >= N_PORTS)
		return;

	/* Get string index */
	speed = pDev->ports[port].speed;
	duplex = pDev->ports[port].duplex;

	if (mode == 1) 
		{	/* supported link mode */
		dprintf("port%d is configured to %s speed and %s duplex.\n",
	        	port, speedStr[speed], duplexStr[duplex]);
		}
	else 
		{			
		dprintf("port%d is disabled", port);
	
		if (mode == 2) 	/* parallel detection for forced 100M LP */
	      dprintf(" because half duplex is not supported.\n");
	   else 		/* no link mode */
	     dprintf(" because no link is established.\n");
	   }
	}

