/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/* ixd2810Int.c - Interrupt control handler

modification history
--------------------
01a, 09oct02,eme	created
01b, 18jun03,tep added include of ixd2810Mmap.h, moved int level def and task 
name list from header file to this module, cheanged dprintf to printf for 
debugging, changed intconnect to pciintconnect, changed intenable to 
pciintenable, added GBIC_ADRS macros to reg accesses, changed intdisable to 
pciintdisable, added GLOBAL_ADRS macros to reg accesses.
01c,24jun03,tep Fixed reg access in enable port function, changed interrupt 
scheme to use GPIO int instaed of PCI int.
01d,25jun03,tep Added gbic_status0 and 1 elements for history to be preserved 
for mod_def int.  Removed printf statments.  Added volatile directive to all 
reg accesses.
01e,26jun03,tep Added callback function to ISR for core components.
*/
#include <ixd2810Int.h>
#include <ixd2810Mmap.h>

/* local variables */
LOCAL int autoFlag[N_DEVICES][N_PORTS];

/* interrupt level definition */
uint32_t intLevelList[] =
	{
	(uint32_t) INT_LVL_GPIO,
	};

/* task names list */
char *taskNames[] = 
	{
	"Dev_0_ISR",
	"Dev_1_ISR"
	};

int		ixd2810CallBackData;
void 	(*pixd2810CallBackFunc)(int);

/* forward prototypes */
LOCAL void getModState(IXD2810_DEV *pDev,	uint32_t port);
LOCAL void ixd2810Intr(IXD2810_DEV *pDev);
LOCAL void ixd2810RxLossIsr(IXD2810_DEV *pDev, uint32_t irq, uint32_t status);
LOCAL void ixd2810TxFaultIsr(IXD2810_DEV *pDev, uint32_t irq, uint32_t status);
LOCAL void ixd2810ModDefIsr(IXD2810_DEV *pDev);

/*******************************************************************************
 * ixd2810IntDevInit - Initialize the interrupt handler for the MAC device.
 *
 * RETURNS:  ERROR in case an error occurs, otherwise OK.
 */
STATUS ixd2810IntDevInit
	(
	IXD2810_DEV *pDev
	) 
	{
	uint32_t status;
	int taskId;
	int	key;
    
	/* launch a task to serve the mod def irq at task level */
	taskId = taskSpawn(pDev->devHdr.name, TASK_PRIORITY, 
				   TASK_OPTS, TASK_STACK_SIZE, (FUNCPTR) ixd2810ModDefIsr,
				   (int) pDev, 0, 0, 0, 0, 0, 0, 0, 0, 0);

	if (taskId == ERROR) 
		{
		dprintf("Unable to spawn task %s, dev %d.\n",
			  (int) pDev->devHdr.name, pDev->devId);
		return ERROR;
		}		   

	pDev->taskId = taskId;
        
	/* connect isr to interrupt handler */
	(void) intConnect((VOIDFUNCPTR*) INUM_TO_IVEC(pDev->intLevel), 
				(VOIDFUNCPTR) ixd2810Intr, (unsigned int) pDev);

	/* enable interrupt */
	intEnable(pDev->intLevel);

	/* Read GBIC status value */
	IXD2810_REG_GET(pDev->vbase + GBIC_ADRS(GBIC_STATUS), pDev->gbicStatus0);

	/* Synch status 0 and 1 */
	pDev->gbicStatus1=pDev->gbicStatus0;

	/* Enable interrupt in GBIC control register */
	IXD2810_REG_SET(pDev->vbase + GBIC_ADRS(GBIC_CONTROL), GC_INT_ENABLE);

	/* Init GPIO CSRs for media IRQ */
    key = intLock();

	*(volatile uint32_t *)(GPIO_PDCR) = 0xF;
	*(volatile uint32_t *)(GPIO_REDR) = 0x0;
	*(volatile uint32_t *)(GPIO_FEDR) = 0x0;
	*(volatile uint32_t *)(GPIO_EDSR) = 0xF;
	*(volatile uint32_t *)(GPIO_LSHR) = 0x0;
	*(volatile uint32_t *)(GPIO_LSLR) = 0x2;
	*(volatile uint32_t *)(GPIO_LDSR) = 0xF;
	*(volatile uint32_t *)(GPIO_INER) = 0x2;
	*(volatile uint32_t *)(GPIO_INSR) = 0x2;

    intUnlock(key);

	return OK;
	}

/*******************************************************************************
 * ixd2810IntDevRelease - Removes the interrupt handler.
 *
 * Uninitialize any configuration previous done by the ixd2810IntDevInit(),
 * releases any resources acquired during the intialization process.
 *
 * RETURNS:  ERROR in case an error occurs, otherwise OK.
 */
STATUS ixd2810IntDevRelease
	(
	IXD2810_DEV *pDev
	) 
	{
	
	int key;
	                                  
	/* Disable interrupt in GBIC control register */	
	IXD2810_REG_SET(pDev->vbase + GBIC_ADRS(GBIC_CONTROL), GC_INT_DISABLE);	

	/* Disable GPIO CSRs for media IRQ */
    key = intLock();

	*(volatile uint32_t *)(GPIO_INER) = 0x0;
	*(volatile uint32_t *)(GPIO_INSR) = 0x0;
	*(volatile uint32_t *)(GPIO_LSLR) = 0x0;
	*(volatile uint32_t *)(GPIO_INCR) = 0x2;
    
    intUnlock(key);

	intDisable(pDev->intLevel);
	taskDelete(pDev->taskId);


	return OK;
	}

/*******************************************************************************
 * ixd2810Intr - IXD2810 Interrupt Service Routine
 *
 * This routine get called to handle interrupts. Call individual ISRs
 * based on GBIC value. 
 *
 * RETURNS:  N/A
 */
void ixd2810Intr
	(
	IXD2810_DEV *pDev
	) 
	{
	uint32_t gstatus;
	uint32_t irq;
    BOOL irqServiced = FALSE;
	int key;
    
	pDev->intCnt++;

	/* Read GBIC status register  */
	IXD2810_REG_GET(pDev->vbase + GBIC_ADRS(GBIC_STATUS), gstatus);

    key = intLock();

	*(volatile uint32_t *)(GPIO_INST) = 0x2;
    
    intUnlock(key);

	    dprintf("gstatus=%x\n",gstatus);
        dprintf("pDev->gbicStatus=%x\n",pDev->gbicStatus0);

	/* Find which bit(s) changed */
	irq = gstatus ^ pDev->gbicStatus0;

	if (irq & GS_Rx_Loss)
        {
    	dprintf("IR\n");
		ixd2810RxLossIsr(pDev, irq, gstatus);
        irqServiced = TRUE;
        }
	if (irq & GS_Tx_Fault)
        {
    	dprintf("IT\n");
		ixd2810TxFaultIsr(pDev, irq, gstatus);
        irqServiced = TRUE;
        }
	if (irq & GS_Mod_Def)
        {
    	dprintf("IM\n");
		semGive(pDev->syncSem);
        irqServiced = TRUE;
        }

    if (irqServiced == FALSE)
        {
        dprintf("Unable to service IRQ %d\n",irq);
        }

	pDev->gbicStatus1=pDev->gbicStatus0;
	pDev->gbicStatus0=gstatus;

	if(pixd2810CallBackFunc != NULL)
	{
		(*pixd2810CallBackFunc)(ixd2810CallBackData);
	}
    dprintf("IE\n");
	}


/*******************************************************************************
 * ixd2810RxLossIsr - ISR to serve the IXD2810 MAC Rx Loss interrupt
 *
 * Interrupt Service Routine for Rx_Fault_Int. 
 * Display Rx Fault on appropriate device and port.
 *
 * RETURNS:  N/A
 */
void ixd2810RxLossIsr
	(
	IXD2810_DEV *pDev, 	/* pointer to servicing device */
	uint32_t irq,   		/* changed bits in GBIC status */
	uint32_t status 		/* current GBIC status */
	)
	{
	uint32_t	rxBitIndex = GS_RX_STARTING_BIT;  /* GBIC_STATUS bits 29:20 */
	int 	port;
	
	for (port = 0; port < N_PORTS; port++, rxBitIndex <<= 1) 
		{
		if (irq & rxBitIndex) 
			{  /* bit changed for this port */
		    if (status & rxBitIndex) 
		    	{
				dprintf("Rx loss appeared on MAC%d port%d.\n", 
						pDev->devId, port);
				ixd2810DisablePort(pDev, port);	
				} 
			else 
				{
				dprintf("Rx loss disappeared on MAC%d port%d.\n ", 
						pDev->devId, port);
				ixd2810EnablePort(pDev, port);
				}
			}
		}
	}

/*******************************************************************************
 * ixd2810TxFaultIsr - ISR to serve the IXD2810 Tx Fault interrupt
 *
 * Interrupt Service Routine for Tx_Fault_Int. 
 * Display Tx Fault on appropriate device and port.
 *
 * RETURNS:  N/A
 */
void ixd2810TxFaultIsr
	(
	IXD2810_DEV *pDev, 	/* pointer to servicing device */
	uint32_t irq,   		/* changed bits in GBIC status */
	uint32_t status 		/* current GBIC status */
	)
	{
	uint32_t	txBitIndex = GS_TX_STARTING_BIT; /* GBIC_STATUS bits 19:10 */
	int 	port;
	
	for (port = 0; port < N_PORTS; port++, txBitIndex <<= 1) 
		{
		if (irq & txBitIndex) 
			{  /* bit changed for this port */
		    if (status & txBitIndex) 
		    	{
                dprintf("Tx fault appeared on MAC%d port%d.\n",
                          pDev->devId, port);
				} 
			else 
				{
	            dprintf("Tx fault disappeared on MAC%d port%d.\n ",
                          pDev->devId, port);
				}
			}
		}
	}

/*******************************************************************************
 * ixd2810ModDefIsr - ISR to serve the IXD2810 MAC Mod Def interrupt
 *
 * RETURNS:  N/A
 */
void ixd2810ModDefIsr
	(
	IXD2810_DEV *pDev 	/* pointer to servicing device */
	)
	{
	uint32_t modBitIndex = GS_MOD_STARTING_BIT;		
	uint32_t status;
	uint32_t irq;
	int port;
    
    dprintf("IMC\n");
	
	FOREVER
		{
		/* exclusive access */
		semTake(pDev->syncSem, WAIT_FOREVER);
	    dprintf("IMC1\n");

		/* Read GBIC status register  */
		IXD2810_REG_GET(pDev->vbase + GBIC_ADRS(GBIC_STATUS), status); /*  bits 9:0 */

		/* Find which bit(s) changed */
		irq = status ^ pDev->gbicStatus1;
	    dprintf("IMC2 irq= %x\n",irq);

		for (port = 0; port < N_PORTS; port++, modBitIndex <<= 1) 
			{
	    	dprintf("IMC3 modBitIndex= %x\n",modBitIndex);
			if (irq & modBitIndex)   /* bit changed for this port */
				{
				/* if bit=1, module is not present */
				if (status & modBitIndex) 
            		{
					ixd2810DisablePort(pDev, port);	
					dprintf("Module removed from MAC%d port%d.\n", 
			  			pDev->devId, port);
					autoFlag[pDev->devId][port] = 0;
					}
				else 
		    		{		/* if bit=0, module is present */
					dprintf("Module inserted into MAC%d port%d.\n",
						pDev->devId, port);
					autoFlag[pDev->devId][port] = 1;		
					}
				} /* end if */
			} /* end for */
	
			/* auto-negotiation each port */
			for (port = 0; port < N_PORTS; port++) 
				{
	    		dprintf("IMC4 port= %x\n",port);
				if (autoFlag[pDev->devId][port] == 1) 
					{
					/* Start auto-negotiation */
					dprintf("Start autoneg!\n");
#ifndef DEBUG
					doAutoneg(pDev, port);	
#endif
					/* Clear auto-negotiation flag */
					autoFlag[pDev->devId][port] = 0;
					break;
					}
				}
			} /* FOREVER */
		}


/*******************************************************************************
 * ixd2810DisablePort - Disable MAC port and set link down.
 *
 * RETURNS:  N/A
 */
void ixd2810DisablePort
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port /* port number. From 0 to 9. */
	) 
	{

	uint32_t data;

	/* sanity check */
	if (port >= N_PORTS)
		return;

	/* bits 9:0 for port enable. 0=disable */
	IXD2810_REG_GET(pDev->vbase + GLOBAL_ADRS(PORT_ENABLE), data);
	data &= (ENABLE_BOUNDARY - PORT_EN(port));
	IXD2810_REG_SET(pDev->vbase + GLOBAL_ADRS(PORT_ENABLE), data);

	/* bits 9:0 for link up. 0=down */
	IXD2810_REG_GET(pDev->vbase + GLOBAL_ADRS(LINK_LED_ENABLE), data);
	data &= (ENABLE_BOUNDARY - LINK_UP(port));
	IXD2810_REG_SET(pDev->vbase + GLOBAL_ADRS(LINK_LED_ENABLE), data);
}

/*******************************************************************************
 *	ixd2810EnablePort - Enable MAC port and set link up.
 *
 *	RETURNS: N/A
 */
void ixd2810EnablePort
	(
	IXD2810_DEV *pDev, /* device with corresponding port to service */
	uint32_t port /* port number. From 0 to 9. */
	) 
	{
	
	uint32_t data;
    
	/* bits 9:0 for port enable. 1=enable */
	
    IXD2810_REG_GET(pDev->vbase + GLOBAL_ADRS(PORT_ENABLE), data);
/*	data |= (ENABLE_BOUNDARY - PORT_EN(port));*/
	data |= PORT_EN(port);
    IXD2810_REG_SET(pDev->vbase + GLOBAL_ADRS(PORT_ENABLE), data);

	/* bits 9:0 for link up. 1=up */
    IXD2810_REG_GET(pDev->vbase + GLOBAL_ADRS(LINK_LED_ENABLE), data);
/*	data |= (ENABLE_BOUNDARY - LINK_UP(port));*/
	data |= LINK_UP(port);  
    IXD2810_REG_SET(pDev->vbase + GLOBAL_ADRS(LINK_LED_ENABLE),data);
}


