/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2002 Intel Corporation. All Rights Reserved.
*****************************************************************************/


/*****************************************************************64
 *     Name:    OSAL C source code file
 *  
 *    Type:     C  source file
 * 
 *    Desc:     This file contains the C source code for OSAL
 *              primitives. The OSAL API mapping code for VxWorks   
 *              is bounded by #ifdef IX_VXWORKS flag. The OSAL 
 *              API mapping code for Linux-user space is bounded 
 *              by #ifdef IX_LINUX_USR_SPACE flag.  
 *
 *    File:     osal.c 
 *
 *    Sid:      osal.c@@/main/1 - Tue Dec 10 10:33:06 2002
 *
 *    Prg:      shabbir ali 
 *
*****************************************************************/
/* VxWorks Specific include files */
#ifdef IX_VXWORKS
#include "VxWorks.h"      /* Main VxWorks defines */
#include "semLib.h"       /* Binary Semaphore */
#include "sysLib.h"       /* Tick Counters */
#include "cacheLib.h"     /* Cache Safe Memory */
#include "stdlib.h"       /* Allocate/Deallocate Memory */
#include "memLib.h"       /* Alocate/Deallocate Memory */
#include "taskLib.h"      /* Task Create/Delete/Exit */
#endif                    /* IX_VXWORKS */

/* Linux Specific include files */
#ifdef IX_LINUX_USR_SPACE
/* Include files for Linux */
#include "pthread.h"  
#include "sys/timeb.h" /* Tick Counters */
#include "stdlib.h"    /* Allocate/Deallocate Memory */
#include "errno.h"     /* Error Code Header File */
#endif                 /* IX_LINUX_USR_SPACE */

/* Common Include files */
#include "ix_osal.h"         /* OSAL header file */

#ifdef IX_VXWORKS
#include "ix_osal_VxWorks.h" /* OSAL VxWorks Sepcific Header File */
#endif                    /* IX_VXWORKS */


/*****************************************************************
 *
 *      Fun:   IX_CreateBinarySem 
 *
 *      Desc:  Create a new instance of Binary Semaphore. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_OUT_OF_MEMORY  - Failure(out of memory)
 *             IX_BAD_PARAMS     - Failure(invalid parameter)
 *             IX_FAILED         - Failure(failed to allocate more sem)   
 *
 *      Notes: Semaphores are create in Empty state. 
 *
 *      File:  osal.c
 *
*****************************************************************/

#ifndef IX_VXWORKS
IX_Err_t IX_CreateBinarySem(IX_Sem_t *s)
{

	/* Linux - User Space  */
#ifdef IX_LINUX_USR_SPACE
	return(IX_CreateBinarySem_LnxUsrSpc(s));
#endif /* IX_LINUX_USR_SPACE */

}
#endif /* IX_VXWORKS */




/*****************************************************************
 *
 *      Fun:   IX_DestroyBinarySem 
 *
 *      Desc:  Destroy an instance of Binary Semaphore. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_UNKNOWN_HANDLE - Failure(invalid parameter)
 *             IX_FAILED         - Failure(failed to destroy the sem)   
 *
 *      Notes:  
 *
 *      File:  osal.c
 *
*****************************************************************/

IX_Err_t IX_DestroyBinarySem(IX_Sem_t s)
{

	/* VxWorks */
#ifdef IX_VXWORKS
	return(IX_DestroyBinarySem_VxWrks(s));
#endif /* IX_VXWORKS */


	/* Linux - User Space */
#ifdef IX_LINUX_USR_SPACE
	return(IX_DestroyBinarySem_LnxUsrSpc(s));
#endif /* IX_LINUX_USR_SPACE */
}


/*****************************************************************
 *
 *      Fun:   IX_GiveBinarySem 
 *
 *      Desc:  Give/Unlock an instance of Binary Semaphore. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_UNKNOWN_HANDLE - Failure(invalid parameter)
 *             IX_FAILED         - Failure(semaphore is already in 
 *                                 given state)   
 *
 *      Notes:  
 *
 *      File:  osal.c
 *
*****************************************************************/

#ifndef IX_LINUX_USR_SPACE
IX_Err_t IX_GiveBinarySem (IX_Sem_t s)
{

	/* VxWorks */
#ifdef IX_VXWORKS
	return(IX_GiveBinarySem_VxWrks(s));
#endif /* IX_VXWORKS */


}
#endif /* IX_LINUX_USR_SPACE */


/*****************************************************************
 *
 *      Fun:   IX_TakeBinarySem 
 *
 *      Desc:  Take/Lock an instance of Binary Semaphore. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_UNKNOWN_HANDLE - Failure(invalid parameter)
 *             IX_FAILED         - Failure(Semaphore already in 
 *                                 lock state and block = IX_TRUE)   
 *
 *      Notes:  
 *
 *      File:  osal.c
 *
*****************************************************************/

IX_Err_t IX_TakeBinarySem(IX_Sem_t s, IX_Boolean_t block)
{

	/* VxWorks */
#ifdef IX_VXWORKS
	return(IX_TakeBinarySem_VxWrks(s,block));
#endif /* IX_VXWORKS */


	/* Linux - User Space */
#ifdef IX_LINUX_USR_SPACE
	return(IX_TakeBinarySem_LnxUsrSpc(s,block));
#endif /* IX_LINUX_USR_SPACE */
}


/*****************************************************************
 *
 *      Fun:   IX_Int_Lock
 *
 *      Desc:  Lock/Disable all the interrupts in the system. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_UNKNOWN_ERROR  - Failure(failed to lock interrupts)   
 *
 *      Notes:  
 *
 *      File:  osal.c
 *
*****************************************************************/

IX_Err_t IX_Int_Lock(IX_U32 *lockId)
{

	/* VxWorks */
#ifdef IX_VXWORKS
	return(IX_Int_Lock_VxWrks(lockId));
#endif /* IX_VXWORKS */


	/* Linux - User Space */
#ifdef IX_LINUX_USR_SPACE
	return(IX_Int_Lock_LnxUsrSpc(lockId));
#endif /* IX_LINUX_USR_SPACE */
}

/*****************************************************************
 *
 *      Fun:   IX_Int_Unlock
 *
 *      Desc:  Unlock/Enable the interrupts in the system. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_UNKNOWN_ERROR  - Failure(failed to unlock interrupts)   
 *
 *      Notes:  
 *
 *      File:  osal.c
 *
*****************************************************************/

IX_Err_t IX_Int_Unlock(IX_U32 lockId)
{

	/* VxWorks */
#ifdef IX_VXWORKS
	return(IX_Int_Unlock_VxWrks(lockId));
#endif /* IX_VXWORKS */


	/* Linux - User Space */
#ifdef IX_LINUX_USR_SPACE
	return(IX_Int_Unlock_LnxUsrSpc(lockId));
#endif /* IX_LINUX_USR_SPACE */
}

/*****************************************************************
 * 
 *      Fun:   IX_Thread_Spawn
 *
 *      Desc:  Spawn a new Task/Thread. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_OUT_OF_MEMORY  - Failure(out of memory)
 *             IX_BAD_PARAMS     - Failure(invalid parameter)
 *             IX_FAILED         - Failure(failed to allocate more sem)   
 *
 *      Notes:  
 *
 *      File:  osal.c
 *
*****************************************************************/

IX_Err_t IX_Thread_Spawn(IX_Function_Ptr_t f, IX_Thread_Info_t *t,
								 IX_Thread_id *i)
{

	/* VxWorks */
#ifdef IX_VXWORKS
	return(IX_Thread_Spawn_VxWrks(f,t,i));
#endif /* IX_VXWORKS */


	/* Linux - User Space */
#ifdef IX_LINUX_USR_SPACE
	return(IX_Thread_Spawn_LnxUsrSpc(f,t,i));	
#endif /* IX_LINUX_USR_SPACE */
}

/*****************************************************************
 *
 *      Fun:   IX_Thread_Stop
 *
 *      Desc:  Stop a Task/Thread. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_UNKNOWN_HANDLE - Failure(Unknown Thread ID)   
 *             IX_FAILED         - Failure(failed to Stop Task/Thread)   
 *
 *      Notes:  
 *
 *      File:  osal.c
 *
*****************************************************************/

IX_Err_t IX_Thread_Stop(IX_Thread_id i)
{

	/* VxWorks */
#ifdef IX_VXWORKS
	return(IX_Thread_Stop_VxWrks(i));
#endif /* IX_VXWORKS */

	/* Linux - User Space */
#ifdef IX_LINUX_USR_SPACE
	return(IX_Thread_Stop_LnxUsrSpc(i));
#endif /* IX_LINUX_USR_SPACE */

}

/*****************************************************************
 *
 *      Fun:   IX_GetLowResClockCount 
 *
 *      Desc:  Get the Tick Count of the low resolution clock 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_FAILED         - Failure(No Clock Tick 
 *											  functionality supported)   
 *
 *      Notes:  
 *
 *
 *      File:  osal.c
 *
*****************************************************************/

#ifndef IX_VXWORKS
IX_Err_t  IX_GetLowResClockCount(IX_U32 *currentTicks)
{

	/* Linux - User Space */
#ifdef IX_LINUX_USR_SPACE
	return(IX_GetLowResClockCount_LnxUsrSpc(currentTicks));
#endif /* IX_LINUX_USR_SPACE */

}
#endif /* ndef IX_VXWORKS */

/*****************************************************************
 *
 *      Fun:   IX_GetHighResClockCount 
 *
 *      Desc:  Get the Tick Count of the low resolution clock 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_FAILED         - Failure(No Clock Tick 
 *											  functionality supported)   
 *
 *      Notes:  
 *
 *
 *      File:  osal.c
 *
*****************************************************************/

#ifndef IX_VXWORKS
IX_Err_t  IX_GetHighResClockCount(IX_U32 *currentTicks)
{
	/* Linux - User Space */
#ifdef IX_LINUX_USR_SPACE
	return(IX_GetHighResClockCount_LnxUsrSpc(currentTicks));
#endif /* IX_LINUX_USR_SPACE */

}
#endif /* ndef IX_VXWORKS */

/**********************************************************************
 
     ver       pat    init                  description
------------ -------- ---- ----------------------------------------------
$SID$          ---    sali 1. initial release.
  
**********************************************************************/
