/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2002 Intel Corporation. All Rights Reserved.
*****************************************************************************/


/*****************************************************************64
 *     Name:    OSAL C source code file for Linux User Space
 *  
 *    Type:     C  source file
 * 
 *    Desc:     This file contains the C source code for OSAL
 *              primitives for Linux User Space.
 *					 The OSAL API mapping code for Linux-user space 
 *					 is bounded by #ifdef IX_LINUX_USR_SPACE flag. 
 *                
 *
 *    File:     ix_osal_Linux.c 
 *
 *    Sid:      ix_osal_Linux.c@@/main/1 - Tue Dec 10 10:33:03 2002
 *
 *    Prg:      shabbir ali 
 *
*****************************************************************/
#ifdef IX_LINUX_USR_SPACE

/* Common Include files */
#include "ix_osal.h"         /* OSAL header file */

/*****************************************************************
 *
 *      Fun:   IX_CreateBinarySem_LnxUsrSpc 
 *
 *      Desc:  Create a new instance of Binary Semaphore. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_OUT_OF_MEMORY  - Failure(out of memory)
 *             IX_BAD_PARAMS     - Failure(invalid parameter)
 *             IX_FAILED         - Failure(failed to allocate more sem)   
 *
 *      Notes: Semaphores are create in Empty state. 
 *
 *      File:  ix_osal_Linux.c
 *
*****************************************************************/

IX_Err_t IX_CreateBinarySem_LnxUsrSpc(IX_Sem_t *s)
{
	/* Check if s is NULL */
	if(s == NULL)
	{
		return(IX_BAD_PARAM);
	}

	/* allocate memory for the semaphore */
	*s = malloc(sizeof(pthread_mutex_t));

	if(*s == NULL)
		return(IX_OUT_OF_MEMORY);

	/* Initialize the mutex */
	pthread_mutex_init(*s,NULL);

	/* Lock the mutex to bring it to empty state */
	return((pthread_mutex_lock(*s) == EINVAL) ? IX_BAD_PARAM : IX_NO_ERROR);
}




/*****************************************************************
 *
 *      Fun:   IX_DestroyBinarySem_LnxUsrSpc 
 *
 *      Desc:  Destroy an instance of Binary Semaphore. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_UNKNOWN_HANDLE - Failure(invalid parameter)
 *             IX_FAILED         - Failure(failed to destroy the sem)   
 *
 *      Notes:  
 *
 *      File:  ix_osal_Linux.c
 *
*****************************************************************/

IX_Err_t IX_DestroyBinarySem_LnxUsrSpc(IX_Sem_t s)
{

	/* Check if s is NULL */
	if( s == NULL)
	{
		return(IX_UNKNOWN_HANDLE);
	}

	/* Unlock the mutex */
	if( (pthread_mutex_unlock(s)) == EINVAL)
	{
		return (IX_UNKNOWN_HANDLE);
	}

	/* Destoy the mutex */
	if( (pthread_mutex_destroy(s)) == EBUSY)
	{
		return(IX_FAILED);
	}

	/* Free the memory Allocated for the mutex */
	free(s);
	
	return(IX_NO_ERROR);
}



/*****************************************************************
 *
 *      Fun:   IX_TakeBinarySem_LnxUsrSpc 
 *
 *      Desc:  Take/Lock an instance of Binary Semaphore. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_UNKNOWN_HANDLE - Failure(invalid parameter)
 *             IX_FAILED         - Failure(Semaphore already in 
 *                                 lock state and block = IX_TRUE)   
 *
 *      Notes:  
 *
 *      File:  ix_osal_Linux.c
 *
*****************************************************************/

IX_Err_t IX_TakeBinarySem_LnxUsrSpc(IX_Sem_t s, IX_Boolean_t block)
{

	/* Check if s is NULL */
	if( s == NULL)
	{
		return(IX_UNKNOWN_HANDLE);
	}

	/* Check if block is false */
	if( block == IX_FALSE)
	{
		/* Non-Blocking Call */
		switch(pthread_mutex_trylock(s))
		{
			case EBUSY:
				return(IX_FAILED);
				break;
			case EINVAL:
				return(IX_UNKNOWN_HANDLE);
				break;
			default:
				return(IX_NO_ERROR);
				break;
		}
	}
	else
	{
		/* Blocking Call */
		if(pthread_mutex_lock(s) == EINVAL)
			return(IX_UNKNOWN_HANDLE);
		else
			return(IX_NO_ERROR);
	}

}




/*****************************************************************
 *
 *      Fun:   IX_Int_Lock_LnxUsrSpc
 *
 *      Desc:  Lock/Disable all the interrupts in the system. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_UNKNOWN_ERROR  - Failure(failed to lock interrupts)   
 *
 *      Notes:  
 *
 *      File:  ix_osal_Linux.c
 *
*****************************************************************/

IX_Err_t IX_Int_Lock_LnxUsrSpc(IX_U32 *lockId)
{

	IX_U32 regVal = 0;
	
	/* Save the register value before locking the interrupts */
	save_flags(regVal); 

	/* Lock/Disable the interrupts */
	cli();  

	/* Save the register value in the lockId */
	*lockId = regVal;

	return(IX_NO_ERROR);
}

/*****************************************************************
 *
 *      Fun:   IX_Int_Unlock_LnxUsrSpc
 *
 *      Desc:  Unlock/Enable the interrupts in the system. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_UNKNOWN_ERROR  - Failure(failed to unlock interrupts)   
 *
 *      Notes:  
 *
 *      File:  ix_osal_Linux.c
 *
*****************************************************************/

IX_Err_t IX_Int_Unlock_LnxUsrSpc(IX_U32 lockId)
{


	/* restore the register value to the one before disabling interrupts */
	restore_flags(lockId); 

	/* Unlock/Enable the interrupts in the system */
	sti(); 

	return(IX_NO_ERROR);
}

/*****************************************************************
 *
 *      Fun:   IX_Thread_Spawn_LnxUsrSpc
 *
 *      Desc:  Spawn a new Task/Thread. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_OUT_OF_MEMORY  - Failure(out of memory)
 *             IX_BAD_PARAMS     - Failure(invalid parameter)
 *             IX_FAILED         - Failure(failed to allocate more sem)   
 *
 *      Notes:  
 *
 *      File:  ix_osal_Linux.c
 *
*****************************************************************/

IX_Err_t IX_Thread_Spawn_LnxUsrSpc(IX_Function_Ptr_t f, IX_Thread_Info_t *t,
								 IX_Thread_id *i)
{

	pthread_attr_t attr = {0};      /* Pthread Attribute */
	struct sched_param param; /* Scheduling Parameter */
	IX_Linux_Conv *temp = NULL;       /* Temporary Linux Conversion */

	/* Allocate the memory for the temp variable */
	temp = malloc(sizeof(IX_Linux_Conv));
	if(temp == NULL)
		return(IX_FAILED);

	/* Check if the function pointer passed is not NULL */
	if(f == NULL)
		return(IX_BAD_PARAM);

	/* Store the function pointer in the temp structure */
	temp->f = f;
	temp->args = t->arg;

	/* Initialize the thread attributes to default value. This
    * sets the following attributes values:-
    * scheduling policy  - SCHED_OTHER (regular non-real time)
    * contention         - system wide
    * cancellation state - PTHREAD_CANCEL_ENABLE
    * cancellation type  - PTHREAD_CANCEL_DEFERRED 
   */
	if( (pthread_attr_init(&attr)) == ENOMEM)
	{
		/* free the memory allocated for the temp variable */
		free(temp);

		return(IX_FAILED);
	}

	/* Set the Thread Priority -- The Linux man page for 
	 * pthread_attr_setschedparam says that for scheduling policy
	 * of type SCHED_OTHER, priority should be set to zero. The 
	 * actual priority value is used only for real time and FIFO
	 * scheduling policies */

	param.sched_priority = 0;
	if( (pthread_attr_setschedparam(&attr, &param)) == EINVAL)
	{
		/* free the memory allocated for the temp variable */
		free(temp);

		return(IX_BAD_PARAM);
	}

	/* Create a new thread with the attribute values set as above */
	if( (pthread_create(i,&attr,ixConversionFunc_LnxUsrSpc,(void *)temp)) == EAGAIN)
	{
		/* free the memory allocated for the temp variable */
		free(temp);

		return(IX_FAILED);
	}

	/* If successfull we are not freeing the memory for temp as it will be done in the conversion function. Else we might have a race condition */

	return(IX_NO_ERROR);	

}
/*****************************************************************
 *
 *      Fun:   ixConversionFunc_LnxUsrSpc 
 *
 *      Desc:  This function is used only for Linux. This function
 *					is used to set the cancellation type of the new thread 
 *					and then call the actual function pointer. 
 *
 *      Ret:   NONE 
 *
 *      Notes:  
 *
 *      File:  ix_osal_Linux.c
 *
*****************************************************************/
void *ixConversionFunc_LnxUsrSpc(void *arg)
{
	IX_Linux_Conv temp = {0,0};  		/* Temporary Linux Conversion */
	IX_Linux_Conv *temp1 = NULL;     /* Temporary Linux Conversion */

	temp1 = (IX_Linux_Conv *)arg;

	temp.f = temp1->f;
	temp.args = temp1->args;

	/* free the memory allocated for the temp variable */
	free(arg);

	/* Set the cancellation type of the thread as asynchronous --
	 * i.e. It allows the thread to be deleted without waiting to 
	 * reach a cancellation point */
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);

	if(temp.f != NULL)
	{
		/* Call the actual entry point Function */
		return( (temp.f)(temp.args) ); 
	}
	return(NULL);
	
}

/*****************************************************************
 *
 *      Fun:   IX_Thread_Stop_LnxUsrSpc
 *
 *      Desc:  Stop a Task/Thread. 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_UNKNOWN_HANDLE - Failure(Unknown Thread ID)   
 *             IX_FAILED         - Failure(failed to Stop Task/Thread)   
 *
 *      Notes:  
 *
 *      File:  ix_osal_Linux.c
 *
*****************************************************************/

IX_Err_t IX_Thread_Stop_LnxUsrSpc(IX_Thread_id i)
{

	
	/* Signal the thread to exit */
	if( (pthread_cancel(i)) == ESRCH) 
	{
		return(IX_UNKNOWN_HANDLE);
	}

	/* Wait for the thread to exit */
	switch(pthread_join(i,NULL))
	{
		case ESRCH:
			return(IX_UNKNOWN_HANDLE);
			break;
		case EINVAL:
		case EDEADLK:
			return(IX_FAILED);
			break;
		default:	
			return(IX_NO_ERROR);
			break;
	}
	
}


/*****************************************************************
 *
 *      Fun:   IX_GetLowResClockCount_LnxUsrSpc 
 *
 *      Desc:  Get the Tick Count of the low resolution clock 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_FAILED         - Failure(No Clock Tick 
 *											  functionality supported)   
 *
 *      Notes:  
 *
 *
 *      File:  ix_osal_Linux.c
 *
*****************************************************************/

IX_Err_t  IX_GetLowResClockCount_LnxUsrSpc(IX_U32 *currentTicks)
{

	struct timeb temp = {0,0,0};
	/* Get the system clock time elapsed since epoch */
	ftime(&temp);

	/* Compute the Tick Count from the clock time */
	*currentTicks = ((temp.time * 1000) + temp.millitm) /
							IX_LOW_RES_CLOCK_FREQ;

	return(IX_NO_ERROR);
}

/*****************************************************************
 *
 *      Fun:   IX_GetHighResClockCount_LnxUsrSpc 
 *
 *      Desc:  Get the Tick Count of the low resolution clock 
 *
 *      Ret:   IX_NO_ERROR       - Success
 *             IX_FAILED         - Failure(No Clock Tick 
 *											  functionality supported)   
 *
 *      Notes:  
 *
 *
 *      File:  ix_osal_Linux.c
 *
*****************************************************************/

IX_Err_t  IX_GetHighResClockCount_LnxUsrSpc(IX_U32 *currentTicks)
{

	struct timeb temp = {0,0,0};
	/* Get the system clock time elapsed since epoch */
	ftime(&temp);

	/* Compute the Tick Count from the clock time */
	*currentTicks = ((temp.time * 1000) + temp.millitm) /
							IX_HIGH_RES_CLOCK_FREQ;

	return(IX_NO_ERROR);
}

#endif /* IX_LINUX_USR_SPACE */
/**********************************************************************
 
     ver       pat    init                  description
------------ -------- ---- ----------------------------------------------
$SID$          ---    sali 1. initial release.
  
**********************************************************************/
